/*
 * OmniVision OV511 Camera-to-USB Bridge Driver
 *
 * Copyright (c) 1999-2001 Mark W. McClelland
 * Original decompression code Copyright 1998-2000 OmniVision Technologies
 * Many improvements by Bret Wallach <bwallac1@san.rr.com>
 * Color fixes by by Orion Sky Lawlor <olawlor@acm.org> (2/26/2000)
 * Snapshot code by Kevin Moore
 * OV7620 fixes by Charl P. Botha <cpbotha@ieee.org>
 * Changes by Claudio Matsuoka <claudio@conectiva.com>
 *
 * Based on the Linux CPiA driver written by Peter Pregler,
 * Scott J. Bertin and Johannes Erdfelt.
 * 
 * Please see the file: linux/Documentation/usb/ov511.txt 
 * and the website at:  http://alpha.dyndns.org/ov511
 * for more info.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

static const char version[] = "1.35";

#define __NO_VERSION__

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
#include <linux/proc_fs.h>
#include <linux/ctype.h>
#include <linux/pagemap.h>
#include <linux/usb.h>
#include <asm/io.h>
#include <asm/semaphore.h>
#include <linux/wrapper.h>

#include "ov511.h"

#define OV511_I2C_RETRIES 3
#define ENABLE_OV511_COMPRESSION
#define ENABLE_Y_QUANTABLE 1
#define ENABLE_UV_QUANTABLE 1

/* Video Size 640 x 480 x 3 bytes for RGB */
#define MAX_FRAME_SIZE (640 * 480 * 3)
#define MAX_DATA_SIZE (MAX_FRAME_SIZE + sizeof(struct timeval))

/* Max size * bytes per YUV420 pixel (1.5) + one extra isoc frame for safety */
#define MAX_RAW_DATA_SIZE (((640 * 480 * 3) / 2) + 1024)

#define GET_SEGSIZE(p) ((p) == VIDEO_PALETTE_GREY ? 256 : 384)

/* PARAMETER VARIABLES: */
static int autoadjust = 1;    /* CCD dynamically changes exposure, etc... */

/* 0=no debug messages
 * 1=init/detection/unload and other significant messages,
 * 2=some warning messages
 * 3=config/control function calls
 * 4=most function calls and data parsing messages
 * 5=highly repetitive mesgs
 * NOTE: This should be changed to 0, 1, or 2 for production kernels
 */
static int debug = 0;

/* Fix vertical misalignment of red and blue at 640x480 */
static int fix_rgb_offset = 0;

/* Snapshot mode enabled flag */
static int snapshot = 0;

/* Sensor detection override (global for all attached cameras) */
static int sensor = 0;

/* Increase this if you are getting "Failed to read sensor ID..." */
static int i2c_detect_tries = 5;

/* For legal values, see the OV7610/7620 specs under register Common F,
 * upper nybble  (set to 0-F) */
static int aperture = -1;

/* Force image to be read in RGB instead of BGR. This option allow
 * programs that expect RGB data (e.g. gqcam) to work with this driver. */
static int force_rgb = 0;

/* Number of seconds before inactive buffers are deallocated */
static int buf_timeout = 5;

/* Number of cameras to stream from simultaneously */
static int cams = 1;

/* Prevent apps from timing out if frame is not done in time */
static int retry_sync = 0;

/* Enable compression. Does not work reliably yet, and needs a fast CPU. */
static int compress = 0;

/* Display test pattern - doesn't work yet either */
static int testpat = 0;

/* Setting this to 1 will make the sensor output GBR422 instead on YUV420. Only
 * affects RGB24 mode. */
static int sensor_gbr = 0;

/* Dump raw pixel data, in one of 3 formats. See ov511_dumppix() for details. */
static int dumppix = 0;

/* LED policy. Only works on some OV511+ cameras. 0=off, 1=on (default), 2=auto
 * (on when open) */
static int led = 1;

/* Set this to 1 to dump the bridge register contents after initialization */
static int dump_bridge = 0;

/* Set this to 1 to dump the sensor register contents after initialization */
static int dump_sensor = 0;

/* Temporary option for debugging "works, but no image" problem. Prints the
 * first 12 bytes of data (potentially a packet header) in each isochronous
 * data frame. */
static int printph = 0;

/* Compression parameters - I'm not exactly sure what these do yet */
static int phy = 0x1f;
static int phuv = 0x05;
static int pvy = 0x06;
static int pvuv = 0x06;
static int qhy = 0x14;
static int qhuv = 0x03;
static int qvy = 0x04;
static int qvuv = 0x04;

/* Light frequency. Set to 50 or 60 (Hz), or zero for default settings */
static int lightfreq = 0;

/* Set this to 1 to enable banding filter by default. Compensates for alternating
 * horizontal light/dark bands caused by (usually fluorescent) lights */
static int bandingfilter = 0;

/* Set this to 1 to do post-processing (decompression, format conversion, etc...) at
 * task-time instead of interrupt-time. This only supports RGB and YUV420/420P, and may
 * increase frame latency since all processing is deferred until the frame is totally
 * received. If the benefits of this (simpler code, less lost interrupts) outweigh the
 * costs, this will be default in the future.
 */
static int ttpp = 0;

MODULE_PARM(autoadjust, "i");
MODULE_PARM_DESC(autoadjust, "CCD dynamically changes exposure");
MODULE_PARM(debug, "i");
MODULE_PARM_DESC(debug, "Debug level: 0=none, 1=init/detection, 2=warning, 3=config/control, 4=function call, 5=max");
MODULE_PARM(fix_rgb_offset, "i");
MODULE_PARM_DESC(fix_rgb_offset, "Fix vertical misalignment of red and blue at 640x480");
MODULE_PARM(snapshot, "i");
MODULE_PARM_DESC(snapshot, "Enable snapshot mode");
MODULE_PARM(sensor, "i");
MODULE_PARM_DESC(sensor, "Override sensor detection");
MODULE_PARM(i2c_detect_tries, "i");
MODULE_PARM_DESC(i2c_detect_tries, "Number of tries to detect sensor");
MODULE_PARM(aperture, "i");
MODULE_PARM_DESC(aperture, "Read the OV7610/7620 specs");
MODULE_PARM(force_rgb, "i");
MODULE_PARM_DESC(force_rgb, "Read RGB instead of BGR");
MODULE_PARM(buf_timeout, "i");
MODULE_PARM_DESC(buf_timeout, "Number of seconds before buffer deallocation");
MODULE_PARM(cams, "i");
MODULE_PARM_DESC(cams, "Number of simultaneous cameras");
MODULE_PARM(retry_sync, "i");
MODULE_PARM_DESC(retry_sync, "Prevent apps from timing out");
MODULE_PARM(compress, "i");
MODULE_PARM_DESC(compress, "Turn on compression (not reliable yet)");
MODULE_PARM(testpat, "i");
MODULE_PARM_DESC(testpat, "Replace image with vertical bar testpattern (only partially working)");
MODULE_PARM(sensor_gbr, "i");
MODULE_PARM_DESC(sensor_gbr, "Make sensor output GBR422 rather than YUV420");
MODULE_PARM(dumppix, "i");
MODULE_PARM_DESC(dumppix, "Dump raw pixel data, in one of 3 formats. See ov511_dumppix() for details");
MODULE_PARM(led, "i");
MODULE_PARM_DESC(led, "Sets LED policy. Only works on some OV511+ cameras. 0=off, 1=on (default), 2=auto (on when open)");
MODULE_PARM(dump_bridge, "i");
MODULE_PARM_DESC(dump_bridge, "Dump the bridge registers");
MODULE_PARM(dump_sensor, "i");
MODULE_PARM_DESC(dump_sensor, "Dump the sensor registers");
MODULE_PARM(printph, "i");
MODULE_PARM_DESC(printph, "Print the (potential) packet header of each isoc frame");
MODULE_PARM(phy, "i");
MODULE_PARM_DESC(phy, "Prediction range (horizontal Y)");
MODULE_PARM(phuv, "i");
MODULE_PARM_DESC(phuv, "Prediction range (horizontal UV)");
MODULE_PARM(pvy, "i");
MODULE_PARM_DESC(pvy, "Prediction range (vertical Y)");
MODULE_PARM(pvuv, "i");
MODULE_PARM_DESC(pvuv, "Prediction range (vertical UV)");
MODULE_PARM(qhy, "i");
MODULE_PARM_DESC(qhy, "Quantization threshold (horizontal Y)");
MODULE_PARM(qhuv, "i");
MODULE_PARM_DESC(qhuv, "Quantization threshold (horizontal UV)");
MODULE_PARM(qvy, "i");
MODULE_PARM_DESC(qvy, "Quantization threshold (vertical Y)");
MODULE_PARM(qvuv, "i");
MODULE_PARM_DESC(qvuv, "Quantization threshold (vertical UV)");
MODULE_PARM(lightfreq, "i");
MODULE_PARM_DESC(lightfreq, "Light frequency. Set to 50 or 60 Hz, or zero for default settings");
MODULE_PARM(bandingfilter, "i");
MODULE_PARM_DESC(bandingfilter, "Enable banding filter (to reduce effects of fluorescent lighting)");
MODULE_PARM(ttpp, "i");
MODULE_PARM_DESC(ttpp, "Do post-processing at task-time instead of interrupt-time (experimental)");

MODULE_AUTHOR("Mark McClelland <mmcclell@bigfoot.com> & Bret Wallach & Orion Sky Lawlor <olawlor@acm.org> & Kevin Moore & Charl P. Botha <cpbotha@ieee.org> & Claudio Matsuoka <claudio@conectiva.com>");
MODULE_DESCRIPTION("OV511 USB Camera Driver");

static struct usb_driver ov511_driver;

/* Function prototypes */
static void ov51x_clear_snapshot(struct usb_ov511 *);
static int ov51x_check_snapshot(struct usb_ov511 *);
static inline int ov7610_get_picture(struct usb_ov511 *, struct video_picture *);

/**********************************************************************
 * List of known OV511-based cameras
 **********************************************************************/

static struct cam_list clist[] = {
	{   0, "generic model (no ID)" },
	{   1, "Mustek WCam 3X" },
	{   3, "D-Link DSB-C300" },
	{   4, "generic OV511/OV7610" },
	{   5, "Puretek PT-6007" },
	{  21, "Creative Labs WebCam 3" },
	{  36, "Koala-Cam" },
	{  38, "Lifeview USB Life TV" },	/* No support yet! */
	{  41, "Samsung Anycam MPC-M10" },
	{  43, "Mtekvision Zeca MV402" },
	{ 100, "Lifeview RoboCam" },
	{ 102, "AverMedia InterCam Elite" },
	{ 112, "MediaForte MV300" },	/* or OV7110 evaluation kit */
	{  -1, NULL }
};

static __devinitdata struct usb_device_id device_table [] = {
	{ USB_DEVICE(0x05a9, 0x0511) },  /* OV511 */
	{ USB_DEVICE(0x05a9, 0xA511) },  /* OV511+ */
	{ USB_DEVICE(0x05a9, 0x0518) },  /* OV518+ */
	{ USB_DEVICE(0x0813, 0x0002) },  /* Intel Play Me2Cam OV511+ */
	{ }  /* Terminating entry */
};

MODULE_DEVICE_TABLE (usb, device_table);

#if defined(CONFIG_PROC_FS) && defined(CONFIG_VIDEO_PROC_FS)
static struct palette_list plist[] = {
	{ VIDEO_PALETTE_GREY,	"GREY" },
	{ VIDEO_PALETTE_HI240,  "HI240" },
	{ VIDEO_PALETTE_RGB565, "RGB565" },
	{ VIDEO_PALETTE_RGB24,	"RGB24" },
	{ VIDEO_PALETTE_RGB32,	"RGB32" },
	{ VIDEO_PALETTE_RGB555, "RGB555" },
	{ VIDEO_PALETTE_YUV422, "YUV422" },
	{ VIDEO_PALETTE_YUYV,   "YUYV" },
	{ VIDEO_PALETTE_UYVY,   "UYVY" },
	{ VIDEO_PALETTE_YUV420, "YUV420" },
	{ VIDEO_PALETTE_YUV411, "YUV411" },
	{ VIDEO_PALETTE_RAW,    "RAW" },
	{ VIDEO_PALETTE_YUV422P,"YUV422P" },
	{ VIDEO_PALETTE_YUV411P,"YUV411P" },
	{ VIDEO_PALETTE_YUV420P,"YUV420P" },
	{ VIDEO_PALETTE_YUV410P,"YUV410P" },
	{ -1, NULL }
};
#endif

static unsigned char yQuanTable[] = YQUANTABLE;
static unsigned char uvQuanTable[] = UVQUANTABLE;

/**********************************************************************
 *
 * Memory management
 *
 * This is a shameless copy from the USB-cpia driver (linux kernel
 * version 2.3.29 or so, I have no idea what this code actually does ;).
 * Actually it seems to be a copy of a shameless copy of the bttv-driver.
 * Or that is a copy of a shameless copy of ... (To the powers: is there
 * no generic kernel-function to do this sort of stuff?)
 *
 * Yes, it was a shameless copy from the bttv-driver. IIRC, Alan says
 * there will be one, but apparentely not yet -jerdfelt
 *
 * So I copied it again for the OV511 driver -claudio
 **********************************************************************/

/* Given PGD from the address space's page table, return the kernel
 * virtual mapping of the physical memory mapped at ADR.
 */
static inline unsigned long uvirt_to_kva(pgd_t *pgd, unsigned long adr)
{
	unsigned long ret = 0UL;
	pmd_t *pmd;
	pte_t *ptep, pte;

	if (!pgd_none(*pgd)) {
		pmd = pmd_offset(pgd, adr);
		if (!pmd_none(*pmd)) {
			ptep = pte_offset(pmd, adr);
			pte = *ptep;
			if (pte_present(pte)) {
				ret = (unsigned long) page_address(pte_page(pte));
				ret |= (adr & (PAGE_SIZE - 1));
			}
		}
	}

	return ret;
}

/* Here we want the physical address of the memory.
 * This is used when initializing the contents of the
 * area and marking the pages as reserved.
 */
static inline unsigned long kvirt_to_pa(unsigned long adr)
{
	unsigned long va, kva, ret;

	va = VMALLOC_VMADDR(adr);
	kva = uvirt_to_kva(pgd_offset_k(va), va);
	ret = __pa(kva);
	return ret;
}

static void *rvmalloc(unsigned long size)
{
	void *mem;
	unsigned long adr, page;

	/* Round it off to PAGE_SIZE */
	size += (PAGE_SIZE - 1);
	size &= ~(PAGE_SIZE - 1);

	mem = vmalloc_32(size);
	if (!mem)
		return NULL;

	memset(mem, 0, size); /* Clear the ram out, no junk to the user */
	adr = (unsigned long) mem;
	while (size > 0) {
		page = kvirt_to_pa(adr);
		mem_map_reserve(virt_to_page(__va(page)));
		adr += PAGE_SIZE;
		if (size > PAGE_SIZE)
			size -= PAGE_SIZE;
		else
			size = 0;
	}

	return mem;
}

static void rvfree(void *mem, unsigned long size)
{
	unsigned long adr, page;

	if (!mem)
		return;

	size += (PAGE_SIZE - 1);
	size &= ~(PAGE_SIZE - 1);

	adr=(unsigned long) mem;
	while (size > 0) {
		page = kvirt_to_pa(adr);
		mem_map_unreserve(virt_to_page(__va(page)));
		adr += PAGE_SIZE;
		if (size > PAGE_SIZE)
			size -= PAGE_SIZE;
		else
			size = 0;
	}
	vfree(mem);
}

/**********************************************************************
 * /proc interface
 * Based on the CPiA driver version 0.7.4 -claudio
 **********************************************************************/

#if defined(CONFIG_PROC_FS) && defined(CONFIG_VIDEO_PROC_FS)

static struct proc_dir_entry *ov511_proc_entry = NULL;
extern struct proc_dir_entry *video_proc_entry;

#define YES_NO(x) ((x) ? "yes" : "no")

/* /proc/video/ov511/<minor#>/info */
static int ov511_read_proc_info(char *page, char **start, off_t off,
                                 int count, int *eof, void *data)
{
	char *out = page;
	int i, j, len;
	struct usb_ov511 *ov511 = data;
	struct video_picture p;

	ov7610_get_picture(ov511, &p);

	/* IMPORTANT: This output MUST be kept under PAGE_SIZE
	 *            or we need to get more sophisticated. */

	out += sprintf (out, "driver_version  : %s\n", version);
	out += sprintf (out, "custom_id       : %d\n", ov511->customid);
	out += sprintf (out, "model           : %s\n", ov511->desc ?
		clist[ov511->desc].description : "unknown");
	out += sprintf (out, "streaming       : %s\n", YES_NO (ov511->streaming));
	out += sprintf (out, "grabbing        : %s\n", YES_NO (ov511->grabbing));
	out += sprintf (out, "compress        : %s\n", YES_NO (ov511->compress));
	out += sprintf (out, "subcapture      : %s\n", YES_NO (ov511->sub_flag));
	out += sprintf (out, "sub_size        : %d %d %d %d\n",
		ov511->subx, ov511->suby, ov511->subw, ov511->subh);
	out += sprintf (out, "data_format     : %s\n", force_rgb ? "RGB" : "BGR");
	out += sprintf (out, "brightness      : %d\n", p.brightness >> 8);
	out += sprintf (out, "colour          : %d\n", p.colour >> 8);
	out += sprintf (out, "contrast        : %d\n", p.contrast >> 8);
	out += sprintf (out, "hue             : %d\n", p.hue >> 8);
	out += sprintf (out, "num_frames      : %d\n", OV511_NUMFRAMES);
	for (i = 0; i < OV511_NUMFRAMES; i++) {
		out += sprintf (out, "frame           : %d\n", i);
		out += sprintf (out, "  depth         : %d\n",
			ov511->frame[i].depth);
		out += sprintf (out, "  size          : %d %d\n",
			ov511->frame[i].width, ov511->frame[i].height);
		out += sprintf (out, "  format        : ");
		for (j = 0; plist[j].num >= 0; j++) {
			if (plist[j].num == ov511->frame[i].format) {
				out += sprintf (out, "%s\n", plist[j].name);
				break;
			}
		}
		if (plist[j].num < 0)
			out += sprintf (out, "unknown\n");
		out += sprintf (out, "  segsize       : %d\n",
			ov511->frame[i].segsize);
		out += sprintf (out, "  data_buffer   : 0x%p\n",
			ov511->frame[i].data);
	}
	out += sprintf (out, "snap_enabled    : %s\n", YES_NO (ov511->snap_enabled));
	out += sprintf (out, "bridge          : %s\n",
		ov511->bridge == BRG_OV511 ? "OV511" :
		ov511->bridge == BRG_OV511PLUS ? "OV511+" :
		ov511->bridge == BRG_OV518PLUS ? "OV518+" :
		"unknown");
	out += sprintf (out, "sensor          : %s\n",
		ov511->sensor == SEN_OV6620 ? "OV6620" :
		ov511->sensor == SEN_OV6630 ? "OV6630" :
		ov511->sensor == SEN_OV7610 ? "OV7610" :
		ov511->sensor == SEN_OV7620 ? "OV7620" :
		ov511->sensor == SEN_OV7620AE ? "OV7620AE" :
		ov511->sensor == SEN_OV8600 ? "OV8600" :
		ov511->sensor == SEN_KS0127 ? "KS0127" :
		"unknown");
	out += sprintf (out, "packet_size     : %d\n", ov511->packet_size);
	out += sprintf (out, "framebuffer     : 0x%p\n", ov511->fbuf);
	
	len = out - page;
	len -= off;
	if (len < count) {
		*eof = 1;
		if (len <= 0) return 0;
	} else
		len = count;

	*start = page + off;

	return len;
}

/* /proc/video/ov511/<minor#>/button
 *
 * When the camera's button is pressed, the output of this will change from a
 * 0 to a 1 (ASCII). It will retain this value until it is read, after which
 * it will reset to zero.
 * 
 * SECURITY NOTE: Since reading this file can change the state of the snapshot
 * status, it is important for applications that open it to keep it locked
 * against access by other processes, using flock() or a similar mechanism. No
 * locking is provided by this driver.
 */
static int ov511_read_proc_button(char *page, char **start, off_t off,
                                  int count, int *eof, void *data)
{
	char *out = page;
	int len, status;
	struct usb_ov511 *ov511 = data;

	status = ov51x_check_snapshot(ov511);
	out += sprintf(out, "%d", status);

	if (status)
		ov51x_clear_snapshot(ov511);

	len = out - page;
	len -= off;
	if (len < count) {
		*eof = 1;
		if (len <= 0) return 0;
	} else
		len = count;

	*start = page + off;

	return len;
}

static void create_proc_ov511_cam(struct usb_ov511 *ov511)
{
	char dirname[4];

	if (!ov511_proc_entry || !ov511)
		return;

	/* Create per-device directory */
	sprintf(dirname, "%d", ov511->vdev.minor);
	PDEBUG(4, "creating /proc/video/ov511/%s/", dirname);
	ov511->proc_devdir = create_proc_entry(dirname, S_IFDIR,
		ov511_proc_entry);
	if (!ov511->proc_devdir)
		return;

	/* Create "info" entry (human readable device information) */
	PDEBUG(4, "creating /proc/video/ov511/%s/info", dirname);
	ov511->proc_info = create_proc_read_entry("info",
		S_IFREG|S_IRUGO|S_IWUSR, ov511->proc_devdir,
		ov511_read_proc_info, ov511);
	if (!ov511->proc_info)
		return;

	/* Don't create it if old snapshot mode on (would cause race cond.) */
	if (!snapshot) {
		/* Create "button" entry (snapshot button status) */
		PDEBUG(4, "creating /proc/video/ov511/%s/button", dirname);
		ov511->proc_button = create_proc_read_entry("button",
			S_IFREG|S_IRUGO|S_IWUSR, ov511->proc_devdir,
			ov511_read_proc_button, ov511);
		if (!ov511->proc_button)
			return;
	}

	// Will create control device entry here soon
}

static void destroy_proc_ov511_cam(struct usb_ov511 *ov511)
{
	char dirname[4];
	
	if (!ov511 || !ov511->proc_devdir)
		return;

	sprintf(dirname, "%d", ov511->vdev.minor);

	// Destroy control device here eventually

	/* Destroy "button" entry */
	if (ov511->proc_button) {
		PDEBUG(4, "destroying /proc/video/ov511/%s/button", dirname);
		remove_proc_entry("button", ov511->proc_devdir);
		ov511->proc_button = NULL;
	}

	/* Destroy "info" entry */
	if (ov511->proc_info) {
		PDEBUG(4, "destroying /proc/video/ov511/%s/info", dirname);
		remove_proc_entry("info", ov511->proc_devdir);
		ov511->proc_info = NULL;
	}

	/* Destroy per-device directory */
	PDEBUG(4, "destroying /proc/video/ov511/%s/", dirname);
	remove_proc_entry(dirname, ov511_proc_entry);
	ov511->proc_devdir = NULL;
}

static void proc_ov511_create(void)
{
	/* No current standard here. Alan prefers /proc/video/ as it keeps
	 * /proc "less cluttered than /proc/randomcardifoundintheshed/"
	 * -claudio
	 */
	if (video_proc_entry == NULL) {
		err("Error: /proc/video/ does not exist");
		return;
	}

	ov511_proc_entry = create_proc_entry("ov511", S_IFDIR, video_proc_entry);

	if (ov511_proc_entry)
		ov511_proc_entry->owner = THIS_MODULE;
	else
		err("Unable to create /proc/video/ov511");
}

static void proc_ov511_destroy(void)
{
	PDEBUG (3, "removing /proc/video/ov511");

	if (ov511_proc_entry == NULL)
		return;

	remove_proc_entry("ov511", video_proc_entry);
}
#endif /* CONFIG_PROC_FS && CONFIG_VIDEO_PROC_FS */

/**********************************************************************
 *
 * Camera interface
 *
 **********************************************************************/

static int ov511_reg_write(struct usb_device *dev,
			   unsigned char reg,
			   unsigned char value)
{
	int rc;

	PDEBUG(5, "0x%02X:0x%02X", reg, value);

	rc = usb_control_msg(dev,
		usb_sndctrlpipe(dev, 0),
		2 /* REG_IO */,
		USB_TYPE_CLASS | USB_RECIP_DEVICE,
		0, (__u16)reg, &value, 1, HZ);	

	if (rc < 0)
		err("reg write: error %d", rc);

	return rc;
}

/* returns: negative is error, pos or zero is data */
static int ov511_reg_read(struct usb_device *dev, unsigned char reg)
{
	int rc;
	unsigned char buffer[1];

	rc = usb_control_msg(dev,
		usb_rcvctrlpipe(dev, 0),
		2 /* REG_IO */,
		USB_DIR_IN | USB_TYPE_CLASS | USB_RECIP_DEVICE,
		0, (__u16)reg, buffer, 1, HZ);
                               
	PDEBUG(5, "0x%02X:0x%02X", reg, buffer[0]);
	
	if (rc < 0) {
		err("reg read: error %d", rc);
		return rc;
	} else {
		return buffer[0];	
	}
}

/*
 * Writes bits at positions specified by mask to a reg. Bits that are in
 * the same position as 1's in "mask" are cleared and set to "value". Bits
 * that are in the same position as 0's in "mask" are preserved, regardless 
 * of their respective state in "value".
 */
static int ov511_reg_write_mask(struct usb_device *dev,
				unsigned char reg,
				unsigned char value,
				unsigned char mask)
{
	int ret;
	unsigned char oldval, newval;

	ret = ov511_reg_read(dev, reg);
	if (ret < 0)
		return ret;

	oldval = (unsigned char) ret;
	oldval &= (~mask);		/* Clear the masked bits */
	value &= mask;			/* Enforce mask on value */
	newval = oldval | value;	/* Set the desired bits */

	return (ov511_reg_write(dev, reg, newval));
}

/* Writes multiple (n) values to a single register. Only valid with certain
 * registers (0x30 and 0xc4 - 0xce). I don't know yet what this really does */
static int ov518_reg_write_multi(struct usb_device *dev,
				unsigned char reg,
				unsigned char *values,
				int n)
{
	int rc;

	PDEBUG(5, "0x%02X:[multiple], n=%d", reg, n);  // FIXME

	if (values == NULL) {
		err("reg write multiple: NULL buffer");
		return -EINVAL;
	}

	rc = usb_control_msg(dev,
		usb_sndctrlpipe(dev, 0),
		2 /* REG_IO */,
		USB_TYPE_CLASS | USB_RECIP_DEVICE,
		0, (__u16)reg, values, n, HZ);	

	if (rc < 0)
		err("reg write multiple: error %d", rc);

	return rc;
}

static int ov511_upload_quan_tables(struct usb_device *dev)
{
	unsigned char *pYTable = yQuanTable;
	unsigned char *pUVTable = uvQuanTable;
	unsigned char val0, val1;
	int i, rc, reg = OV511_OMNICE_Y_LUT_BEGIN;

	PDEBUG(4, "Uploading quantization tables");

	for(i = 0; i < QUANTABLESIZE / 2; i++)
	{
		if (ENABLE_Y_QUANTABLE)
		{
			val0 = *pYTable++;
			val1 = *pYTable++;
			val0 &= 0x0f;
			val1 &= 0x0f;
			val0 |= val1 << 4;
			rc = ov511_reg_write(dev, reg, val0);
			if (rc < 0)
				return rc;
		}

		if (ENABLE_UV_QUANTABLE)
		{
			val0 = *pUVTable++;
			val1 = *pUVTable++;
			val0 &= 0x0f;
			val1 &= 0x0f;
			val0 |= val1 << 4;
			rc = ov511_reg_write(dev, reg + QUANTABLESIZE / 2, val0);
			if (rc < 0)
				return rc;
		}

		reg++;
	}

	return 0;
}

/* The OV518 I2C I/O procedure is different, hence, this function.
 * This is normally only called from ov511_i2c_write(). Note that this function
 * always succeeds regardless of whether the sensor is present and working.
 */
static int ov518_i2c_write(struct usb_device *dev,
			   unsigned char reg,
			   unsigned char value)
{
	int rc;

	PDEBUG(5, "0x%02X:0x%02X", reg, value);

	/* Select camera register */
	rc = ov511_reg_write(dev, OV511_REG_I2C_SUB_ADDRESS_3_BYTE, reg);
	if (rc < 0) goto error;

	/* Write "value" to I2C data port of OV511 */
	rc = ov511_reg_write(dev, OV511_REG_I2C_DATA_PORT, value);
	if (rc < 0) goto error;

	/* Initiate 3-byte write cycle */
	rc = ov511_reg_write(dev, OV518_REG_I2C_CONTROL, 0x01);
	if (rc < 0) goto error;

	return 0;

error:
	err("ov518 i2c write: error %d", rc);
	return rc;
}

static int ov511_i2c_write(struct usb_device *dev,
			   unsigned char reg,
			   unsigned char value)
{
	int rc, retries;

	/* OV518 needs a different write algorithm */
	if (dev->descriptor.idProduct == 0x0518)
		return (ov518_i2c_write(dev, reg, value));

	PDEBUG(5, "0x%02X:0x%02X", reg, value);

	/* Three byte write cycle */
	for (retries = OV511_I2C_RETRIES; ; ) {
		/* Select camera register */
		rc = ov511_reg_write(dev, OV511_REG_I2C_SUB_ADDRESS_3_BYTE, reg);
		if (rc < 0) goto error;

		/* Write "value" to I2C data port of OV511 */
		rc = ov511_reg_write(dev, OV511_REG_I2C_DATA_PORT, value);	
		if (rc < 0) goto error;

		/* Initiate 3-byte write cycle */
		rc = ov511_reg_write(dev, OV511_REG_I2C_CONTROL, 0x01);
		if (rc < 0) goto error;

		do rc = ov511_reg_read(dev, OV511_REG_I2C_CONTROL);
		while (rc > 0 && ((rc&1) == 0)); /* Retry until idle */
		if (rc < 0) goto error;

		if ((rc&2) == 0) /* Ack? */
			break;
#if 0
		/* I2C abort */	
		ov511_reg_write(dev, OV511_REG_I2C_CONTROL, 0x10);
#endif
		if (--retries < 0) {
			err("i2c write retries exhausted");
			rc = -1;
			goto error;
		}
	}

	return 0;

error:
	err("i2c write: error %d", rc);
	return rc;
}

/* The OV518 I2C I/O procedure is different, hence, this function.
 * This is normally only called from ov511_i2c_read(). Note that this function
 * always succeeds regardless of whether the sensor is present and working.
 */
static int ov518_i2c_read(struct usb_device *dev, unsigned char reg)
{
	int rc, value;

	/* Select camera register */
	rc = ov511_reg_write(dev, OV511_REG_I2C_SUB_ADDRESS_2_BYTE, reg);
	if (rc < 0) goto error;

	/* Initiate 2-byte write cycle */
	rc = ov511_reg_write(dev, OV518_REG_I2C_CONTROL, 0x03);
	if (rc < 0) goto error;

	/* Initiate 2-byte read cycle */
	rc = ov511_reg_write(dev, OV518_REG_I2C_CONTROL, 0x05);
	if (rc < 0) goto error;

	value = ov511_reg_read(dev, OV511_REG_I2C_DATA_PORT);

	PDEBUG(5, "0x%02X:0x%02X", reg, value);

	return value;

error:
	err("ov518 i2c read: error %d", rc);
	return rc;
}

/* returns: negative is error, pos or zero is data */
static int ov511_i2c_read(struct usb_device *dev, unsigned char reg)
{
	int rc, value, retries;

	/* OV518 needs a different read algorithm */
	if (dev->descriptor.idProduct == 0x0518)
		return (ov518_i2c_read(dev, reg));

	/* Two byte write cycle */
	for (retries = OV511_I2C_RETRIES; ; ) {
		/* Select camera register */
		rc = ov511_reg_write(dev, OV511_REG_I2C_SUB_ADDRESS_2_BYTE, reg);
		if (rc < 0) goto error;

		/* Initiate 2-byte write cycle */
		rc = ov511_reg_write(dev, OV511_REG_I2C_CONTROL, 0x03);
		if (rc < 0) goto error;

		do rc = ov511_reg_read(dev, OV511_REG_I2C_CONTROL);
		while (rc > 0 && ((rc&1) == 0)); /* Retry until idle */
		if (rc < 0) goto error;

		if ((rc&2) == 0) /* Ack? */
			break;

		/* I2C abort */	
		ov511_reg_write(dev, OV511_REG_I2C_CONTROL, 0x10);

		if (--retries < 0) {
			err("i2c write retries exhausted");
			rc = -1;
			goto error;
		}
	}

	/* Two byte read cycle */
	for (retries = OV511_I2C_RETRIES; ; ) {
		/* Initiate 2-byte read cycle */
		rc = ov511_reg_write(dev, OV511_REG_I2C_CONTROL, 0x05);
		if (rc < 0) goto error;

		do rc = ov511_reg_read(dev, OV511_REG_I2C_CONTROL);
		while (rc > 0 && ((rc&1) == 0)); /* Retry until idle */
		if (rc < 0) goto error;

		if ((rc&2) == 0) /* Ack? */
			break;

		/* I2C abort */	
		rc = ov511_reg_write(dev, OV511_REG_I2C_CONTROL, 0x10);
		if (rc < 0) goto error;

		if (--retries < 0) {
			err("i2c read retries exhausted");
			rc = -1;
			goto error;
		}
	}

	value = ov511_reg_read(dev, OV511_REG_I2C_DATA_PORT);

	PDEBUG(5, "0x%02X:0x%02X", reg, value);
		
	/* This is needed to make ov511_i2c_write() work */
	rc = ov511_reg_write(dev, OV511_REG_I2C_CONTROL, 0x05);
	if (rc < 0)
		goto error;
	
	return value;

error:
	err("i2c read: error %d", rc);
	return rc;
}

/* Writes bits at positions specified by mask to an I2C reg. Bits that are in
 * the same position as 1's in "mask" are cleared and set to "value". Bits
 * that are in the same position as 0's in "mask" are preserved, regardless 
 * of their respective state in "value".
 */
static int ov511_i2c_write_mask(struct usb_device *dev,
				unsigned char reg,
				unsigned char value,
				unsigned char mask)
{
	int ret;
	unsigned char oldval, newval;

	ret = ov511_i2c_read(dev, reg);
	if (ret < 0)
		return ret;

	oldval = (unsigned char) ret;
	oldval &= (~mask);		/* Clear the masked bits */
	value &= mask;			/* Enforce mask on value */
	newval = oldval | value;	/* Set the desired bits */

	return (ov511_i2c_write(dev, reg, newval));
}

static int ov511_write_regvals(struct usb_device *dev,
			       struct ov511_regvals * pRegvals)
{
	int rc;

	while (pRegvals->bus != OV511_DONE_BUS) {
		if (pRegvals->bus == OV511_REG_BUS) {
			if ((rc = ov511_reg_write(dev, pRegvals->reg,
			                           pRegvals->val)) < 0)
				goto error;
		} else if (pRegvals->bus == OV511_I2C_BUS) {
			if ((rc = ov511_i2c_write(dev, pRegvals->reg, 
			                           pRegvals->val)) < 0)
				goto error;
		} else {
			err("Bad regval array");
			rc = -1;
			goto error;
		}
		pRegvals++;
	}
	return 0;

error:
	err("write regvals: error %d", rc);
	return rc;
}

#ifdef OV511_DEBUG 
static void ov511_dump_i2c_range(struct usb_device *dev, int reg1, int regn)
{
	int i;
	int rc;
	for(i = reg1; i <= regn; i++) {
		rc = ov511_i2c_read(dev, i);
		info("OV7610[0x%X] = 0x%X", i, rc);
	}
}

static void ov511_dump_i2c_regs(struct usb_device *dev)
{
	info("I2C REGS");
	ov511_dump_i2c_range(dev, 0x00, 0x7C);
}

static void ov511_dump_reg_range(struct usb_device *dev, int reg1, int regn)
{
	int i;
	int rc;
	for(i = reg1; i <= regn; i++) {
	  rc = ov511_reg_read(dev, i);
	  info("OV511[0x%X] = 0x%X", i, rc);
	}
}

static void ov511_dump_regs(struct usb_device *dev)
{
	info("CAMERA INTERFACE REGS");
	ov511_dump_reg_range(dev, 0x10, 0x1f);
	info("DRAM INTERFACE REGS");
	ov511_dump_reg_range(dev, 0x20, 0x23);
	info("ISO FIFO REGS");
	ov511_dump_reg_range(dev, 0x30, 0x31);
	info("PIO REGS");
	ov511_dump_reg_range(dev, 0x38, 0x39);
	ov511_dump_reg_range(dev, 0x3e, 0x3e);
	info("I2C REGS");
	ov511_dump_reg_range(dev, 0x40, 0x49);
	info("SYSTEM CONTROL REGS");
	ov511_dump_reg_range(dev, 0x50, 0x55);
	ov511_dump_reg_range(dev, 0x5e, 0x5f);
	info("OmniCE REGS");
	ov511_dump_reg_range(dev, 0x70, 0x79);
	/* NOTE: Quantization tables are not readable. You will get the value
	 * in reg. 0x79 for every table register */
	ov511_dump_reg_range(dev, 0x80, 0x9f);
	ov511_dump_reg_range(dev, 0xa0, 0xbf);

}
#endif

static int ov511_reset(struct usb_ov511 *ov511, unsigned char reset_type)
{
	int rc;
		
	/* Setting bit 0 not allowed on 518/518Plus */
	if (ov511->bridge == BRG_OV518PLUS)
		reset_type &= 0xfe;

	PDEBUG(4, "Reset: type=0x%X", reset_type);

	rc = ov511_reg_write(ov511->dev, OV511_REG_SYSTEM_RESET, reset_type);
	rc = ov511_reg_write(ov511->dev, OV511_REG_SYSTEM_RESET, 0);

	if (rc < 0)
		err("reset: command failed");

	return rc;
}

/* Temporarily stops OV511 from functioning. Must do this before changing
 * registers while the camera is streaming */
static inline int ov511_stop(struct usb_ov511 *ov511)
{
	PDEBUG(4, "stopping");
	if (ov511->bridge == BRG_OV518PLUS)
		return (ov511_reg_write(ov511->dev, OV511_REG_SYSTEM_RESET, 0x3c));
	else
		return (ov511_reg_write(ov511->dev, OV511_REG_SYSTEM_RESET, 0x3d));
}

/* Restarts OV511 after ov511_stop() is called */
static inline int ov511_restart(struct usb_ov511 *ov511)
{
	PDEBUG(4, "restarting");
	return (ov511_reg_write(ov511->dev, OV511_REG_SYSTEM_RESET, 0x00));
}

/* Resets the hardware snapshot button */
static void ov51x_clear_snapshot(struct usb_ov511 *ov511)
{
	if (ov511->bridge == BRG_OV511 || ov511->bridge == BRG_OV511PLUS) {
		ov511_reg_write(ov511->dev, OV511_REG_SYSTEM_SNAPSHOT, 0x01);
		ov511_reg_write(ov511->dev, OV511_REG_SYSTEM_SNAPSHOT, 0x03);
		ov511_reg_write(ov511->dev, OV511_REG_SYSTEM_SNAPSHOT, 0x01);
	} else if (ov511->bridge == BRG_OV518PLUS) {
		warn("snapshot reset not supported yet on OV518(+)");
	} else {
		err("clear snap: invalid sensor type");
	}
	
}

/* Checks the status of the snapshot button. Returns 1 if it was pressed since
 * it was last cleared, and zero in all other cases (including errors) */
static int ov51x_check_snapshot(struct usb_ov511 *ov511)
{
	int ret, status = 0;

	if (ov511->bridge == BRG_OV511 || ov511->bridge == BRG_OV511PLUS) {
		ret = ov511_reg_read(ov511->dev, OV511_REG_SYSTEM_SNAPSHOT);
		if (ret < 0) {
			err("Error checking snspshot status (%d)", ret);
		} else if (ret & 0x08) {
			status = 1;
		}
	} else if (ov511->bridge == BRG_OV518PLUS) {
		warn("snapshot check not supported yet on OV518(+)");
	} else {
		err("check snap: invalid sensor type");
	}

	return status;
}

/* Sets I2C read and write slave IDs. Returns <0 for error */
static int ov51x_set_slave_ids(struct usb_ov511 *ov511,
			       unsigned char write_id,
			       unsigned char read_id)
{
	struct usb_device *dev = ov511->dev;

	if (ov511_reg_write(dev, OV511_REG_I2C_SLAVE_ID_WRITE, write_id) < 0)
		return -EIO;

	if (ov511_reg_write(dev, OV511_REG_I2C_SLAVE_ID_READ, read_id) < 0)
		return -EIO;

	if (ov511_reset(ov511, OV511_RESET_NOREGS) < 0)
		return -EIO;

	return 0;
}

/* This does an initial reset of an OmniVision sensor and ensures that I2C
 * is synchronized. Returns <0 for failure.
 */
static int ov51x_init_ov_sensor(struct usb_ov511 *ov511)
{
	struct usb_device *dev = ov511->dev;
	int i, success;

	/* Reset the sensor */ 
	if (ov511_i2c_write(dev, 0x12, 0x80) < 0) return -EIO;

	/* Wait for it to initialize */ 
	schedule_timeout (1 + 150 * HZ / 1000);

	for (i = 0, success = 0; i < i2c_detect_tries && !success; i++) {
		if ((ov511_i2c_read(dev, OV7610_REG_ID_HIGH) == 0x7F) &&
		    (ov511_i2c_read(dev, OV7610_REG_ID_LOW) == 0xA2)) {
			success = 1;
			continue;
		}

		/* Reset the sensor */ 
		if (ov511_i2c_write(dev, 0x12, 0x80) < 0) return -EIO;
		/* Wait for it to initialize */ 
		schedule_timeout (1 + 150 * HZ / 1000);
		/* Dummy read to sync I2C */
		if (ov511_i2c_read(dev, 0x00) < 0) return -EIO;
	}

	if (!success)
		return -EIO;
	
	PDEBUG(1, "I2C synced in %d attempt(s)", i);

	return 0;
}

static int ov511_set_packet_size(struct usb_ov511 *ov511, int size)
{
	int alt, mult;

	if (ov511_stop(ov511) < 0)
		return -EIO;

	mult = size >> 5;

	if (ov511->bridge == BRG_OV511) {
		if (size == 0) alt = OV511_ALT_SIZE_0;
		else if (size == 257) alt = OV511_ALT_SIZE_257;
		else if (size == 513) alt = OV511_ALT_SIZE_513;
		else if (size == 769) alt = OV511_ALT_SIZE_769;
		else if (size == 993) alt = OV511_ALT_SIZE_993;
		else {
			err("Set packet size: invalid size (%d)", size);
			return -EINVAL;
		}
	} else if (ov511->bridge == BRG_OV511PLUS) {
		if (size == 0) alt = OV511PLUS_ALT_SIZE_0;
		else if (size == 33) alt = OV511PLUS_ALT_SIZE_33;
		else if (size == 129) alt = OV511PLUS_ALT_SIZE_129;
		else if (size == 257) alt = OV511PLUS_ALT_SIZE_257;
		else if (size == 385) alt = OV511PLUS_ALT_SIZE_385;
		else if (size == 513) alt = OV511PLUS_ALT_SIZE_513;
		else if (size == 769) alt = OV511PLUS_ALT_SIZE_769;
		else if (size == 961) alt = OV511PLUS_ALT_SIZE_961;
		else {
			err("Set packet size: invalid size (%d)", size);
			return -EINVAL;
		}
	} else if (ov511->bridge == BRG_OV518PLUS) {
		if (size == 0) alt = OV518PLUS_ALT_SIZE_0;
		else if (size == 128) alt = OV518PLUS_ALT_SIZE_128;
		else if (size == 256) alt = OV518PLUS_ALT_SIZE_256;
		else if (size == 384) alt = OV518PLUS_ALT_SIZE_384;
		else if (size == 512) alt = OV518PLUS_ALT_SIZE_512;
		else if (size == 640) alt = OV518PLUS_ALT_SIZE_640;
		else if (size == 768) alt = OV518PLUS_ALT_SIZE_768;
		else if (size == 896) alt = OV518PLUS_ALT_SIZE_896;
		else {
			err("Set packet size: invalid size (%d)", size);
			return -EINVAL;
		}
	} else {
		err("Set packet size: Invalid bridge type");
		return -EINVAL;
	}

	PDEBUG(3, "set packet size: %d, mult=%d, alt=%d", size, mult, alt);

	// FIXME: Don't know how to do this on OV518 yet
	if (ov511->bridge != BRG_OV518PLUS)
		if (ov511_reg_write(ov511->dev, OV511_REG_FIFO_PACKET_SIZE, mult) < 0)
			return -EIO;
	
	if (usb_set_interface(ov511->dev, ov511->iface, alt) < 0) {
		err("Set packet size: set interface error");
		return -EBUSY;
	}

	/* I don't know what this does */
	if (ov511->bridge == BRG_OV518PLUS)
		if (ov511_reg_write(ov511->dev, 0x2f, 0x80) < 0)
			return -EIO;

	// FIXME - Should we only reset the FIFO?
	if (ov511_reset(ov511, OV511_RESET_NOREGS) < 0)
		return -EIO;

	ov511->packet_size = size;

	if (ov511_restart(ov511) < 0)
		return -EIO;

	return 0;
}


static inline int
ov7610_set_picture(struct usb_ov511 *ov511, struct video_picture *p)
{
	int ret;
	struct usb_device *dev = ov511->dev;

	PDEBUG(4, "ov511_set_picture");

	if (ov511_stop(ov511) < 0)
		return -EIO;

	ov511->contrast = p->contrast;
	ov511->brightness = p->brightness;
	ov511->colour = p->colour;
	ov511->hue = p->hue;
	ov511->whiteness = p->whiteness;

	if ((ret = ov511_i2c_read(dev, OV7610_REG_COM_B)) < 0)
		return -EIO;
#if 0
	/* disable auto adjust mode */
	if (ov511_i2c_write(dev, OV7610_REG_COM_B, ret & 0xfe) < 0)
		return -EIO;
#endif
	if (ov511->sensor == SEN_OV7610 || ov511->sensor == SEN_OV7620AE
		|| ov511->sensor == SEN_OV6620)
		if (ov511_i2c_write(dev, OV7610_REG_SAT, p->colour >> 8) < 0)
			return -EIO;

	if (ov511->sensor == SEN_OV7610 || ov511->sensor == SEN_OV6620) {
		if (ov511_i2c_write(dev, OV7610_REG_CNT, p->contrast >> 8) < 0)
			return -EIO;

		if (ov511_i2c_write(dev, OV7610_REG_RED, 0xFF - (p->hue >> 8)) < 0)
			return -EIO;

		if (ov511_i2c_write(dev, OV7610_REG_BLUE, p->hue >> 8) < 0)
			return -EIO;

	} else if (ov511->sensor == SEN_OV7620) {
		/* Use gamma control instead. Bit 0 enables it. */
		if (ov511_i2c_write(dev, 0x64, (p->contrast >> 8) | 1) < 0)
			return -EIO;

		/* These regs are undocumented. They may be backwards */
		if (ov511_i2c_write(dev, 0x7a, 0xFF - (p->hue >> 8)) < 0)
			return -EIO;

		if (ov511_i2c_write(dev, 0x79, p->hue >> 8) < 0)
			return -EIO;

#if 0
		int cur_sat, new_sat, tmp;

		cur_sat = ov511_i2c_read(dev, OV7610_REG_BLUE);

		tmp = (p->hue >> 8) - cur_sat;
		new_sat = (tmp < 0) ? (-tmp) | 0x80 : tmp;

	        PDEBUG(1, "cur=%d target=%d diff=%d", cur_sat, p->hue >> 8, tmp); 

		if (ov511_i2c_write(dev, OV7610_REG_BLUE, new_sat) < 0)
			return -EIO;

	        // DEBUG_CODE
	        PDEBUG(1, "hue=%d", ov511_i2c_read(dev, OV7610_REG_BLUE)); 

#endif
	} else if (ov511->sensor == SEN_OV7620AE) {
	}

	if (ov511_i2c_write(dev, OV7610_REG_BRT, p->brightness >> 8) < 0)
		return -EIO;

	if (ov511_restart(ov511) < 0)
		return -EIO;

	return 0;
}

static inline int
ov7610_get_picture(struct usb_ov511 *ov511, struct video_picture *p)
{
	int ret;
	struct usb_device *dev = ov511->dev;

	PDEBUG(4, "ov511_get_picture");

	if (ov511_stop(ov511) < 0)
		return -EIO;

	if ((ret = ov511_i2c_read(dev, OV7610_REG_SAT)) < 0) return -EIO;
	p->colour = ret << 8;

	if ((ret = ov511_i2c_read(dev, OV7610_REG_CNT)) < 0) return -EIO;
	p->contrast = ret << 8;

	if ((ret = ov511_i2c_read(dev, OV7610_REG_BRT)) < 0) return -EIO;
	p->brightness = ret << 8;

	/* This may not be the best way to do it */
	if ((ret = ov511_i2c_read(dev, OV7610_REG_BLUE)) < 0) return -EIO;
	p->hue = ret << 8;

	p->whiteness = 105 << 8;

	/* Can we get these from frame[0]? -claudio? */
	p->depth = ov511->frame[0].depth;
	p->palette = ov511->frame[0].format;

	if (ov511_restart(ov511) < 0)
		return -EIO;

	return 0;
}

/* Turns on or off the LED. Only has an effect with OV511+/OV518+ */
static inline void 
ov511_led_control(struct usb_ov511 *ov511, int enable) {
	PDEBUG(4, " (%s)", enable ? "turn on" : "turn off");

	if (ov511->bridge == BRG_OV511PLUS)
		ov511_reg_write(ov511->dev, OV511_REG_SYSTEM_LED_CTL, 
			        enable ? 1 : 0);
	else if (ov511->bridge == BRG_OV518PLUS)
		ov511_reg_write_mask(ov511->dev, OV518_REG_GPIO_OUT, 
				     enable ? 0x02 : 0x00, 0x02);
	return;
}

/* Matches the sensor's internal frame rate to the lighting frequency.
 * Valid frequencies are:
 *	50 - 50Hz, for European and Asian lighting
 *	60 - 60Hz, for American lighting
 *
 * Tested with: OV7610, OV7620, OV7620AE, OV6620
 * Returns: 0 for success
 */
static inline int
ov51x_set_light_freq(struct usb_ov511 *ov511, int freq) {
	int sixty;

	PDEBUG(4, "%d Hz", freq);

	if (freq == 60)
		sixty = 1;
	else if (freq == 50)
		sixty = 0;
	else {
		err("Invalid light freq (%d Hz)", freq);
		return -EINVAL;
	}
	
	switch (ov511->sensor) {
	case SEN_OV7610:
		ov511_i2c_write_mask(ov511->dev, 0x2a, sixty?0x00:0x80, 0x80);
		ov511_i2c_write(ov511->dev, 0x2b, sixty?0x00:0xac);
		ov511_i2c_write_mask(ov511->dev, 0x13, 0x10, 0x10);
		ov511_i2c_write_mask(ov511->dev, 0x13, 0x00, 0x10);
		break;
	case SEN_OV7620:
	case SEN_OV7620AE:
	case SEN_OV8600:
		ov511_i2c_write_mask(ov511->dev, 0x2a, sixty?0x00:0x80, 0x80);
		ov511_i2c_write(ov511->dev, 0x2b, sixty?0x00:0xac);
		ov511_i2c_write_mask(ov511->dev, 0x76, 0x01, 0x01);
		break;		
	case SEN_OV6620:
	case SEN_OV6630:
		ov511_i2c_write(ov511->dev, 0x2b, sixty?0xa8:0x28);
		ov511_i2c_write(ov511->dev, 0x2a, sixty?0x84:0xa4);
		break;
	default:
		err("Sensor not supported for set_light_freq");
		return -EINVAL;
	}

	return 0;
}

/* If enable is true, turn on the sensor's banding filter, otherwise turn it off. This
 * filter tries to reduce the pattern of horizontal light/dark bands caused by some
 * (usually fluorescent) lighting. The light frequency must be set either before or
 * after enabling it with ov51x_set_light_freq().
 *
 * Tested with: OV7610, OV7620AE, OV6620. Ruins exposure with OV7620.
 * Returns: 0 for success
 */
static inline int
ov51x_set_banding_filter(struct usb_ov511 *ov511, int enable)
{	
	PDEBUG(4, " (%s)", enable ? "turn on" : "turn off");

	return (ov511_i2c_write_mask(ov511->dev, 0x2d, enable?0x04:0x00, 0x04));
}

/* Returns number of bits per pixel (regardless of where they are located; planar or
 * not), or zero for unsupported format.
 */
static int ov511_get_depth(int palette)
{
	switch (palette) {
	case VIDEO_PALETTE_GREY:    return 8;
	case VIDEO_PALETTE_RGB565:  return 16;
	case VIDEO_PALETTE_RGB24:   return 24;  
	case VIDEO_PALETTE_YUV422:  return 16;
	case VIDEO_PALETTE_YUYV:    return 16;
	case VIDEO_PALETTE_YUV420:  return 24;
	case VIDEO_PALETTE_YUV422P: return 24; /* Planar */
	case VIDEO_PALETTE_YUV420P: return 24; /* Planar */
	default:		    return 0;  /* Invalid format */
	}
}

/* LNCNT values fixed by Lawrence Glaister <lg@jfm.bc.ca> */
static struct mode_list_511 mlist511[] = {
	/* W    H   C  PXCNT LNCNT PXDIV LNDIV M420  COMA  COML */
	{ 640, 480, 0, 0x4f, 0x3b, 0x00, 0x00, 0x03, 0x24, 0x9e },
	{ 640, 480, 1, 0x4f, 0x3b, 0x00, 0x00, 0x03, 0x24, 0x9e },
	{ 320, 240, 0, 0x27, 0x1d, 0x00, 0x00, 0x03, 0x04, 0x1e },
	{ 320, 240, 1, 0x27, 0x1d, 0x00, 0x00, 0x03, 0x04, 0x1e },
	{ 352, 288, 0, 0x2b, 0x25, 0x00, 0x00, 0x03, 0x04, 0x1e },
	{ 352, 288, 1, 0x2b, 0x25, 0x00, 0x00, 0x03, 0x04, 0x1e },
	{ 384, 288, 0, 0x2f, 0x25, 0x00, 0x00, 0x03, 0x04, 0x1e },
	{ 384, 288, 1, 0x2f, 0x25, 0x00, 0x00, 0x03, 0x04, 0x1e },
	{ 448, 336, 0, 0x37, 0x29, 0x00, 0x00, 0x03, 0x04, 0x1e },
	{ 448, 336, 1, 0x37, 0x29, 0x00, 0x00, 0x03, 0x04, 0x1e },
	{ 176, 144, 0, 0x13, 0x13, 0x00, 0x00, 0x03, 0x04, 0x1e },
	{ 176, 144, 1, 0x13, 0x13, 0x00, 0x00, 0x03, 0x04, 0x1e },
	{ 160, 120, 0, 0x13, 0x0e, 0x00, 0x00, 0x03, 0x04, 0x1e },
	{ 160, 120, 1, 0x13, 0x0e, 0x00, 0x00, 0x03, 0x04, 0x1e },
	{ 0, 0 }
};

static int
ov511_mode_init_regs(struct usb_ov511 *ov511,
		     int width, int height, int mode, int sub_flag)
{
	int i;
	struct usb_device *dev = ov511->dev;
	int hwsbase, hwebase, vwsbase, vwebase, hwsize, vwsize; 
	int hoffset, voffset, hwscale = 0, vwscale = 0;

	PDEBUG(3, "width:%d, height:%d, mode:%d, sub:%d",
	       width, height, mode, sub_flag);

	if (ov511_stop(ov511) < 0)
		return -EIO;

	/* Dumppix only works with RGB24 */
	if (dumppix && (mode != VIDEO_PALETTE_RGB24)) {
		err("dumppix only supported with RGB 24");
		return -EINVAL;
	}

	if (mode == VIDEO_PALETTE_GREY) {
		ov511_reg_write(dev, 0x16, 0x00);
		if (ov511->sensor == SEN_OV7610
		    || ov511->sensor == SEN_OV7620AE) {
			/* these aren't valid on the OV6620/OV7620/6630? */
			ov511_i2c_write(dev, 0x0e, 0x44);
		}
		ov511_i2c_write_mask(dev, 0x13, autoadjust ? 0x21 : 0x20, 0x21);

		/* For snapshot */
		ov511_reg_write(dev, 0x1e, 0x00);
		ov511_reg_write(dev, 0x1f, 0x01);
	} else {
		ov511_reg_write(dev, 0x16, 0x01);
		if (ov511->sensor == SEN_OV7610
		    || ov511->sensor == SEN_OV7620AE) {
			/* not valid on the OV6620/OV7620/6630? */
			ov511_i2c_write(dev, 0x0e, 0x04);
		}
		ov511_i2c_write_mask(dev, 0x13, autoadjust ? 0x01 : 0x00, 0x21);

		/* For snapshot */
		ov511_reg_write(dev, 0x1e, 0x01);
		ov511_reg_write(dev, 0x1f, 0x03);
	}

	/* The different sensor ICs handle setting up of window differently.
	 * IF YOU SET IT WRONG, YOU WILL GET ALL ZERO ISOC DATA FROM OV511!!! */
	switch (ov511->sensor) {
	case SEN_OV7610:
	case SEN_OV7620AE:
		hwsbase = 0x38;
		hwebase = 0x3a;
		vwsbase = vwebase = 0x05;
		break;
	case SEN_OV6620:
	case SEN_OV6630:	// FIXME: Is this right?
		hwsbase = 0x38;
		hwebase = 0x3a;
		vwsbase = 0x05;
		vwebase = 0x06;
		break;
	case SEN_OV7620:
		hwsbase = 0x2c;
		hwebase = 0x2d;
		vwsbase = vwebase = 0x05;
		break;
	default:
		err("Invalid sensor");
		return -EINVAL;
	}

	/* Bit 5 of COM C register varies with sensor */ 
	if (ov511->sensor == SEN_OV6620) {
		if (width > 176 && height > 144) {  /* CIF */
			ov511_i2c_write(dev, 0x14, 0x04);
			hwscale = 1;
			vwscale = 1;  /* The datasheet says 0; it's wrong */
			hwsize = 352;
			vwsize = 288;
		} else {			    /* QCIF */
			ov511_i2c_write(dev, 0x14, 0x24);
			hwsize = 176;
			vwsize = 144;
		}
	}
	else {
		if (width > 320 && height > 240) {  /* VGA */
			ov511_i2c_write(dev, 0x14, 0x04);
			hwscale = 2;
			vwscale = 1;
			hwsize = 640;
			vwsize = 480;
		} else {			    /* QVGA */
			ov511_i2c_write(dev, 0x14, 0x24);
			hwscale = 1;
			hwsize = 320;
			vwsize = 240;
		}	
	}

	/* Center the window */
	hoffset = ((hwsize - width) / 2) >> hwscale;
	voffset = ((vwsize - height) / 2) >> vwscale;

	/* FIXME! - This needs to be changed to support 160x120 and 6620!!! */
	if (sub_flag) {
		ov511_i2c_write(dev, 0x17, hwsbase+(ov511->subx>>hwscale));
		ov511_i2c_write(dev, 0x18, hwebase+((ov511->subx+ov511->subw)>>hwscale));
		ov511_i2c_write(dev, 0x19, vwsbase+(ov511->suby>>vwscale));
		ov511_i2c_write(dev, 0x1a, vwebase+((ov511->suby+ov511->subh)>>vwscale));
	} else {
		ov511_i2c_write(dev, 0x17, hwsbase + hoffset);
		ov511_i2c_write(dev, 0x18, hwebase + hoffset + (hwsize>>hwscale));
		ov511_i2c_write(dev, 0x19, vwsbase + voffset);
		ov511_i2c_write(dev, 0x1a, vwebase + voffset + (vwsize>>vwscale));
	}

	for (i = 0; mlist511[i].width; i++) {
		int lncnt, pxcnt, clock;

		if (width != mlist511[i].width || height != mlist511[i].height)
			continue;

		if (!mlist511[i].color && mode != VIDEO_PALETTE_GREY)
			continue;

		/* Here I'm assuming that snapshot size == image size.
		 * I hope that's always true. --claudio
		 */
		pxcnt = sub_flag ? (ov511->subw >> 3) - 1 : mlist511[i].pxcnt;
		lncnt = sub_flag ? (ov511->subh >> 3) - 1 : mlist511[i].lncnt;

		ov511_reg_write(dev, 0x12, pxcnt);
		ov511_reg_write(dev, 0x13, lncnt);
		ov511_reg_write(dev, 0x14, mlist511[i].pxdv);
		ov511_reg_write(dev, 0x15, mlist511[i].lndv);
		ov511_reg_write(dev, 0x18, mlist511[i].m420);

		/* Snapshot additions */
		ov511_reg_write(dev, 0x1a, pxcnt);
		ov511_reg_write(dev, 0x1b, lncnt);
                ov511_reg_write(dev, 0x1c, mlist511[i].pxdv);
                ov511_reg_write(dev, 0x1d, mlist511[i].lndv);

		/* The OV6620 needs special handling. This prevents the 
		 * severe banding that normally occurs */
		if (ov511->sensor == SEN_OV6620)
		{
			/* Clock down */
			ov511_i2c_write(dev, 0x2a, 0x04);
			ov511_i2c_write(dev, 0x11, 0x03);
			ov511_i2c_write(dev, 0x2a, 0x84);
			/* This next setting is critical. It seems to improve
			 * the gain or the contrast. The "reserved" bits seem
			 * to have some effect in this case. */
			ov511_i2c_write(dev, 0x2d, 0x85);

		}
		else
		{
			/* Calculate and set the clock divisor */
			clock = ((sub_flag ? ov511->subw * ov511->subh : width * height)
			* (mlist511[i].color ? 3 : 2) / 2) / 66000;
#if 0
			clock *= cams;
#endif
			ov511_i2c_write(dev, 0x11, clock);
		}

		/* We only have code to convert GBR -> RGB24 */
		if ((mode == VIDEO_PALETTE_RGB24) && sensor_gbr)
			ov511_i2c_write(dev, 0x12, mlist511[i].common_A | (testpat?0x0a:0x08));
		else
			ov511_i2c_write(dev, 0x12, mlist511[i].common_A | (testpat?0x02:0x00));

		/* Enable/Disable AGC */
		if (ov511_i2c_write_mask(dev, 0x12, autoadjust ? 0x20 : 0x00, 0x20) < 0)
			return -1;

		/* 7620/6620/6630? don't have register 0x35, so play it safe */
		if (ov511->sensor == SEN_OV7610 ||
		    ov511->sensor == SEN_OV7620AE)
			ov511_i2c_write(dev, 0x35, mlist511[i].common_L);

		break;
	}

	if (lightfreq)
		if (ov51x_set_light_freq(ov511, lightfreq))
			return -1;

	if (ov51x_set_banding_filter(ov511, bandingfilter))
			return -1;

	if (compress) {
		ov511_reg_write(dev, 0x78, 0x07); // Turn on Y and UV compression
		ov511_reg_write(dev, 0x79, 0x03); // Enable LUTs
		ov511_reset(ov511, OV511_RESET_OMNICE);
	}

	if (ov511_restart(ov511) < 0)
		return -EIO;

	if (mlist511[i].width == 0) {
		err("Unknown mode (%d, %d): %d", width, height, mode);
		return -EINVAL;
	}

#ifdef OV511_DEBUG
	if (dump_sensor)
		ov511_dump_i2c_regs(dev);
#endif

	return 0;
}

static struct mode_list_518 mlist518[] = {
	/* W    H   reg28 reg29 reg2a reg2c reg2e reg24 reg25 */
	{ 352, 288, 0x80, 0x16, 0x48, 0x00, 0x00, 0x9f, 0x90 },
	{ 320, 240, 0x80, 0x14, 0x3c, 0x10, 0x18, 0x9f, 0x90 },
	{ 176, 144, 0x85, 0x0b, 0x24, 0x00, 0x00, 0xff, 0xf0 },
	{ 160, 120, 0x85, 0x0a, 0x1e, 0x08, 0x0c, 0xff, 0xf0 },
	{ 0, 0 }
};

/* OV518 needs a completely different approach, until we can figure out what
 * the individual registers do. Many register ops are commented out until we
 * can find out if they are still valid. Also, only 15 FPS is supported now.
 */
static int
ov518_mode_init_regs(struct usb_ov511 *ov511,
		     int width, int height, int mode, int sub_flag)
{
	int i;
	struct usb_device *dev = ov511->dev;
	int hwsbase, hwebase, vwsbase, vwebase, hwsize, vwsize; 
	int hoffset, voffset, hwscale = 0, vwscale = 0;
	unsigned char b[3]; /* Multiple-value reg buffer */

	PDEBUG(3, "width:%d, height:%d, mode:%d, sub:%d",
	       width, height, mode, sub_flag);

	if (ov511_stop(ov511) < 0)
		return -EIO;

	/* Dumppix only works with RGB24 */
	if (dumppix && (mode != VIDEO_PALETTE_RGB24)) {
		err("dumppix only supported with RGB 24");
		return -EINVAL;
	}

	/* I don't know how to set YUV400 mode on OV518 yet */
	if (mode == VIDEO_PALETTE_GREY) {
//		ov511_reg_write(dev, 0x16, 0x00);
		if (ov511->sensor == SEN_OV7610
		    || ov511->sensor == SEN_OV7620AE) {
			/* these aren't valid on the OV6620/OV7620/6630? */
			ov511_i2c_write(dev, 0x0e, 0x44);
		}
		ov511_i2c_write_mask(dev, 0x13, autoadjust ? 0x21 : 0x20, 0x21);
	} else {
//		ov511_reg_write(dev, 0x16, 0x01);
		if (ov511->sensor == SEN_OV7610
		    || ov511->sensor == SEN_OV7620AE) {
			/* not valid on the OV6620/OV7620/6630? */
			ov511_i2c_write(dev, 0x0e, 0x04);
		}
		ov511_i2c_write_mask(dev, 0x13, autoadjust ? 0x01 : 0x00, 0x21);
	}

	/* The different sensor ICs handle setting up of window differently */
	switch (ov511->sensor) {
	case SEN_OV7610:
	case SEN_OV7620AE:
		hwsbase = 0x38;
		hwebase = 0x3a;
		vwsbase = vwebase = 0x05;
		break;
	case SEN_OV6620:
	case SEN_OV6630:	// FIXME: Is this right?
		hwsbase = 0x38;
		hwebase = 0x3a;
		vwsbase = 0x05;
		vwebase = 0x06;
		break;
	case SEN_OV7620:
		hwsbase = 0x2c;
		hwebase = 0x2d;
		vwsbase = vwebase = 0x05;
		break;
	default:
		err("Invalid sensor");
		return -EINVAL;
	}

	/* Bit 5 of COM C register varies with sensor */ 
	if (ov511->sensor == SEN_OV6630) {
		if (width > 176 && height > 144) {  /* CIF */
			ov511_i2c_write(dev, 0x14, 0x04);
			hwscale = 1;
			vwscale = 1;  /* The datasheet says 0; it's wrong */
			hwsize = 352;
			vwsize = 288;
		} else {			    /* QCIF */
			ov511_i2c_write(dev, 0x14, 0x24);
			hwsize = 176;
			vwsize = 144;
		}
	}
	else {
		if (width > 320 && height > 240) {  /* VGA */
			ov511_i2c_write(dev, 0x14, 0x04);
			hwscale = 2;
			vwscale = 1;
			hwsize = 640;
			vwsize = 480;
		} else {			    /* QVGA */
			ov511_i2c_write(dev, 0x14, 0x24);
			hwscale = 1;
			hwsize = 320;
			vwsize = 240;
		}	
	}

	/* Center the window */
	hoffset = ((hwsize - width) / 2) >> hwscale;
	voffset = ((vwsize - height) / 2) >> vwscale;

	/* FIXME! - This needs to be changed to support 160x120 and 6620!!! */
	if (sub_flag) {
		ov511_i2c_write(dev, 0x17, hwsbase+(ov511->subx>>hwscale));
		ov511_i2c_write(dev, 0x18, hwebase+((ov511->subx+ov511->subw)>>hwscale));
		ov511_i2c_write(dev, 0x19, vwsbase+(ov511->suby>>vwscale));
		ov511_i2c_write(dev, 0x1a, vwebase+((ov511->suby+ov511->subh)>>vwscale));
	} else {
		ov511_i2c_write(dev, 0x17, hwsbase + hoffset);
		ov511_i2c_write(dev, 0x18, hwebase + hoffset + (hwsize>>hwscale));
		ov511_i2c_write(dev, 0x19, vwsbase + voffset);
		ov511_i2c_write(dev, 0x1a, vwebase + voffset + (vwsize>>vwscale));
	}

	for (i = 0; mlist518[i].width; i++) {
//		int lncnt, pxcnt;
		int clock;

		if (width != mlist518[i].width || height != mlist518[i].height)
			continue;

// FIXME: Subcapture won't be possible until we know what the registers do
// FIXME: We can't handle anything but YUV420 so far

//		/* Here I'm assuming that snapshot size == image size.
//		 * I hope that's always true. --claudio
//		 */
//		pxcnt = sub_flag ? (ov511->subw >> 3) - 1 : mlist[i].pxcnt;
//		lncnt = sub_flag ? (ov511->subh >> 3) - 1 : mlist[i].lncnt;
//
//		ov511_reg_write(dev, 0x12, pxcnt);
//		ov511_reg_write(dev, 0x13, lncnt);

		/******** Set the mode ********/		

		/* Mode independent regs */
		ov511_reg_write(dev, 0x2b, 0x00);
		ov511_reg_write(dev, 0x2d, 0x00);
		ov511_reg_write(dev, 0x3b, 0x00);
		ov511_reg_write(dev, 0x3d, 0x00);

		/* Mode dependent regs. Regs 38 - 3e are always the same as
		 * regs 28 - 2e */
		ov511_reg_write_mask(dev, 0x28, mlist518[i].reg28, 0x0f);
		ov511_reg_write(dev, 0x29, mlist518[i].reg29);
		ov511_reg_write(dev, 0x2a, mlist518[i].reg2a);
		ov511_reg_write(dev, 0x2c, mlist518[i].reg2c);
		ov511_reg_write(dev, 0x2e, mlist518[i].reg2e);
		ov511_reg_write_mask(dev, 0x38, mlist518[i].reg28, 0x0f);
		ov511_reg_write(dev, 0x39, mlist518[i].reg29);
		ov511_reg_write(dev, 0x3a, mlist518[i].reg2a);
		ov511_reg_write(dev, 0x3c, mlist518[i].reg2c);
		ov511_reg_write(dev, 0x3e, mlist518[i].reg2e);
		ov511_reg_write(dev, 0x24, mlist518[i].reg24);
		ov511_reg_write(dev, 0x25, mlist518[i].reg25);

		/* Windows driver does this here; who knows why */
		ov511_reg_write(dev, 0x2f, 0x80);

		/******** Set the framerate (to 15 FPS) ********/		

		/* Mode independent, but framerate dependent, regs */
		/* These are for 15 FPS only */
		ov511_reg_write(dev, 0x51, 0x08);
		ov511_reg_write(dev, 0x22, 0x18);
		ov511_reg_write(dev, 0x23, 0xff);
		ov511_reg_write(dev, 0x71, 0x19);  /* Compression-related? */
		/* Bit 5 is what matters here. Of course, it is "reserved" */
		ov511_i2c_write(dev, 0x54, 0x23);

		ov511_reg_write(dev, 0x2f, 0x80);

		/* Mode dependent regs */
		if ((width == 352 && height == 288) ||
		    (width == 320 && height == 240)) {
			b[0]=0x80; b[1]=0x02; ov518_reg_write_multi(dev, 0x30, b, 2);
			b[0]=0x90; b[1]=0x01; ov518_reg_write_multi(dev, 0xc4, b, 2);
			b[0]=0xf4; b[1]=0x01; ov518_reg_write_multi(dev, 0xc6, b, 2);
			b[0]=0xf4; b[1]=0x01; ov518_reg_write_multi(dev, 0xc7, b, 2);
			b[0]=0x8e; b[1]=0x00; ov518_reg_write_multi(dev, 0xc8, b, 2);
			b[0]=0x1a; b[1]=0x00; b[2]=0x02; ov518_reg_write_multi(dev, 0xca, b, 3);
			b[0]=0x14; b[1]=0x02; ov518_reg_write_multi(dev, 0xcb, b, 2);
			b[0]=0xd0; b[1]=0x07; ov518_reg_write_multi(dev, 0xcc, b, 2);
			b[0]=0x20; b[1]=0x00; ov518_reg_write_multi(dev, 0xcd, b, 2);
			b[0]=0x60; b[1]=0x02; ov518_reg_write_multi(dev, 0xce, b, 2);

		} else if ((width == 176 && height == 144) ||
			   (width == 160 && height == 120)) {
			b[0]=0x80; b[1]=0x01; ov518_reg_write_multi(dev, 0x30, b, 2);
			b[0]=0xc8; b[1]=0x00; ov518_reg_write_multi(dev, 0xc4, b, 2);
			b[0]=0x40; b[1]=0x01; ov518_reg_write_multi(dev, 0xc6, b, 2);
			b[0]=0x40; b[1]=0x01; ov518_reg_write_multi(dev, 0xc7, b, 2);
			b[0]=0x60; b[1]=0x00; ov518_reg_write_multi(dev, 0xc8, b, 2);
			b[0]=0x0f; b[1]=0x33; b[2]=0x01; ov518_reg_write_multi(dev, 0xca, b, 3);
			b[0]=0x40; b[1]=0x01; ov518_reg_write_multi(dev, 0xcb, b, 2);
			b[0]=0xec; b[1]=0x04; ov518_reg_write_multi(dev, 0xcc, b, 2);
			b[0]=0x13; b[1]=0x00; ov518_reg_write_multi(dev, 0xcd, b, 2);
			b[0]=0x6d; b[1]=0x01; ov518_reg_write_multi(dev, 0xce, b, 2);
		} else {
			/* Can't happen, since we already handled this case */
			err("ov518_mode_init_regs(): **** logic error ****");
		}

		ov511_reg_write(dev, 0x2f, 0x80);

		/* The OV6620 needs special handling. This prevents the 
		 * severe banding that normally occurs */
		if (ov511->sensor == SEN_OV6620)
		{
			/* Clock down */
			ov511_i2c_write(dev, 0x2a, 0x04);
			ov511_i2c_write(dev, 0x11, 0x03);
			ov511_i2c_write(dev, 0x2a, 0x84);
			/* This next setting is critical. It seems to improve
			 * the gain or the contrast. The "reserved" bits seem
			 * to have some effect in this case. */
			ov511_i2c_write(dev, 0x2d, 0x85);

		}
		else
		{
			/* Calculate and set the clock divisor */
// Color is always on
//			clock = ((sub_flag ? ov511->subw * ov511->subh : width * height)
//			* (mlist518[i].color ? 3 : 2) / 2) / 66000;
			clock = ((sub_flag ? ov511->subw * ov511->subh : width * height)
			* (3 / 2)) / 66000;
#if 0
			clock *= cams;
#endif
			ov511_i2c_write(dev, 0x11, clock);
		}

		/* We only have code to convert GBR -> RGB24 */
		if ((mode == VIDEO_PALETTE_RGB24) && sensor_gbr)
			ov511_i2c_write(dev, 0x12, 0x04 | (testpat?0x0a:0x08));
		else
			ov511_i2c_write(dev, 0x12, 0x04 | (testpat?0x02:0x00));

		/* Enable/Disable AGC */
		if (ov511_i2c_write_mask(dev, 0x12, autoadjust ? 0x20 : 0x00, 0x20) < 0)
			return -1;

		/* 7620/6620/6630? don't have register 0x35, so play it safe */
		if (ov511->sensor == SEN_OV7610 ||
		    ov511->sensor == SEN_OV7620AE)
			ov511_i2c_write(dev, 0x35, 0x1e);

		break;
	}

	if (lightfreq)
		if (ov51x_set_light_freq(ov511, lightfreq))
			return -1;

	if (ov51x_set_banding_filter(ov511, bandingfilter))
			return -1;

	if (compress) {
//		ov511_reg_write(dev, 0x78, 0x03); // Turn on Y compression
//		ov511_reg_write(dev, 0x79, 0x00); // Disable LUTs
	}

	if (ov511_restart(ov511) < 0)
		return -EIO;

	if (mlist518[i].width == 0) {
		err("Unknown mode (%d, %d): %d", width, height, mode);
		return -EINVAL;
	}

#ifdef OV511_DEBUG
	if (dump_sensor)
		ov511_dump_i2c_regs(dev);
#endif

	return 0;
}

/* This is a wrapper around the OV511, OV518, and sensor specific functions */
static int
mode_init_regs(struct usb_ov511 *ov511,
	       int width, int height, int mode, int sub_flag)
{
	if (ov511->bridge == BRG_OV518PLUS)
		return ov518_mode_init_regs(ov511, width, height, mode, sub_flag);
	else
		return ov511_mode_init_regs(ov511, width, height, mode, sub_flag);
}

/* This sets the default image parameters (Size = max, RGB24). This is
 * useful for apps that use read() and do not set these.
 */
static int ov51x_set_default_params(struct usb_ov511 *ov511)
{
	int i;

	/* Set default sizes in case IOCTL (VIDIOCMCAPTURE) is not used
	 * (using read() instead). */
	for (i = 0; i < OV511_NUMFRAMES; i++) {
		ov511->frame[i].width = ov511->maxwidth;
		ov511->frame[i].height = ov511->maxheight;
		ov511->frame[i].depth = 24;
		ov511->frame[i].bytes_read = 0;
		ov511->frame[i].segment = 0;
		ov511->frame[i].format = VIDEO_PALETTE_RGB24;
		ov511->frame[i].segsize = GET_SEGSIZE(ov511->frame[i].format);
	}

	/* Initialize to max width/height, RGB24 */
	if (mode_init_regs(ov511, ov511->maxwidth, ov511->maxheight,
			   VIDEO_PALETTE_RGB24, 0) < 0)
		return -EINVAL;

	return 0;
}

/**********************************************************************
 *
 * Color correction functions
 *
 **********************************************************************/

/*
 * Turn a YUV4:2:0 block into an RGB block
 *
 * Video4Linux seems to use the blue, green, red channel
 * order convention-- rgb[0] is blue, rgb[1] is green, rgb[2] is red.
 *
 * Color space conversion coefficients taken from the excellent
 * http://www.inforamp.net/~poynton/ColorFAQ.html
 * In his terminology, this is a CCIR 601.1 YCbCr -> RGB.
 * Y values are given for all 4 pixels, but the U (Pb)
 * and V (Pr) are assumed constant over the 2x2 block.
 *
 * To avoid floating point arithmetic, the color conversion
 * coefficients are scaled into 16.16 fixed-point integers.
 * They were determined as follows:
 *
 *	double brightness = 1.0;  (0->black; 1->full scale) 
 *	double saturation = 1.0;  (0->greyscale; 1->full color)
 *	double fixScale = brightness * 256 * 256;
 *	int rvScale = (int)(1.402 * saturation * fixScale);
 *	int guScale = (int)(-0.344136 * saturation * fixScale);
 *	int gvScale = (int)(-0.714136 * saturation * fixScale);
 *	int buScale = (int)(1.772 * saturation * fixScale);
 *	int yScale = (int)(fixScale);	
 */

/* LIMIT: convert a 16.16 fixed-point value to a byte, with clipping. */
#define LIMIT(x) ((x)>0xffffff?0xff: ((x)<=0xffff?0:((x)>>16)))

static inline void
ov511_move_420_block(int yTL, int yTR, int yBL, int yBR, int u, int v, 
	int rowPixels, unsigned char * rgb, int bits)
{
	const int rvScale = 91881;
	const int guScale = -22553;
	const int gvScale = -46801;
	const int buScale = 116129;
	const int yScale  = 65536;
	int r, g, b;

	g = guScale * u + gvScale * v;
	if (force_rgb) {
		r = buScale * u;
		b = rvScale * v;
	} else {
		r = rvScale * v;
		b = buScale * u;
	}

	yTL *= yScale; yTR *= yScale;
	yBL *= yScale; yBR *= yScale;

	if (bits == 24) {
		/* Write out top two pixels */
		rgb[0] = LIMIT(b+yTL); rgb[1] = LIMIT(g+yTL); rgb[2] = LIMIT(r+yTL);
		rgb[3] = LIMIT(b+yTR); rgb[4] = LIMIT(g+yTR); rgb[5] = LIMIT(r+yTR);

		/* Skip down to next line to write out bottom two pixels */
		rgb += 3 * rowPixels;
		rgb[0] = LIMIT(b+yBL); rgb[1] = LIMIT(g+yBL); rgb[2] = LIMIT(r+yBL);
		rgb[3] = LIMIT(b+yBR); rgb[4] = LIMIT(g+yBR); rgb[5] = LIMIT(r+yBR);
	} else if (bits == 16) {
		/* Write out top two pixels */
		rgb[0] = ((LIMIT(b+yTL) >> 3) & 0x1F) | ((LIMIT(g+yTL) << 3) & 0xE0);
		rgb[1] = ((LIMIT(g+yTL) >> 5) & 0x07) | (LIMIT(r+yTL) & 0xF8);

		rgb[2] = ((LIMIT(b+yTR) >> 3) & 0x1F) | ((LIMIT(g+yTR) << 3) & 0xE0);
		rgb[3] = ((LIMIT(g+yTR) >> 5) & 0x07) | (LIMIT(r+yTR) & 0xF8);

		/* Skip down to next line to write out bottom two pixels */
		rgb += 2 * rowPixels;

		rgb[0] = ((LIMIT(b+yBL) >> 3) & 0x1F) | ((LIMIT(g+yBL) << 3) & 0xE0);
		rgb[1] = ((LIMIT(g+yBL) >> 5) & 0x07) | (LIMIT(r+yBL) & 0xF8);

		rgb[2] = ((LIMIT(b+yBR) >> 3) & 0x1F) | ((LIMIT(g+yBR) << 3) & 0xE0);
		rgb[3] = ((LIMIT(g+yBR) >> 5) & 0x07) | (LIMIT(r+yBR) & 0xF8);
	}
}

/*
 * For a 640x480 YUV4:2:0 images, data shows up in 1200 384 byte segments.
 * The first 64 bytes of each segment are U, the next 64 are V.  The U and
 * V are arranged as follows:
 *
 *      0  1 ...  7
 *      8  9 ... 15
 *           ...   
 *     56 57 ... 63
 *
 * U and V are shipped at half resolution (1 U,V sample -> one 2x2 block).
 *
 * The next 256 bytes are full resolution Y data and represent 4 squares
 * of 8x8 pixels as follows:
 *
 *      0  1 ...  7    64  65 ...  71   ...  192 193 ... 199
 *      8  9 ... 15    72  73 ...  79        200 201 ... 207
 *           ...              ...                    ...
 *     56 57 ... 63   120 121 ... 127   ...  248 249 ... 255
 *
 * Note that the U and V data in one segment represents a 16 x 16 pixel
 * area, but the Y data represents a 32 x 8 pixel area.
 *
 * If dumppix module param is set, _parse_data just dumps the incoming segments,
 * verbatim, in order, into the frame. When used with vidcat -f ppm -s 640x480
 * this puts the data on the standard output and can be analyzed with the
 * parseppm.c utility I wrote.  That's a much faster way for figuring out how
 * this data is scrambled.
 */

#define HDIV 8
#define WDIV (256/HDIV)

static void
ov511_parse_gbr422_to_rgb24(unsigned char *pIn0, unsigned char *pOut0,
		       int iOutY, int iOutUV, int iHalf, int iWidth)
{
	int k, l, m;
	unsigned char *pIn;
	unsigned char *pOut, *pOut1;

	pIn = pIn0;
	pOut = pOut0 + iOutUV + (force_rgb ? 2 : 0);
		
	for (k = 0; k < 8; k++) {
		pOut1 = pOut;
		for (l = 0; l < 8; l++) {
			*pOut1 = *(pOut1 + 3) = *(pOut1 + iWidth*3) =
				*(pOut1 + iWidth*3 + 3) = *pIn++;
			pOut1 += 6;
		}
		pOut += iWidth*3*2;
	}

	pIn = pIn0 + 64;
	pOut = pOut0 + iOutUV + (force_rgb ? 0 : 2);
	for (k = 0; k < 8; k++) {
		pOut1 = pOut;
		for (l = 0; l < 8; l++) {
			*pOut1 = *(pOut1 + 3) = *(pOut1 + iWidth*3) =
				*(pOut1 + iWidth*3 + 3) = *pIn++;
			pOut1 += 6;
		}
		pOut += iWidth*3*2;
	}

	pIn = pIn0 + 128;
	pOut = pOut0 + iOutY + 1;
	for (k = 0; k < 4; k++) {
		pOut1 = pOut;
		for (l = 0; l < 8; l++) {
			for (m = 0; m < 8; m++) {
				*pOut1 = *pIn++;
				pOut1 += 3;
			}
			pOut1 += (iWidth - 8) * 3;
		}
		pOut += 8 * 3;
	}
}

static void
ov511_parse_yuv420_to_rgb(unsigned char *pIn0, unsigned char *pOut0,
		       int iOutY, int iOutUV, int iHalf, int iWidth, int bits)
{
	int k, l, m;
	int bytes = bits >> 3;
	unsigned char *pIn;
	unsigned char *pOut, *pOut1;

	/* Just copy the Y's if in the first stripe */
	if (!iHalf) {
		pIn = pIn0 + 128;
		pOut = pOut0 + iOutY;
		for (k = 0; k < 4; k++) {
			pOut1 = pOut;
			for (l = 0; l < 8; l++) {
				for (m = 0; m < 8; m++) {
					*pOut1 = *pIn++;
					pOut1 += bytes;
				}
				pOut1 += (iWidth - 8) * bytes;
			}
			pOut += 8 * bytes;
		}
	}

	/* Use the first half of VUs to calculate value */
	pIn = pIn0;
	pOut = pOut0 + iOutUV;
	for (l = 0; l < 4; l++) {
		for (m=0; m<8; m++) {
			int y00 = *(pOut);
			int y01 = *(pOut+bytes);
			int y10 = *(pOut+iWidth*bytes);
			int y11 = *(pOut+iWidth*bytes+bytes);
			int v   = *(pIn+64) - 128;
			int u   = *pIn++ - 128;
			ov511_move_420_block(y00, y01, y10, y11, u, v, iWidth,
				pOut, bits);
			pOut += 2 * bytes;
		}
		pOut += (iWidth*2 - 16) * bytes;
	}

	/* Just copy the other UV rows */
	for (l = 0; l < 4; l++) {
		for (m = 0; m < 8; m++) {
			*pOut++ = *(pIn + 64);
			*pOut = *pIn++;
			pOut += 2 * bytes - 1;
		}
		pOut += (iWidth*2 - 16) * bytes;
	}

	/* Calculate values if it's the second half */
	if (iHalf) {
		pIn = pIn0 + 128;
		pOut = pOut0 + iOutY;
		for (k = 0; k < 4; k++) {
			pOut1 = pOut;
			for (l=0; l<4; l++) {
				for (m=0; m<4; m++) {
					int y10 = *(pIn+8);
					int y00 = *pIn++;
					int y11 = *(pIn+8);
					int y01 = *pIn++;
					int v   = *pOut1 - 128;
					int u   = *(pOut1+1) - 128;
					ov511_move_420_block(y00, y01, y10,
						y11, u, v, iWidth, pOut1, bits);
					pOut1 += 2 * bytes;
				}
				pOut1 += (iWidth*2 - 8) * bytes;
				pIn += 8;
			}
			pOut += 8 * bytes;
		}
	}
}

static void
ov511_dumppix(struct ov511_frame *frame, 
	      unsigned char *pIn0, unsigned char *pOut0,
	      int iOutY, int iOutUV, int iHalf, int iWidth)
{
	int i, j, k;
	unsigned char *pIn, *pOut, *pOut1;

	switch (dumppix) {
	case 1: /* Just dump YUV data straight out for debug */
		pOut0 += iOutY;
		for (i = 0; i < HDIV; i++) {
			for (j = 0; j < WDIV; j++) {
				*pOut0++ = *pIn0++;
				*pOut0++ = *pIn0++;
				*pOut0++ = *pIn0++;
			}
			pOut0 += (iWidth - WDIV) * 3;
		}
		break;
	case 2: /* This converts the Y data to "black-and-white" RGB data */
		/* Useful for experimenting with compression */
		pIn = pIn0 + 128;
		pOut = pOut0 + iOutY;
		for (i = 0; i < 4; i++) {
			pOut1 = pOut;
			for (j = 0; j < 8; j++) {
				for (k = 0; k < 8; k++) {
					*pOut1++ = *pIn;
					*pOut1++ = *pIn;
					*pOut1++ = *pIn++;
				}
				pOut1 += (iWidth - 8) * 3;
			}
			pOut += 8 * 3;
		}
		break;
	case 3: /* This will dump only the Y channel data stream as-is */
		pIn = pIn0 + 128;
		pOut = pOut0 + frame->output_offset;
		for (i = 0; i < 256; i++) {
			*pOut++ = *pIn;
			*pOut++ = *pIn;
			*pOut++ = *pIn++;
			frame->output_offset += 3;
		}
		break;
	case 4: /* This will dump the whole unformatted data stream as-is */
		pIn = pIn0;
		pOut = pOut0 + frame->output_offset;
		for (i = 0; i < 384; i++) {
			*pOut++ = *pIn;
			*pOut++ = *pIn;
			*pOut++ = *pIn++;
			frame->output_offset += 3;
		}
		break;
	} /* End switch (dumppix) */
}

/* This converts YUV420 segments to YUYV */
static void
ov511_parse_data_yuv422(unsigned char *pIn0, unsigned char *pOut0,
		        int iOutY, int iOutUV, int iWidth)
{
	int k, l, m;
	unsigned char *pIn, *pOut, *pOut1;

	pIn = pIn0 + 128;
	pOut = pOut0 + iOutY;
	for (k = 0; k < 4; k++) {
		pOut1 = pOut;
		for (l = 0; l < 8; l++) {
			for (m = 0; m < 8; m++) {
				*pOut1 = (*pIn++);
				pOut1 += 2;
			}
			pOut1 += (iWidth - 8) * 2;
		}
		pOut += 8 * 2;
	}

	pIn = pIn0;
	pOut = pOut0 + iOutUV + 1;
	for (l = 0; l < 8; l++) {
		for (m=0; m<8; m++) {
			int v   = *(pIn+64);
			int u   = *pIn++;
			
			*pOut = u;
			*(pOut+2) = v;
			*(pOut+iWidth) = u;
			*(pOut+iWidth+2) = v;
			pOut += 4;
		}
		pOut += (iWidth*4 - 32);
	}
}

static void
ov511_parse_data_yuv420(unsigned char *pIn0, unsigned char *pOut0,
		        int iOutY, int iOutUV, int iWidth, int iHeight)
{
	int k, l, m;
	unsigned char *pIn;
	unsigned char *pOut, *pOut1;
	unsigned a = iWidth * iHeight;
	unsigned w = iWidth / 2;

	pIn = pIn0;
	pOut = pOut0 + iOutUV + a;
	for (k = 0; k < 8; k++) {
		pOut1 = pOut;
		for (l = 0; l < 8; l++) *pOut1++ = *pIn++;
		pOut += w;
	}

	pIn = pIn0 + 64;
	pOut = pOut0 + iOutUV + a + a/4;
	for (k = 0; k < 8; k++) {
		pOut1 = pOut;
		for (l = 0; l < 8; l++) *pOut1++ = *pIn++;
		pOut += w;
	}

	pIn = pIn0 + 128;
	pOut = pOut0 + iOutY;
	for (k = 0; k < 4; k++) {
		pOut1 = pOut;
		for (l = 0; l < 8; l++) {
			for (m = 0; m < 8; m++)
				*pOut1++ =*pIn++;
			pOut1 += iWidth - 8;
		}
		pOut += 8;
	}
}

static void
ov511_parse_data_yuv422p(unsigned char *pIn0, unsigned char *pOut0,
		       int iOutY, int iOutUV, int iWidth, int iHeight)
{
	int k, l, m;
	unsigned char *pIn;
	unsigned char *pOut, *pOut1;
	unsigned a = iWidth * iHeight;
	unsigned w = iWidth / 2;

	pIn = pIn0;
	pOut = pOut0 + iOutUV + a;
	for (k = 0; k < 8; k++) {
		pOut1 = pOut;
		for (l = 0; l < 8; l++) {
			*pOut1 = *(pOut1 + w) = *pIn++;
			pOut1++;
		}
		pOut += iWidth;
	}

	pIn = pIn0 + 64;
	pOut = pOut0 + iOutUV + a + a/2;
	for (k = 0; k < 8; k++) {
		pOut1 = pOut;
		for (l = 0; l < 8; l++) {
			*pOut1 = *(pOut1 + w) = *pIn++;
			pOut1++;
		}
		pOut += iWidth;
	}

	pIn = pIn0 + 128;
	pOut = pOut0 + iOutY;
	for (k = 0; k < 4; k++) {
		pOut1 = pOut;
		for (l = 0; l < 8; l++) {
			for (m = 0; m < 8; m++)
				*pOut1++ =*pIn++;
			pOut1 += iWidth - 8;
		}
		pOut += 8;
	}
}

/*
 * For 640x480 RAW BW images, data shows up in 1200 256 byte segments.
 * The segments represent 4 squares of 8x8 pixels as follows:
 *
 *      0  1 ...  7    64  65 ...  71   ...  192 193 ... 199
 *      8  9 ... 15    72  73 ...  79        200 201 ... 207
 *           ...              ...                    ...
 *     56 57 ... 63   120 121 ... 127        248 249 ... 255
 *
 */ 
static void
ov511_parse_data_grey(unsigned char *pIn0, unsigned char *pOut0,
		      int iOutY, int iWidth)		    
{
	int k, l, m;
	unsigned char *pIn;
	unsigned char *pOut, *pOut1;

	pIn = pIn0;
	pOut = pOut0 + iOutY;
	for (k = 0; k < 4; k++) {
		pOut1 = pOut;
		for (l = 0; l < 8; l++) {
			for (m = 0; m < 8; m++) {
				*pOut1++ = *pIn++;
			}
			pOut1 += iWidth - 8;
		}
		pOut += 8;
	}
}

/*
 * fixFrameRGBoffset--
 * My camera seems to return the red channel about 1 pixel
 * low, and the blue channel about 1 pixel high. After YUV->RGB
 * conversion, we can correct this easily. OSL 2/24/2000.
 */
static void fixFrameRGBoffset(struct ov511_frame *frame)
{
	int x, y;
	int rowBytes = frame->width*3, w = frame->width;
	unsigned char *rgb = frame->data;
	const int shift = 1;  /* Distance to shift pixels by, vertically */

	/* Don't bother with little images */
	if (frame->width < 400) 
		return;

	/* Shift red channel up */
	for (y = shift; y < frame->height; y++)	{
		int lp = (y-shift)*rowBytes;     /* Previous line offset */
		int lc = y*rowBytes;             /* Current line offset */
		for (x = 0; x < w; x++)
			rgb[lp+x*3+2] = rgb[lc+x*3+2]; /* Shift red up */
	}

	/* Shift blue channel down */
	for (y = frame->height-shift-1; y >= 0; y--) {
		int ln = (y + shift) * rowBytes;  /* Next line offset */
		int lc = y * rowBytes;            /* Current line offset */
		for (x = 0; x < w; x++)
			rgb[ln+x*3+0] = rgb[lc+x*3+0]; /* Shift blue down */
	}
}

/**********************************************************************
 *
 * Decompression
 *
 **********************************************************************/

/* Large memory buffer allocation. */
#define ALLOCATE(a, size, b) vmalloc(size)
#define FREE(addr) vfree(addr)

#ifdef ENABLE_OV511_COMPRESSION

/* Tables */
static UCHAR YQuanTable[] = YQUANTABLE;
static UCHAR UVQuanTable[] = UVQUANTABLE;
static SHORT IDCTTable[] = IDCTTABLE;
static ULONG ZigZagTable[] = ZIGZAGTABLE;

static PUCHAR GetYQuanTable(void)
{
	return YQuanTable;
}

static PUCHAR GetUVQuanTable(void)
{
	return UVQuanTable;
}

static PSHORT GetIDCTTable(void)
{
	return IDCTTable;
}

static PULONG GetZigZagTable(void)
{
	return ZigZagTable;
}

/************************************************************************/
/* Avoid changing anything below this point, to make synchronizing with */
/* newer code from OmniVision easier.					*/
/************************************************************************/

static int DecompressYHINoMMX(	LPBYTE    lpDestBuf,
				LPBYTE    lpTMPBUF,
				ULONG     index,
				DWORD     IPTRR,
				DWORD     X_WIDE,
				PUCHAR    TableBufferYUV,
				PSHORT    IDCT_Coef,
				PULONG    ZZLN,
				DWORDLONG UV_Scalor,
				ULONG     YUVFlag,
				struct    ov511_frame *frame)
{
	short ZigZag[64];
	int temp[64];
	int Zcnt_Flag=0;
	int Num8_Flag=0;
	int tmp, tmp1, tmp2, tmp3;
	unsigned char header, ZTable[64];
	short tmpl, tmph, half_byte, idx, count;
	DWORD ZigZag_length=0, ZT_length, i, j;
//	DWORD Start, k;
	short DeZigZag[64]; 

	const short a = 11584;
	const short b = 16068;
	const short c = 15136;
	const short d = 13624;
	const short e =  9104;
	const short f =  6270;
	const short g =  3196;

	int	IPTRR_YUV;

#if 0
	int	 coef[64] = 
	{
		11584,  11584,  11584,  11584,  11584,  11584,  11584,  11584,
		16068,  13624,   9104,   3196,  -3196,  -9104, -13624, -16068,
		15136,   6270,  -6270, -15136, -15136,  -6270,   6270,  15136,
		13624,  -3196, -16068,  -9104,   9104,  16068,   3196, -13624,
		11584, -11584, -11584,  11584,  11584, -11584, -11584,  11584,
		 9104, -16068,   3196,  13624, -13624,  -3196,  16068,  -9104,
		 6270, -15136,  15136,  -6270,  -6270,  15136, -15136,   6270,
		 3196,  -9104,  13624, -16068,  16068, -13624,   9104,  -3196
	};

	int	 coeft[64] = 
	{
		11584,  16068,  15136,  13624,  11584,   9104,   6270,   3196, 
		11584,  13624,   6270,  -3196, -11584, -16068, -15136,  -9104, 
		11584,   9104,  -6270, -16068, -11584,   3196,  15136,  13624, 
		11584,   3196, -15136,  -9104,  11584,  13624,  -6270, -16068, 
		11584,  -3196, -15136,   9104,  11584, -13624,  -6270,  16068, 
		11584,  -9104,  -6270,  16068, -11584,  -3196,  15136, -13624, 
		11584, -13624,   6270,   3196, -11584,  16068, -15136,   9104, 
		11584, -16068,  15136, -13624,  11584,  -9104,   6270,  -3196
	};
#endif

	for(i=0; i<64; i++)
	{
		ZigZag[i] = 0;
	}


// Read in the Y header byte  


	header = lpTMPBUF[index];
	index++;

	ZigZag_length = header & 0x3f;
	ZigZag_length = ZigZag_length + 1;

	Num8_Flag = header & 0x40;
	Zcnt_Flag = header & 0x80;

// Read in the Y content
#if 1
	if(Zcnt_Flag==0)  // Without Zero Table Read contents directly
	{

		// Read in ZigZag[0]
		ZigZag[0] = lpTMPBUF[index++];
		tmpl = lpTMPBUF[index++];
		tmph = tmpl<<8;
		ZigZag[0] = ZigZag[0] | tmph;
		ZigZag[0] = ZigZag[0]<<4;
		ZigZag[0] = ZigZag[0]>>4;

		if(Num8_Flag)  // 8 Bits
		{
			for(i=1; i<ZigZag_length; i++)
			{
				ZigZag[i] = lpTMPBUF[index++];
				ZigZag[i] = ZigZag[i]<<8;
				ZigZag[i] = ZigZag[i]>>8;

			}
		}
		else     // 12 bits and has no Zero table
		{

				idx=1;
				half_byte=0;
				for(i=1; i<ZigZag_length; i++)
				{
					if(half_byte==0)
					{
						ZigZag[i] = lpTMPBUF[index++];
						tmpl = lpTMPBUF[index++];
						tmph = tmpl<<8;
						tmph = tmph&0x0f00;
						ZigZag[i] = ZigZag[i] | tmph;
						ZigZag[i] = ZigZag[i]<<4;
						ZigZag[i] = ZigZag[i]>>4;
						half_byte=1;
					}
					else
					{
						ZigZag[i] = lpTMPBUF[index++];
						ZigZag[i] = ZigZag[i]<<8;
						tmpl = tmpl & 0x00f0;
						ZigZag[i] = ZigZag[i] | tmpl;
						ZigZag[i] = ZigZag[i]>>4;
						half_byte=0;
					}
				}

		}	// end of Num8_Flag=0
	}	//  end of without Zero Table
	else  // Has Zero Table
	{

			// calculate Z-Table length
			ZT_length = ZigZag_length/8;
			tmp = ZigZag_length%8;
			if(tmp> 0)
			{
				ZT_length = ZT_length + 1;
			}

			// read in Zero Table
			for(j=0; j<ZT_length; j++)
			{
 
				ZTable[j]=lpTMPBUF[index++];

			}

			// Read in ZigZag[0]
			ZigZag[0] = lpTMPBUF[index++];
			tmpl = lpTMPBUF[index++];
			tmph = tmpl<<8;
			ZigZag[0] = ZigZag[0] | tmph;
			ZigZag[0] = ZigZag[0]<<4;
			ZigZag[0] = ZigZag[0]>>4;

			// Decode ZiaZag
			idx=0;
			ZTable[idx]=ZTable[idx]<<1;
			count=7;

		if(Num8_Flag)  // 8 Bits and has zero table
		{

			for( i=1; i<ZigZag_length; i++)
			{
//				tmp=(ZTable[idx]&0x80);
//				if(tmp)
				if((ZTable[idx]&0x80))
				{
					ZigZag[i] = lpTMPBUF[index++];
					ZigZag[i] = ZigZag[i]<<8;
					ZigZag[i] = ZigZag[i]>>8;
				}
				
				ZTable[idx]=ZTable[idx]<<1;
				count--;
				if(count==0)
				{
					count=8;
					idx++;
						
				}
			}

		}// end of 8 bits with zero table
		else   // 12bits has zero table
		{


			half_byte=0;
			for( i=1; i<ZigZag_length; i++)
			{

				if(ZTable[idx]&0x80)
				{
					if(half_byte==0)
					{
						ZigZag[i] = lpTMPBUF[index++];
						tmpl = lpTMPBUF[index++];
						tmph = tmpl <<8;
						tmph = tmph & 0x0f00;
						ZigZag[i] = ZigZag[i] | tmph;
						ZigZag[i] = ZigZag[i]<<4;
						ZigZag[i] = ZigZag[i]>>4;
						half_byte=1;
					}
					else
					{
						ZigZag[i] = lpTMPBUF[index++];
						ZigZag[i] = ZigZag[i]<<8;
						tmpl = tmpl & 0x00f0;
						ZigZag[i] = ZigZag[i] | tmpl;
						ZigZag[i] = ZigZag[i]>>4;
						half_byte=0;
					}
				}
				
				ZTable[idx]=ZTable[idx]<<1;
				count--;
				if(count==0)
				{
					count=8;
					idx++;
						
				}
			}

		}// end of 12 bits with zero table
			
	}		
#endif 


#if 0
	//DeZigZag
	idx=0;
	for(i=0; i<8; i++)
	{
		for(j=0; j<8; j++)
		{

			DeZigZag[idx] = ZigZag[ZZTB[idx]];
			idx++;
		}
	}
#endif
#if 1	
	for(j=0; j<64; j++)
	{

		DeZigZag[j] = 0;
	}

	if(YUVFlag == 1)
	{
	DeZigZag[0] = ZigZag[0];
	DeZigZag[1] = ZigZag[1]<<1;
	DeZigZag[2] = ZigZag[5]<<1;
	DeZigZag[3] = ZigZag[6]<<2;

	DeZigZag[8] = ZigZag[2]<<1;
	DeZigZag[9] = ZigZag[4]<<1;
	DeZigZag[10] = ZigZag[7]<<1;
	DeZigZag[11] = ZigZag[13]<<2;

	DeZigZag[16] = ZigZag[3]<<1;
	DeZigZag[17] = ZigZag[8]<<1;
	DeZigZag[18] = ZigZag[12]<<2;
	DeZigZag[19] = ZigZag[17]<<2;

	DeZigZag[24] = ZigZag[9]<<2;
	DeZigZag[25] = ZigZag[11]<<2;
	DeZigZag[26] = ZigZag[18]<<2;
	DeZigZag[27] = ZigZag[24]<<3;
	}
	else 
	{
	DeZigZag[0] = ZigZag[0];
	DeZigZag[1] = ZigZag[1]<<2;
	DeZigZag[2] = ZigZag[5]<<2;
	DeZigZag[3] = ZigZag[6]<<3;

	DeZigZag[8] = ZigZag[2]<<2;
	DeZigZag[9] = ZigZag[4]<<2;
	DeZigZag[10] = ZigZag[7]<<2;
	DeZigZag[11] = ZigZag[13]<<4;

	DeZigZag[16] = ZigZag[3]<<2;
	DeZigZag[17] = ZigZag[8]<<2;
	DeZigZag[18] = ZigZag[12]<<3;
	DeZigZag[19] = ZigZag[17]<<4;

	DeZigZag[24] = ZigZag[9]<<3;
	DeZigZag[25] = ZigZag[11]<<4;
	DeZigZag[26] = ZigZag[18]<<4;
	DeZigZag[27] = ZigZag[24]<<4;
	}
#endif 

#if 0
	// Dequante
	if(YUVFlag == 1)
	{
		// Dequante Y
		for(i=0; i<64; i++)
		{
			ZigZag[i] = DeZigZag[i]<<DQTY[i];
		}
	}
	else
	{
		// Dequante UV
		for(i=0; i<64; i++)
		{
			ZigZag[i] = DeZigZag[i]<<DQTUV[i];
		}
	}
#endif 
#if 1
	// =============
	// == IDCT 1D ==
	// =============


//	for(i=0;i<8; i++)
//	{
//		for(j=0; j<4; j++)
//		{
			tmp1=a*DeZigZag[0]+c*DeZigZag[2];
			tmp2=b*DeZigZag[1];
			tmp3=d*DeZigZag[3];

			tmp=tmp1+tmp2+tmp3;
			temp[0] = tmp>>15;

			tmp=tmp1-tmp2-tmp3;
			temp[56] = tmp>>15;

			tmp1=a*DeZigZag[8]+c*DeZigZag[10];
			tmp2=b*DeZigZag[9];
			tmp3=d*DeZigZag[11];

			tmp=tmp1+tmp2+tmp3;
			temp[1] = tmp>>15;

			tmp=tmp1-tmp2-tmp3;
			temp[57] = tmp>>15;

			tmp1=a*DeZigZag[16]+c*DeZigZag[18];
			tmp2=b*DeZigZag[17];
			tmp3=d*DeZigZag[19];

			tmp=tmp1+tmp2+tmp3;
			temp[2] = tmp>>15;

			tmp=tmp1-tmp2-tmp3;
			temp[58] = tmp>>15;

			tmp1=a*DeZigZag[24]+c*DeZigZag[26];
			tmp2=b*DeZigZag[25];
			tmp3=d*DeZigZag[27];

			tmp=tmp1+tmp2+tmp3;
			temp[3] = tmp>>15;

			tmp=tmp1-tmp2-tmp3;
			temp[59] = tmp>>15;

//		for(j=0; j<4; j++)
//		{
			tmp1=a*DeZigZag[0]+f*DeZigZag[2];
			tmp2=d*DeZigZag[1];
			tmp3=g*DeZigZag[3];

			tmp=tmp1+tmp2-tmp3;
			temp[8] = tmp>>15;

			tmp=tmp1-tmp2+tmp3;
			temp[48] = tmp>>15;

			tmp1=a*DeZigZag[8]+f*DeZigZag[10];
			tmp2=d*DeZigZag[9];
			tmp3=g*DeZigZag[11];

			tmp=tmp1+tmp2-tmp3;
			temp[9] = tmp>>15;

			tmp=tmp1-tmp2+tmp3;
			temp[49] = tmp>>15;

			tmp1=a*DeZigZag[16]+f*DeZigZag[18];
			tmp2=d*DeZigZag[17];
			tmp3=g*DeZigZag[19];

			tmp=tmp1+tmp2-tmp3;
			temp[10] = tmp>>15;

			tmp=tmp1-tmp2+tmp3;
			temp[50] = tmp>>15;

			tmp1=a*DeZigZag[24]+f*DeZigZag[26];
			tmp2=d*DeZigZag[25];
			tmp3=g*DeZigZag[27];

			tmp=tmp1+tmp2-tmp3;
			temp[11] = tmp>>15;

			tmp=tmp1-tmp2+tmp3;
			temp[51] = tmp>>15;
//		}

//		for(j=0; j<4; j++)
//		{
			tmp1=a*DeZigZag[0]-f*DeZigZag[2];
			tmp2=e*DeZigZag[1];
			tmp3=b*DeZigZag[3];

			tmp=tmp1+tmp2-tmp3;
			temp[16] = tmp>>15;

			tmp=tmp1-tmp2+tmp3;
			temp[40] = tmp>>15;

			tmp1=a*DeZigZag[8]-f*DeZigZag[10];
			tmp2=e*DeZigZag[9];
			tmp3=b*DeZigZag[11];

			tmp=tmp1+tmp2-tmp3;
			temp[17] = tmp>>15;

			tmp=tmp1-tmp2+tmp3;
			temp[41] = tmp>>15;

			tmp1=a*DeZigZag[16]-f*DeZigZag[18];
			tmp2=e*DeZigZag[17];
			tmp3=b*DeZigZag[19];

			tmp=tmp1+tmp2-tmp3;
			temp[18] = tmp>>15;

			tmp=tmp1-tmp2+tmp3;
			temp[42] = tmp>>15;

			tmp1=a*DeZigZag[24]-f*DeZigZag[26];
			tmp2=e*DeZigZag[25];
			tmp3=b*DeZigZag[27];

			tmp=tmp1+tmp2-tmp3;
			temp[19] = tmp>>15;

			tmp=tmp1-tmp2+tmp3;
			temp[43] = tmp>>15;
//		}

//		for(j=0; j<4; j++)
//		{
			tmp1=a*DeZigZag[0]-c*DeZigZag[2];
			tmp2=g*DeZigZag[1];
			tmp3=e*DeZigZag[3];

			tmp=tmp1+tmp2-tmp3;
			temp[24] = tmp>>15;

			tmp=tmp1-tmp2+tmp3;
			temp[32] = tmp>>15;

			tmp1=a*DeZigZag[8]-c*DeZigZag[10];
			tmp2=g*DeZigZag[9];
			tmp3=e*DeZigZag[11];

			tmp=tmp1+tmp2-tmp3;
			temp[25] = tmp>>15;

			tmp=tmp1-tmp2+tmp3;
			temp[33] = tmp>>15;

			tmp1=a*DeZigZag[16]-c*DeZigZag[18];
			tmp2=g*DeZigZag[17];
			tmp3=e*DeZigZag[19];

			tmp=tmp1+tmp2-tmp3;
			temp[26] = tmp>>15;

			tmp=tmp1-tmp2+tmp3;
			temp[34] = tmp>>15;

			tmp1=a*DeZigZag[24]-c*DeZigZag[26];
			tmp2=g*DeZigZag[25];
			tmp3=e*DeZigZag[27];

			tmp=tmp1+tmp2-tmp3;
			temp[27] = tmp>>15;

			tmp=tmp1-tmp2+tmp3;
			temp[35] = tmp>>15;
//		}
		
//		for(j=0; j<4; j++)
//		{
/*			
			tmp = 0;
			for(k=0; k<4; k++)
			{
				tmp = tmp + coeft[32+k]*DeZigZag[8*j + k];
			}
			temp[32+j] = tmp>>15;
*/	
//		}

//		for(j=0; j<4; j++)
//		{
			
/*			tmp=a*DeZigZag[0] - e*DeZigZag[1] - f*DeZigZag[2] + b*DeZigZag[3];
			temp[40] = tmp>>15;
			
			tmp=a*DeZigZag[8] - e*DeZigZag[9] - f*DeZigZag[10] + b*DeZigZag[11];
			temp[41] = tmp>>15;
			
			tmp=a*DeZigZag[16] - e*DeZigZag[17] - f*DeZigZag[18] + b*DeZigZag[19];
			temp[42] = tmp>>15;
			
			tmp=a*DeZigZag[24] - e*DeZigZag[25] - f*DeZigZag[26] + b*DeZigZag[27];
			temp[43] = tmp>>15;
*//*
			tmp = 0;
			for(k=0; k<4; k++)
			{
				tmp = tmp + coeft[40 + k]*DeZigZag[8*j + k];
			}
			temp[40+j] = tmp>>15;
*/
//		}

//		for(j=0; j<4; j++)
//		{
			
/*			tmp=a*DeZigZag[0] - d*DeZigZag[1]+ f*DeZigZag[2] + g*DeZigZag[3];
			temp[48] = tmp>>15;
			
			tmp=a*DeZigZag[8] - d*DeZigZag[9] + f*DeZigZag[10] + g*DeZigZag[11];
			temp[49] = tmp>>15;
			
			tmp=a*DeZigZag[16] - d*DeZigZag[17] + f*DeZigZag[18] + g*DeZigZag[19];
			temp[50] = tmp>>15;
			
			tmp=a*DeZigZag[24] - d*DeZigZag[25] + f*DeZigZag[26] + g*DeZigZag[27];
			temp[51] = tmp>>15;
*/
//		}

//		for(j=0; j<4; j++)
//		{
			
/*			
			tmp = a*DeZigZag[0] - b*DeZigZag[1]+ c*DeZigZag[2] - d*DeZigZag[3];
			temp[56] = tmp>>15;
			
			tmp = a*DeZigZag[8] - b*DeZigZag[9]+ c*DeZigZag[10] - d*DeZigZag[11];
			temp[57] = tmp>>15;
			
			tmp = a*DeZigZag[16] - b*DeZigZag[17]+ c*DeZigZag[18] - d*DeZigZag[19];
			temp[58] = tmp>>15;
			
			tmp = a*DeZigZag[24] - b*DeZigZag[25]+ c*DeZigZag[26] - d*DeZigZag[27];
			temp[59] = tmp>>15;
*/

//		}


//	}

#endif 
	// =============
	// == IDCT 2D ==
	// =============
#if 1
//	for(i=0; i<8; i++)
//	{
			IPTRR_YUV = IPTRR;

//      i=0	
//		for(j=0; j<8; j++)
//		{ 
			// j=0
			tmp = temp[0]*a+temp[1]*b+temp[2]*c+temp[3]*d;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[0+IPTRR] = (BYTE) tmp;		

			// j=1
			tmp = temp[0]*a+temp[1]*d+temp[2]*f+temp[3]*(-g);
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[1+IPTRR] = (BYTE) tmp;		

			// j=2
			tmp = temp[0]*a+temp[1]*e+temp[2]*(-f)+temp[3]*(-b);
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[2+IPTRR] = (BYTE) tmp;		

			// j=3
			tmp = temp[0]*a+temp[1]*g+temp[2]*(-c)+temp[3]*(-e);
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[3+IPTRR] = (BYTE) tmp;	

			// j=4
			tmp = temp[0]*a+temp[1]*(-g)+temp[2]*(-c)+temp[3]*e;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[4+IPTRR] = (BYTE) tmp;		

			// j=5
			tmp = temp[0]*a+temp[1]*(-e)+temp[2]*(-f)+temp[3]*b;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[5+IPTRR] = (BYTE) tmp;		

			// j=6
			tmp = temp[0]*a+temp[1]*(-d)+temp[2]*f+temp[3]*g;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[6+IPTRR] = (BYTE) tmp;		

			// j=7
			tmp = temp[0]*a+temp[1]*(-b)+temp[2]*c+temp[3]*(-d);
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[7+IPTRR] = (BYTE) tmp;		
//		}

//		i=1
//		for(j=0; j<8; j++)
//		{ 
			IPTRR_YUV = X_WIDE+IPTRR_YUV;
			// j=0
			tmp = temp[8]*a+temp[9]*b+temp[10]*c+temp[11]*d;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV] = (BYTE) tmp;		

			// j=1
			tmp = temp[8]*a+temp[9]*d+temp[10]*f+temp[11]*(-g);
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+1] = (BYTE) tmp;

			// j=2
			tmp = temp[8]*a+temp[9]*e+temp[10]*(-f)+temp[11]*(-b);
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+2] = (BYTE) tmp;

			// j=3
			tmp = temp[8]*a+temp[9]*g+temp[10]*(-c)+temp[11]*(-e);
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+3] = (BYTE) tmp;		

			// j=4
			tmp = temp[8]*a+temp[9]*(-g)+temp[10]*(-c)+temp[11]*e;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+4] = (BYTE) tmp;		

			// j=5
			tmp = temp[8]*a+temp[9]*(-e)+temp[10]*(-f)+temp[11]*b;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+5] = (BYTE) tmp;		

			// j=6
			tmp = temp[8]*a+temp[9]*(-d)+temp[10]*f+temp[11]*g;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+6] = (BYTE) tmp;		

			// j=7
			tmp = temp[8]*a+temp[9]*(-b)+temp[10]*c+temp[11]*(-d);
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+7] = (BYTE) tmp;		
//		}

		IPTRR_YUV = X_WIDE+IPTRR_YUV;

//		i=2
//		for(j=0; j<8; j++)
//		{ 
			// j=0
			tmp = temp[16]*a+temp[17]*b+temp[18]*c+temp[19]*d;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV] = (BYTE) tmp;		

			// j=1
			tmp = temp[16]*a+temp[17]*d+temp[18]*f+temp[19]*(-g);
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+1] = (BYTE) tmp;

			// j=2
			tmp = temp[16]*a+temp[17]*e+temp[18]*(-f)+temp[19]*(-b);
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+2] = (BYTE) tmp;

			// j=3
			tmp = temp[16]*a+temp[17]*g+temp[18]*(-c)+temp[19]*(-e);
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+3] = (BYTE) tmp;		

			// j=4
			tmp = temp[16]*a+temp[17]*(-g)+temp[18]*(-c)+temp[19]*e;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+4] = (BYTE) tmp;		

			// j=5
			tmp = temp[16]*a+temp[17]*(-e)+temp[18]*(-f)+temp[19]*b;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+5] = (BYTE) tmp;		

			// j=6
			tmp = temp[16]*a+temp[17]*(-d)+temp[18]*f+temp[19]*g;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+6] = (BYTE) tmp;		

			// j=7
			tmp = temp[16]*a+temp[17]*(-b)+temp[18]*c+temp[19]*(-d);
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+7] = (BYTE) tmp;		
//		}
		IPTRR_YUV = X_WIDE+IPTRR_YUV;
//		i=3
//		for(j=0; j<8; j++)
//		{ 
			// j=0
			tmp = temp[24]*a+temp[25]*b+temp[26]*c+temp[27]*d;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV] = (BYTE) tmp;		

			// j=1
			tmp = temp[24]*a+temp[25]*d+temp[26]*f+temp[27]*(-g);
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+1] = (BYTE) tmp;

			// j=2
			tmp = temp[24]*a+temp[25]*e+temp[26]*(-f)+temp[27]*(-b);
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+2] = (BYTE) tmp;

			// j=3
			tmp = temp[24]*a+temp[25]*g+temp[26]*(-c)+temp[27]*(-e);
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+3] = (BYTE) tmp;		

			// j=4
			tmp = temp[24]*a+temp[25]*(-g)+temp[26]*(-c)+temp[27]*e;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+4] = (BYTE) tmp;		

			// j=5
			tmp = temp[24]*a+temp[25]*(-e)+temp[26]*(-f)+temp[27]*b;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+5] = (BYTE) tmp;		

			// j=6
			tmp = temp[24]*a+temp[25]*(-d)+temp[26]*f+temp[27]*g;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+6] = (BYTE) tmp;		

			// j=7
			tmp = temp[24]*a+temp[25]*(-b)+temp[26]*c+temp[27]*(-d);
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+7] = (BYTE) tmp;		
//		}
		IPTRR_YUV = X_WIDE+IPTRR_YUV;
//		i=4
//		for(j=0; j<8; j++)
//		{ 
			// j=0
			tmp = temp[32]*a+temp[33]*b+temp[34]*c+temp[35]*d;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV] = (BYTE) tmp;		

			// j=1
			tmp = temp[32]*a+temp[33]*d+temp[34]*f+temp[35]*(-g);
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+1] = (BYTE) tmp;

			// j=2
			tmp = temp[32]*a+temp[33]*e+temp[34]*(-f)+temp[35]*(-b);
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+2] = (BYTE) tmp;

			// j=3
			tmp = temp[32]*a+temp[33]*g+temp[34]*(-c)+temp[35]*(-e);
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+3] = (BYTE) tmp;		

			// j=4
			tmp = temp[32]*a+temp[33]*(-g)+temp[34]*(-c)+temp[35]*e;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+4] = (BYTE) tmp;		

			// j=5
			tmp = temp[32]*a+temp[33]*(-e)+temp[34]*(-f)+temp[35]*b;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+5] = (BYTE) tmp;		

			// j=6
			tmp = temp[32]*a+temp[33]*(-d)+temp[34]*f+temp[35]*g;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+6] = (BYTE) tmp;		

			// j=7
			tmp = temp[32]*a+temp[33]*(-b)+temp[34]*c+temp[35]*(-d);
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+7] = (BYTE) tmp;		
//		}
		IPTRR_YUV = X_WIDE+IPTRR_YUV;
//		i=5
//		for(j=0; j<8; j++)
//		{ 
			// j=0
			tmp = temp[40]*a+temp[41]*b+temp[42]*c+temp[43]*d;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV] = (BYTE) tmp;		

			// j=1
			tmp = temp[40]*a+temp[41]*d+temp[42]*f+temp[43]*(-g);
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+1] = (BYTE) tmp;

			// j=2
			tmp = temp[40]*a+temp[41]*e+temp[42]*(-f)+temp[43]*(-b);
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+2] = (BYTE) tmp;

			// j=3
			tmp = temp[40]*a+temp[41]*g+temp[42]*(-c)+temp[43]*(-e);
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+3] = (BYTE) tmp;		

			// j=4
			tmp = temp[40]*a+temp[41]*(-g)+temp[42]*(-c)+temp[43]*e;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+4] = (BYTE) tmp;		

			// j=5
			tmp = temp[40]*a+temp[41]*(-e)+temp[42]*(-f)+temp[43]*b;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+5] = (BYTE) tmp;		

			// j=6
			tmp = temp[40]*a+temp[41]*(-d)+temp[42]*f+temp[43]*g;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+6] = (BYTE) tmp;		

			// j=7
			tmp = temp[40]*a+temp[41]*(-b)+temp[42]*c+temp[43]*(-d);
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+7] = (BYTE) tmp;		

//		}
		IPTRR_YUV = X_WIDE+IPTRR_YUV;
//		i=6
//		for(j=0; j<8; j++)
//		{ 
			tmp1=temp[48]*a+temp[50]*c;
			tmp2=temp[49]*b;
			tmp3=temp[51]*d;
			// j=0
			tmp=tmp1+tmp2+tmp3;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV] = (BYTE) tmp;		

			// j=7
			tmp=tmp1-tmp2-tmp3;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+7] = (BYTE) tmp;		

			tmp1=temp[48]*a+temp[50]*f;
			tmp2=temp[49]*d;
			tmp3=temp[51]*g;
			// j=1
			tmp=tmp1+tmp2-tmp3;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+1] = (BYTE) tmp;

			// j=6
			tmp=tmp1-tmp2+tmp3;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+6] = (BYTE) tmp;		
			
			tmp1=temp[48]*a-temp[50]*f;
			tmp2=temp[49]*e;
			tmp3=temp[51]*b;
			// j=2
			tmp=tmp1+tmp2-tmp3;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+2] = (BYTE) tmp;

			// j=5
			tmp=tmp1-tmp2+tmp3;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+5] = (BYTE) tmp;		

			tmp1=temp[48]*a-temp[50]*c;
			tmp2=temp[49]*g;
			tmp3=temp[51]*e;
			// j=3
			tmp=tmp1+tmp2-tmp3;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+3] = (BYTE) tmp;		

			// j=4
			tmp=tmp1-tmp2+tmp3;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+4] = (BYTE) tmp;		
//		}
		IPTRR_YUV = X_WIDE+IPTRR_YUV;
//		i=7
//		for(j=0; j<8; j++)
//		{ 
			tmp1=temp[56]*a+temp[58]*c;
			tmp2=temp[57]*b;
			tmp3=temp[59]*d;

			// j=0
			tmp=tmp1+tmp2+tmp3;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV] = (BYTE) tmp;		

			// j=7
			tmp=tmp1-tmp2-tmp3;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+7] = (BYTE) tmp;		

			tmp1=temp[56]*a+temp[58]*f;
			tmp2=temp[57]*d;
			tmp3=temp[59]*g;

			// j=1
			tmp=tmp1+tmp2-tmp3;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+1] = (BYTE) tmp;

			// j=6
			tmp=tmp1-tmp2+tmp3;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+6] = (BYTE) tmp;		

			tmp1=temp[56]*a-temp[58]*f;
			tmp2=temp[57]*e;
			tmp3=temp[59]*b;

			// j=2
			tmp=tmp1+tmp2-tmp3;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+2] = (BYTE) tmp;

			// j=5
			tmp=tmp1-tmp2+tmp3;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+5] = (BYTE) tmp;		

			tmp1=temp[56]*a-temp[58]*c;
			tmp2=temp[57]*g;
			tmp3=temp[59]*e;

			// j=3
			tmp=tmp1+tmp2-tmp3;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+3] = (BYTE) tmp;		

			// j=4
			tmp=tmp1-tmp2+tmp3;
			tmp = tmp>>15;
			tmp=tmp+128;
			if(tmp >255) tmp=255;
			if(tmp < 0 ) tmp=0;
			lpDestBuf[IPTRR_YUV+4] = (BYTE) tmp;		

//		}

	
//	}

#endif 
	frame->TempInfo.index=index;
	frame->TempInfo.IPTRR_U=IPTRR+8;

	return 0;
} 

static NTSTATUS 
Decompress420HiNoMMX(
	PUCHAR  pY,
	ULONG   FrameLength,
	ULONG   X_SIZE, 
	ULONG   Y_SIZE, 
	PUCHAR	TableBuffer,
	PUCHAR	TableUVBuffer,
	PSHORT  CoefBuffer,
	PULONG  ZigZagTable,
	PUCHAR  RawFrameBuffer,
	ULONG   ProcessedDataLength,	
	PUCHAR  TempBuffer,
	struct  ov511_frame *frame)

{

// Begin of decoder definition 
//POVTCAM_DEVICE_CONTEXT deviceContext = DeviceContext;
LPBYTE lpYUV, lpU, lpV, lpTMPBUF;
ULONG	PTRX, PTRY;
// ULONG i;
ULONG	IPTRR, IPTRR_U, IPTRR_V;
//ULONG   ipm, tms, Tcnt=1600;
ULONG   index;
ULONG   SrcUVDataLength;
//ULONG   SrcYDataLength;
//short  a,b,c,d,e,mid;

PUCHAR  TempBuffU = NULL; 
PUCHAR  TempBuffV = NULL; 
PUCHAR  TempBuffY = NULL; 

DWORDLONG   Y_Scalor=0x7fff7fff7fff7fff;

//   u=0.564
DWORDLONG   U_Scalor=0x4831483148314831;
//   v=0.713
DWORDLONG   V_Scalor=0x5b435b435b435b43;

//DWORDLONG   U_Scalor=0x7fff7fff7fff7fff;
//DWORDLONG   V_Scalor=0x7fff7fff7fff7fff;
//DWORDLONG   CONST128=0x8080808080808080;
//DWORDLONG   CONST128=0x0080008000800080; /* Previously disabled by OVT */
//DWORDLONG   Const4000=0x4000400040004000;


// 	TempBuffer = deviceContext->TempBuffer;
//	TableBuffer = deviceContext->TableBuffer;  //TempBuffer;
//	CoefBuffer = deviceContext->CoefBuffer;  //TempBuffer;
 //	ZigZagTable = deviceContext->ZigZagTable;
 	lpTMPBUF = RawFrameBuffer; // + 9;  //TempBuffer;

//ProcessedDataLength = deviceContext->ProcessedDataLength;
//ProcessedDataLength = ProcessedDataLength/8;
//ProcessedDataLength = ProcessedDataLength*8 + 8;

SrcUVDataLength = X_SIZE  * Y_SIZE/4;
//SrcYDataLength = X_SIZE  * Y_SIZE;

TempBuffU= TempBuffer;
TempBuffV= TempBuffU + SrcUVDataLength;
TempBuffY= TempBuffV + SrcUVDataLength;

		   lpYUV = pY;
		     lpU = lpYUV+X_SIZE*Y_SIZE;
		     lpV = lpU + X_SIZE*Y_SIZE/4;

index=0;
// ===== 320x240 images    
 for (PTRY=0; PTRY<Y_SIZE; PTRY=PTRY+16)
 {// start Y Loop

IPTRR=X_SIZE*PTRY;
IPTRR_U=(IPTRR)>>2;
IPTRR_V=IPTRR_U; 
if(index > ProcessedDataLength)
return STATUS_UNSUCCESSFUL;
 for (PTRX=0; PTRX<X_SIZE; PTRX=PTRX+32)
	{ // start X loop
		// ##############################
		// # Decompression				#
		// ##############################
	     /* lpDestBuf, lpTMPBUF, index, IPTRR, X_WIDE, TableBufferYUV, IDCT_Coef, ZZLN, UV_Scalor, YUVFlag */
//DecompressYHI(lpU, lpTMPBUF, index, IPTRR_U, X_SIZE/2,  TableBuffer+64, CoefBuffer, ZigZagTable, U_Scalor, 2, frame);
DecompressYHINoMMX(lpU, lpTMPBUF, index, IPTRR_U, X_SIZE/2,  TableUVBuffer, CoefBuffer, ZigZagTable, U_Scalor, 2, frame);
index=frame->TempInfo.index;
IPTRR_U=frame->TempInfo.IPTRR_U;

//DecompressYHINoMMX(lpV, lpTMPBUF, index, IPTRR_V, X_SIZE/2, TableBuffer+64, CoefBuffer, ZigZagTable, V_Scalor, 2, frame);
DecompressYHINoMMX(lpV, lpTMPBUF, index, IPTRR_V, X_SIZE/2, TableUVBuffer, CoefBuffer, ZigZagTable, V_Scalor, 2, frame);
index=frame->TempInfo.index;
IPTRR_V=frame->TempInfo.IPTRR_U;

DecompressYHINoMMX(lpYUV, lpTMPBUF, index, IPTRR, X_SIZE, TableBuffer, CoefBuffer, ZigZagTable, Y_Scalor, 1, frame);
//DecompressYHINoMMX(TempBuffY, lpTMPBUF, index, IPTRR, X_SIZE, TableBuffer, CoefBuffer, ZigZagTable, Y_Scalor, 1, frame);
index=frame->TempInfo.index;
IPTRR=frame->TempInfo.IPTRR_U;

DecompressYHINoMMX(lpYUV, lpTMPBUF, index, IPTRR, X_SIZE, TableBuffer, CoefBuffer, ZigZagTable, Y_Scalor, 1, frame);
//DecompressYHINoMMX(TempBuffY, lpTMPBUF, index, IPTRR, X_SIZE, TableBuffer, CoefBuffer, ZigZagTable, Y_Scalor, 1, frame);
index=frame->TempInfo.index;
IPTRR=frame->TempInfo.IPTRR_U;
 
DecompressYHINoMMX(lpYUV, lpTMPBUF, index, IPTRR, X_SIZE,  TableBuffer, CoefBuffer, ZigZagTable, Y_Scalor, 1, frame);
//DecompressYHINoMMX(TempBuffY, lpTMPBUF, index, IPTRR, X_SIZE, TableBuffer, CoefBuffer, ZigZagTable, Y_Scalor, 1, frame);
index=frame->TempInfo.index;
IPTRR=frame->TempInfo.IPTRR_U;

DecompressYHINoMMX(lpYUV, lpTMPBUF, index, IPTRR, X_SIZE,  TableBuffer, CoefBuffer, ZigZagTable, Y_Scalor, 1, frame);
//DecompressYHINoMMX(TempBuffY, lpTMPBUF, index, IPTRR, X_SIZE, TableBuffer, CoefBuffer, ZigZagTable, Y_Scalor, 1, frame);
index=frame->TempInfo.index;
IPTRR=frame->TempInfo.IPTRR_U;
	 } //Stop X


IPTRR=X_SIZE*(PTRY+8);
IPTRR_V=(X_SIZE*PTRY + X_SIZE); // /4;
IPTRR_V=IPTRR_V>>2;
IPTRR_U=IPTRR_V;
 
 for (PTRX=0; PTRX<X_SIZE; PTRX=PTRX+32)
	{ // start X loop
		// ##############################
		// # Decompression				#
		// ##############################

//DecompressYHI(lpU, lpTMPBUF, index, IPTRR_U, X_SIZE/2, TableBuffer+64,CoefBuffer, ZigZagTable, U_Scalor, 2, frame);
DecompressYHINoMMX(lpU, lpTMPBUF, index, IPTRR_U, X_SIZE/2,  TableUVBuffer, CoefBuffer, ZigZagTable, U_Scalor, 2, frame);
index=frame->TempInfo.index;
IPTRR_U=frame->TempInfo.IPTRR_U;

//DecompressYHINoMMX(lpV, lpTMPBUF, index, IPTRR_V, X_SIZE/2, TableBuffer+64, CoefBuffer, ZigZagTable, V_Scalor, 2, frame);
DecompressYHINoMMX(lpV, lpTMPBUF, index, IPTRR_V, X_SIZE/2, TableUVBuffer, CoefBuffer, ZigZagTable, V_Scalor, 2, frame);
index=frame->TempInfo.index;
IPTRR_V=frame->TempInfo.IPTRR_U;

DecompressYHINoMMX(lpYUV, lpTMPBUF, index, IPTRR, X_SIZE, TableBuffer, CoefBuffer, ZigZagTable, Y_Scalor, 1, frame);
//DecompressYHINoMMX(TempBuffY, lpTMPBUF, index, IPTRR, X_SIZE, TableBuffer, CoefBuffer, ZigZagTable, Y_Scalor, 1, frame);
index=frame->TempInfo.index;
IPTRR=frame->TempInfo.IPTRR_U;

DecompressYHINoMMX(lpYUV, lpTMPBUF, index, IPTRR, X_SIZE,  TableBuffer, CoefBuffer, ZigZagTable, Y_Scalor, 1, frame);
//DecompressYHINoMMX(TempBuffY, lpTMPBUF, index, IPTRR, X_SIZE, TableBuffer, CoefBuffer, ZigZagTable, Y_Scalor, 1, frame);
index=frame->TempInfo.index;
IPTRR=frame->TempInfo.IPTRR_U;
 
DecompressYHINoMMX(lpYUV, lpTMPBUF, index, IPTRR, X_SIZE,  TableBuffer, CoefBuffer, ZigZagTable, Y_Scalor, 1, frame);
//DecompressYHINoMMX(TempBuffY, lpTMPBUF, index, IPTRR, X_SIZE, TableBuffer, CoefBuffer, ZigZagTable, Y_Scalor, 1, frame);
index=frame->TempInfo.index;
IPTRR=frame->TempInfo.IPTRR_U;

DecompressYHINoMMX(lpYUV, lpTMPBUF, index, IPTRR, X_SIZE, TableBuffer, CoefBuffer, ZigZagTable, Y_Scalor, 1, frame);
//DecompressYHINoMMX(TempBuffY, lpTMPBUF, index, IPTRR, X_SIZE, TableBuffer, CoefBuffer, ZigZagTable, Y_Scalor, 1, frame);
index=frame->TempInfo.index;
IPTRR=frame->TempInfo.IPTRR_U;
	 } //Stop X
}// stop Y


	return 0;
}




/* Input (RawFrameBuffer) format is raw isoc. data (with header and packet 
 * nunber stripped).
 * Output (FrameBuffer) format is planar YUV420
 * Returns uncompressed data length if success, or zero if error
 */
static ULONG 
Decompress420(	ULONG SrcWidth,
		ULONG SrcHeight,
		PVOID FrameBuffer,
		ULONG FrameLength,
		PVOID RawFrameBuffer,
		ULONG RawFrameLength,
		ULONG ActualRawFrameLength,
		struct ov511_frame *frame)
{
	ULONG dwLength = SrcWidth * SrcHeight;
	PUCHAR pBuffer;
//	PUCHAR pOutBuffer;
	PUCHAR pYQuanTable;
	PUCHAR pUVQuanTable;
	PSHORT pIDCTTable;
	PULONG pZigAzgTable;
	NTSTATUS ntStatus = STATUS_SUCCESS;
//	UCHAR YTable[QUANTABLESIZE], UVTable[QUANTABLESIZE];
//	UCHAR Temp;
//	ULONG i;

	PDEBUG(4, "w=%lu h=%lu fb=%p fblen=%lu rawfb=%p rawlen=%lu, actlen=%lu",
		SrcWidth, SrcHeight, FrameBuffer, FrameLength, RawFrameBuffer,
		RawFrameLength, ActualRawFrameLength);

//#ifdef OV518
//	pgFrameContext = m_pCurrentFrameContext;
////	OVTCAM_KdPrint(MIN_TRACE, ("OVTCAM: YUV=420 \n")); 
//#endif
	if(frame->m_pTempBuffer && frame->m_TempBufferLength < dwLength)
	{
		FREE(frame->m_pTempBuffer);
		frame->m_pTempBuffer = NULL;
		frame->m_TempBufferLength = 0;
	}
	if(frame->m_pTempBuffer == NULL)
	{
		frame->m_pTempBuffer = ALLOCATE(NonPagedPool, dwLength,0);
		frame->m_TempBufferLength = dwLength ;
	}
	if(frame->m_pTempBuffer == NULL)
		return 0;

	pBuffer = (PUCHAR)frame->m_pTempBuffer;
//	pOutBuffer = (PUCHAR)FrameBuffer;
	pYQuanTable = GetYQuanTable();
	pUVQuanTable = GetUVQuanTable();
	pIDCTTable = GetIDCTTable();
	pZigAzgTable = GetZigZagTable();

//#if 0
//	//def OV518
//		for(int i = 0; i < QUANTABLESIZE; i++)
//		{
//			Temp = *pYQuanTable++;
//			if(Temp > 4) YTable[i] = 3;
//			else YTable[i] = Temp - 2;
//
//			Temp = *pUVQuanTable++;
//			if(Temp > 4) UVTable[i] = 3;
//			else UVTable[i] = Temp - 2;
//		}
//#endif

//	if(m_bSupportMMX)
//	{
//#ifdef OV518
//
//#ifdef  TABLE_DBG
//		ntStatus = Decompress420OV518NoMMXTABLE((PUCHAR) FrameBuffer, FrameLength, SrcWidth, SrcHeight, (PUCHAR)pYQuanTable, (PUCHAR)pUVQuanTable,pIDCTTable,pZigAzgTable,(PUCHAR) RawFrameBuffer,ActualRawFrameLength,pBuffer); 
//#else
//		ntStatus = Decompress420OV518((PUCHAR) FrameBuffer, FrameLength, SrcWidth, SrcHeight, (PUCHAR)YTable, (PUCHAR)UVTable,pIDCTTable,pZigAzgTable,(PUCHAR) RawFrameBuffer,ActualRawFrameLength,pBuffer); 
//#endif
//
//#else
//		ntStatus = Decompress420Hi((PUCHAR) FrameBuffer, FrameLength, SrcWidth, SrcHeight, pYQuanTable, pUVQuanTable,pIDCTTable,pZigAzgTable,(PUCHAR) RawFrameBuffer,ActualRawFrameLength,pBuffer);
//#endif
//	}
//	else
//	{

//#ifdef OV518
//#ifdef  TABLE_DBG
//		ntStatus = Decompress420OV518NoMMXTABLE((PUCHAR) FrameBuffer, FrameLength, SrcWidth, SrcHeight, (PUCHAR)pYQuanTable, (PUCHAR)pUVQuanTable,pIDCTTable,pZigAzgTable,(PUCHAR) RawFrameBuffer,ActualRawFrameLength,pBuffer); 
//#else
//		ntStatus = Decompress420OV518NoMMX((PUCHAR) FrameBuffer, FrameLength, SrcWidth, SrcHeight, (PUCHAR)YTable, (PUCHAR)UVTable,pIDCTTable,pZigAzgTable,(PUCHAR) RawFrameBuffer,ActualRawFrameLength,pBuffer); 
//#endif
//#else
		ntStatus = Decompress420HiNoMMX((PUCHAR) FrameBuffer, FrameLength, SrcWidth, SrcHeight, pYQuanTable, pUVQuanTable,pIDCTTable,pZigAzgTable,(PUCHAR) RawFrameBuffer,ActualRawFrameLength,pBuffer,frame);
//#endif

// Disabled by Mark
//		//for color saturation 8/22/00
//		{
//			LONG u , v;
//			PUCHAR pU = (PUCHAR) FrameBuffer + dwLength;
//			PUCHAR pV = pU + dwLength/4;
//			for(ULONG i = 0; i < dwLength/4; i++)
//			{
//				u = (LONG)*pU;
//				v = (LONG)*pV;
//				u -= 128;
//				v -= 128;
//				u = (u * 577) >> 10;
//				v = (v * 730) >> 10;
//				u += 128;
//				v += 128;
//				*pU++ = (UCHAR)u;
//				*pV++ = (UCHAR)v;
////   u=0.564
////   v=0.713
//			}
//		}


//	}

	// Added by Steven 11/06/2000
	if(ntStatus) return 0;

//	if(m_EnColorCorrect)
//	{
//		ColorCorrect420((PUCHAR) FrameBuffer, FrameLength, SrcWidth, SrcHeight);
//	}

//	OVTCAM_CopyMemory(pOutBuffer, RawFrameBuffer, dwLength * 3 / 2);
//	if(!NT_SUCCESS(ntStatus))
//		return 0;
	return dwLength * 3 / 2;
}
#endif

static void ov511_decompress(struct ov511_frame *frame)
{
	unsigned long rc;

//	err("Not implemented yet (but given %d bytes)", frame->bytes_recvd);

#ifdef ENABLE_OV511_COMPRESSION

	PDEBUG(4, "Decompressing %d bytes", frame->bytes_recvd);

	rc = Decompress420(	frame->width,
				frame->height,
				frame->data,		/* Output buffer */
				MAX_DATA_SIZE,		/* Output buffer len */
				frame->rawdata,		/* Raw buffer */
				MAX_RAW_DATA_SIZE,	/* Raw buffer len */
				frame->bytes_recvd,	/* Actual raw len */
				frame);

	PDEBUG(4, "DEBUG: Decompress420 returned %ld", rc);

#else

// Enabling this screws up the preprocessor (cpp-2.96-69)
//	warn("Compression was not compiled in!");
//	warn("You must load driver with compress=0 or you will get no image!);

#endif

}

/* Converts frame->data from planar YUV420 to RGB24. This is currently only
 * useful for postprocessing decompressed data, but will eventually replace
 * ov511_parse_yuv420_to_rgb().
 *
 * NOTE: Uses frame->rawdata as a scratch buffer
 */
static void ov511_yuv420p_to_rgb(struct ov511_frame *frame, int bits)
{
	const int numpix = frame->width * frame->height;
	const int bytes = bits >> 3;
	int i, j, y00, y01, y10, y11, u, v;
	unsigned char *pY = frame->rawdata;
	unsigned char *pU = pY + numpix;
	unsigned char *pV = pU + numpix / 4;
	unsigned char *pOut = frame->data;

	/* Move data to a safe place */
	memmove(frame->rawdata, frame->data, MAX_RAW_DATA_SIZE);

//	/* Clear buffer for testing purposes */
//	memset(frame->data, 0, MAX_DATA_SIZE);

	for (j = 0; j <= frame->height - 2; j += 2) {
		for (i = 0; i <= frame->width - 2; i += 2) {
			y00 = *pY;
			y01 = *(pY + 1);
			y10 = *(pY + frame->width);
			y11 = *(pY + frame->width + 1);
			u = (*pU++) - 128;
			v = (*pV++) - 128;

			ov511_move_420_block(y00, y01, y10, y11, u, v,
					     frame->width, pOut, bits);
	
			pY += 2;
			pOut += 2 * bytes;

		}
		pY += frame->width;
		pOut += frame->width * bytes;
	}
}

/* Post-processes the specified frame. This consists of:
 * 	1. Decompress frame (from frame->rawfbuf into frame->data), if necessary
 * 	2. Convert frame->data from YUV420P to destination format, if necessary
 * 	3. Fix the RGB offset, if necessary
 */
static void ov511_postprocess(struct ov511_frame *frame)
{
	if (frame->compressed)
		ov511_decompress(frame);

	switch (frame->format) {
	case VIDEO_PALETTE_RGB565:
		ov511_yuv420p_to_rgb(frame, 16);
		break;
	case VIDEO_PALETTE_RGB24:
		ov511_yuv420p_to_rgb(frame, 24);
		break;
	case VIDEO_PALETTE_YUV420:
	case VIDEO_PALETTE_YUV420P:
	default:
		err("Cannot convert data to this format");
	}

	if (fix_rgb_offset)
		fixFrameRGBoffset(frame);
}

/**********************************************************************
 *
 * OV511 data transfer, IRQ handler
 *
 **********************************************************************/

static int ov511_move_data(struct usb_ov511 *ov511, urb_t *urb)
{
	unsigned char *cdata;
	int i, totlen = 0;
	int aPackNum[10];
	struct ov511_frame *frame;
	unsigned char *pData;
	int iPix, data_size, pn, num, offset;
	struct timeval *ts;

	PDEBUG (4, "Moving %d packets", urb->number_of_packets);

	/* OV518+ has no packet numbering */
	if (ov511->bridge == BRG_OV518PLUS) {
		pn = 0;
		data_size = ov511->packet_size;
	} else {
		pn = 1;
		data_size = ov511->packet_size - 1;
	}

	for (i = 0; i < urb->number_of_packets; i++) {
		int n = urb->iso_frame_desc[i].actual_length;
		int st = urb->iso_frame_desc[i].status;

		urb->iso_frame_desc[i].actual_length = 0;
		urb->iso_frame_desc[i].status = 0;

		cdata = urb->transfer_buffer + urb->iso_frame_desc[i].offset;

		if (pn)
			aPackNum[i] = n ? cdata[ov511->packet_size - 1] : -1;

		if (!n || ov511->curframe == -1)
			continue;

		if (st)
			PDEBUG(2, "data error: [%d] len=%d, status=%d", i, n, st);

		frame = &ov511->frame[ov511->curframe];

		/* SOF/EOF packets have 1st to 8th bytes zeroed and the 9th
		 * byte non-zero. The EOF packet has image width/height in the
		 * 10th and 11th bytes. The 9th byte is given as follows:
		 *
		 * bit 7: EOF
		 *     6: compression enabled
		 *     5: 422/420/400 modes
		 *     4: 422/420/400 modes
		 *     3: 1
		 *     2: snapshot button on
		 *     1: snapshot frame
		 *     0: even/odd field
		 */

#if 0
		{
			int d;
			/* Print all data */
			for (d = data_size; d < data_size; d += 16) {
				info("%4x: %2x %2x %2x %2x %2x %2x %2x %2x %2x %2x %2x %2x %2x %2x %2x %2x", d,
					cdata[d], cdata[d+1], cdata[d+2], cdata[d+3],
					cdata[d+4], cdata[d+5], cdata[d+6], cdata[d+7],
					cdata[d+8], cdata[d+9], cdata[d+10], cdata[d+11],
					cdata[d+12], cdata[d+13], cdata[d+14], cdata[d+15]);
			}
		}
#endif

		if (printph) {
			info("packet header: %x %x %x %x %x %x %x %x %x %x %x %x",
				cdata[0], cdata[1], cdata[2], cdata[3], cdata[4], cdata[5],
				cdata[6], cdata[7], cdata[8], cdata[9], cdata[10], cdata[11]);
		}

		if (ov511->bridge == BRG_OV518PLUS) {
			int trailer = data_size - 9;

			// Could a packet be both SOF and EOF?

			if (compress)
				trailer -= 64;

			if (printph) {
				info("trailer: %x %x %x %x %x %x %x %x", 
					cdata[trailer], cdata[trailer+1], cdata[trailer+2], cdata[trailer+3],
					cdata[trailer+4], cdata[trailer+5], cdata[trailer+6], cdata[trailer+7]);
			}

			/* A false positive here is likely, until OVT gives me
			 * the definitive SOF/EOF format */
			if (!(cdata[0] | cdata[1] | cdata[2] | cdata[3] |
			      cdata[5]) && cdata[6]) {
				goto sof;
			} else if (!(cdata[trailer] | cdata[trailer+1] |
				     cdata[trailer+2] | cdata[trailer+3] |
				     cdata[trailer+5]) && cdata[trailer+6]) {
				goto eof;
			} else {
				goto check_middle;
			}
		} else {
			/* Check for SOF/EOF packet */
			if ((cdata[0] | cdata[1] | cdata[2] | cdata[3] |
			     cdata[4] | cdata[5] | cdata[6] | cdata[7]) ||
			     (~cdata[8] & 0x08))
				goto check_middle;
		}

		/* Frame end */
		if (cdata[8] & 0x80) {
eof:
			ts = (struct timeval *)(frame->data + MAX_FRAME_SIZE);
			do_gettimeofday (ts);

	 		PDEBUG(4, "Frame end, curframe = %d, packnum=%d, hw=%d, vw=%d, recvd=%d",
				ov511->curframe,
				(int)(pn ? cdata[ov511->packet_size - 1] : -1),
				(int)(cdata[9]), (int)(cdata[10]), frame->bytes_recvd);

			if (frame->scanstate == STATE_LINES) {
		    		int iFrameNext;

				if (!ttpp) {
					if (frame->compressed) {
						ov511_decompress(frame);

						switch (frame->format) {
						case VIDEO_PALETTE_RGB565:
							ov511_yuv420p_to_rgb(frame, 16);
							break;
						case VIDEO_PALETTE_RGB24:
							ov511_yuv420p_to_rgb(frame, 24);
							break;
						case VIDEO_PALETTE_YUV420:
						case VIDEO_PALETTE_YUV420P:
							memmove(frame->data, frame->rawdata, MAX_RAW_DATA_SIZE);
							break;
						default:
							err("Cannot convert decompressed data to this format");
						}
					}

					if (fix_rgb_offset)
						fixFrameRGBoffset(frame);
				}

				frame->grabstate = FRAME_DONE;	// FIXME: Is this right?

				if (waitqueue_active(&frame->wq)) {
					frame->grabstate = FRAME_DONE;
					wake_up_interruptible(&frame->wq);
				}

				/* If next frame is ready or grabbing,
                                 * point to it */
				iFrameNext = (ov511->curframe + 1) % OV511_NUMFRAMES;
				if (ov511->frame[iFrameNext].grabstate == FRAME_READY
				    || ov511->frame[iFrameNext].grabstate == FRAME_GRABBING) {
					ov511->curframe = iFrameNext;
					ov511->frame[iFrameNext].scanstate = STATE_SCANNING;
				} else {
					if (frame->grabstate == FRAME_DONE) {
						PDEBUG(4, "Frame done! congratulations");
					} else {
						PDEBUG(4, "Frame not ready? state = %d",
							ov511->frame[iFrameNext].grabstate);
					}

					ov511->curframe = -1;
				}
			}
			/* Image corruption caused by misplaced frame->segment = 0
			 * fixed by carlosf@conectiva.com.br
			 */
		} else {
sof:
			/* Frame start */
			PDEBUG(4, "Frame start, framenum = %d", ov511->curframe);

#if 0
			/* Make sure no previous data carries over; necessary
			 * for compression experimentation */
			memset(frame->data, 0, MAX_DATA_SIZE);
#endif

			/* Check to see if it's a snapshot frame */
			/* FIXME?? Should the snapshot reset go here? Performance? */
			if (cdata[8] & 0x02) {
				frame->snapshot = 1;
				PDEBUG(3, "snapshot detected");
			}

			frame->scanstate = STATE_LINES;
			frame->segment = 0;
			frame->bytes_recvd = 0;
			frame->output_offset = 0;
			frame->compressed = cdata[8] & 0x40;
		}

check_middle:
		/* Are we in a frame? */
		if (frame->scanstate != STATE_LINES)
			continue;

		/* If frame start, skip header */
		if (frame->bytes_recvd == 0)
			offset = 9;
		else
			offset = 0;

		num = n - offset - (pn ? 1 : 0);
		frame->bytes_recvd += num;

		if (frame->compressed) {
			if (frame->bytes_recvd <= MAX_RAW_DATA_SIZE)
				memmove(frame->rawdata + frame->bytes_recvd - num,
					&cdata[offset], num);
			else
				err("Raw data buffer overrun!!");

			continue; /* Skip all other processing */
		}

		/* Deal with leftover from last segment, if any */
		if (frame->segment) {
			pData = ov511->scratch;
			iPix = -ov511->scratchlen;
			memmove(pData + ov511->scratchlen, cdata,
				iPix+frame->segsize);
		} else {
			pData = &cdata[iPix = 9]; /* Start of frame */
	 	}

		/* Parse the segments */
		while (iPix <= data_size - frame->segsize &&
		    frame->segment < frame->width * frame->height / PIXELS_PER_SEG) {
			int iSegY, iSegUV;
			int iY, jY, iUV, jUV;
			int iOutY, iOutYP, iOutUV, iOutUVP;
			int format;
			unsigned char *pOut;

			iSegY = iSegUV = frame->segment;
			pOut = frame->data;
			frame->segment++;
			iPix += frame->segsize;

			/* Handle subwindow */
			if (frame->sub_flag) {
				int iSeg1;

				iSeg1 = iSegY / (ov511->subw / 32);
				iSeg1 *= frame->width / 32;
				iSegY = iSeg1 + (iSegY % (ov511->subw / 32));
				if (iSegY >= frame->width * ov511->subh / PIXELS_PER_SEG)
					break;

				iSeg1 = iSegUV / (ov511->subw / 16);
				iSeg1 *= frame->width / 16;
				iSegUV = iSeg1 + (iSegUV % (ov511->subw / 16));

				pOut += (ov511->subx + ov511->suby * frame->width) *
					(frame->depth >> 3);
			}

			/* 
			 * i counts segment lines
			 * j counts segment columns
			 * iOut is the offset (in bytes) of the upper left corner
			 */
			iY = iSegY / (frame->width / WDIV);
			jY = iSegY - iY * (frame->width / WDIV);
			iOutYP = iY*HDIV*frame->width + jY*WDIV;
			iOutY = iOutYP * (frame->depth >> 3);
			iUV = iSegUV / (frame->width / WDIV * 2);
			jUV = iSegUV - iUV * (frame->width / WDIV * 2);
			iOutUVP = iUV*HDIV*2*frame->width + jUV*WDIV/2;
			iOutUV = iOutUVP * (frame->depth >> 3);

			/* If ttpp is set just decode the frame to YUV420P, and
			 * post-process it at task-time. Else, do it the old way. */
			if (ttpp)
				format = VIDEO_PALETTE_YUV420P;
			else
				format = frame->format;

			switch (format) {
			case VIDEO_PALETTE_GREY:
				ov511_parse_data_grey (pData, pOut, iOutY, frame->width);
				break;
			case VIDEO_PALETTE_RGB24:
				if (dumppix)
					ov511_dumppix(frame, pData, pOut, iOutY, iOutUV,
						iY & 1, frame->width);
				else if (sensor_gbr)
					ov511_parse_gbr422_to_rgb24(pData, pOut, iOutY, iOutUV,
						iY & 1, frame->width);
				else
					ov511_parse_yuv420_to_rgb(pData, pOut, iOutY, iOutUV,
						iY & 1, frame->width, 24);
				break;
			case VIDEO_PALETTE_RGB565:
				ov511_parse_yuv420_to_rgb(pData, pOut, iOutY, iOutUV,
					iY & 1, frame->width, 16);
				break;
			case VIDEO_PALETTE_YUV422:
			case VIDEO_PALETTE_YUYV:
				ov511_parse_data_yuv422(pData, pOut, iOutY, iOutUV, frame->width);
				break;
			case VIDEO_PALETTE_YUV420:
			case VIDEO_PALETTE_YUV420P:
				ov511_parse_data_yuv420 (pData, pOut, iOutYP, iUV*HDIV*frame->width/2 + jUV*WDIV/4,
					frame->width, frame->height);
				break;
			case VIDEO_PALETTE_YUV422P:
				ov511_parse_data_yuv422p (pData, pOut, iOutYP, iOutUVP/2,
					frame->width, frame->height);
				break;
			default:
				err("Unsupported format: %d", frame->format);
			}

			pData = &cdata[iPix];
		}

		/* Save extra data for next time */
		if (frame->segment < frame->width * frame->height / PIXELS_PER_SEG) {
			ov511->scratchlen = data_size - iPix;
			if (ov511->scratchlen < frame->segsize)
				memmove(ov511->scratch, pData, ov511->scratchlen);
			else
				ov511->scratchlen = 0;
		}
	}

	if (pn) {
		PDEBUG(5, "pn: %d %d %d %d %d %d %d %d %d %d",
			aPackNum[0], aPackNum[1], aPackNum[2], aPackNum[3], aPackNum[4],
			aPackNum[5],aPackNum[6], aPackNum[7], aPackNum[8], aPackNum[9]);
	}

	return totlen;
}

static void ov511_isoc_irq(struct urb *urb)
{
	int len;
	struct usb_ov511 *ov511;
	struct ov511_sbuf *sbuf;

	if (!urb->context) {
		PDEBUG(4, "no context");
		return;
	}

	ov511 = (struct usb_ov511 *) urb->context;

	if (!ov511->dev || !ov511->user) {
		PDEBUG(4, "no device, or not open");
		return;
	}

	if (!ov511->streaming) {
		PDEBUG(4, "hmmm... not streaming, but got interrupt");
		return;
	}

	sbuf = &ov511->sbuf[ov511->cursbuf];

	if (sbuf == NULL) {
		err("null scrath buffer pointer!");
		return;
	}

	/* Copy the data received into our scratch buffer */
	if (ov511->curframe >= 0) {
		len = ov511_move_data(ov511, urb);
	} else if (waitqueue_active(&ov511->wq)) {
		wake_up_interruptible(&ov511->wq);
	}

	/* Move to the next sbuf */
	ov511->cursbuf = (ov511->cursbuf + 1) % OV511_NUMSBUF;

	urb->dev = ov511->dev;

	return;
}

static int ov511_init_isoc(struct usb_ov511 *ov511)
{
	urb_t *urb;
	int fx, err, n, size;

	PDEBUG(3, "*** Initializing capture ***");

	ov511->curframe = -1;
	ov511->cursbuf = 0;
	ov511->scratchlen = 0;

	if (ov511->bridge == BRG_OV511) {
		if (cams == 1)				size = 993;
		else if (cams == 2)			size = 513;
		else if (cams == 3 || cams == 4)	size = 257;
		else {
			err("\"cams\" parameter too high!");
			return -1;
		}
	} else if (ov511->bridge == BRG_OV511PLUS) {
		if (cams == 1)				size = 961;
		else if (cams == 2)			size = 513;
		else if (cams == 3 || cams == 4)	size = 257;
		else if (cams >= 5 && cams <= 8)	size = 129;
		else if (cams >= 9 && cams <= 31)	size = 33;
		else {
			err("\"cams\" parameter too high!");
			return -1;
		}
	} else if (ov511->bridge == BRG_OV518PLUS) {
		if (cams == 1)				size = 896;
		else if (cams == 2)			size = 512;
		else if (cams == 3 || cams == 4)	size = 256;
		else if (cams >= 5 && cams <= 8)	size = 128;
		else {
			err("\"cams\" parameter too high!");
			return -1;
		}
	} else {
		err("invalid bridge type");
		return -1;
	}

	// FIXME: OV518 is hardcoded to 15 FPS (alternate 5) for now
	if (ov511->bridge == BRG_OV518PLUS)
		ov511_set_packet_size(ov511, 640);
	else
		ov511_set_packet_size(ov511, size);

	for (n = 0; n < OV511_NUMSBUF; n++) {
		urb = usb_alloc_urb(FRAMES_PER_DESC);
	
		if (!urb) {
			err("init isoc: usb_alloc_urb ret. NULL");
			return -ENOMEM;
		}
		ov511->sbuf[n].urb = urb;
		urb->dev = ov511->dev;
		urb->context = ov511;
		urb->pipe = usb_rcvisocpipe(ov511->dev, OV511_ENDPOINT_ADDRESS);
		urb->transfer_flags = USB_ISO_ASAP;
		urb->transfer_buffer = ov511->sbuf[n].data;
		urb->complete = ov511_isoc_irq;
		urb->number_of_packets = FRAMES_PER_DESC;
		urb->transfer_buffer_length = ov511->packet_size * FRAMES_PER_DESC;
		for (fx = 0; fx < FRAMES_PER_DESC; fx++) {
			urb->iso_frame_desc[fx].offset = ov511->packet_size * fx;
			urb->iso_frame_desc[fx].length = ov511->packet_size;
		}
	}

	ov511->sbuf[OV511_NUMSBUF - 1].urb->next = ov511->sbuf[0].urb;
	for (n = 0; n < OV511_NUMSBUF - 1; n++)
		ov511->sbuf[n].urb->next = ov511->sbuf[n+1].urb;

	for (n = 0; n < OV511_NUMSBUF; n++) {
		ov511->sbuf[n].urb->dev = ov511->dev;
		err = usb_submit_urb(ov511->sbuf[n].urb);
		if (err)
			err("init isoc: usb_submit_urb(%d) ret %d", n, err);
	}

	ov511->streaming = 1;

	return 0;
}

static void ov511_stop_isoc(struct usb_ov511 *ov511)
{
	int n;

	if (!ov511->streaming || !ov511->dev)
		return;

	PDEBUG (3, "*** Stopping capture ***");

	ov511_set_packet_size(ov511, 0);

	ov511->streaming = 0;

	/* Unschedule all of the iso td's */
	for (n = OV511_NUMSBUF - 1; n >= 0; n--) {
		if (ov511->sbuf[n].urb) {
			ov511->sbuf[n].urb->next = NULL;
			usb_unlink_urb(ov511->sbuf[n].urb);
			usb_free_urb(ov511->sbuf[n].urb);
			ov511->sbuf[n].urb = NULL;
		}
	}
}

static int ov511_new_frame(struct usb_ov511 *ov511, int framenum)
{
	struct ov511_frame *frame;

	PDEBUG(4, "ov511->curframe = %d, framenum = %d", ov511->curframe,
		framenum);
	if (!ov511->dev)
		return -1;

	/* If we're not grabbing a frame right now and the other frame is */
	/* ready to be grabbed into, then use it instead */
	if (ov511->curframe == -1) {
		if (ov511->frame[(framenum - 1 + OV511_NUMFRAMES) % OV511_NUMFRAMES].grabstate == FRAME_READY)
			framenum = (framenum - 1 + OV511_NUMFRAMES) % OV511_NUMFRAMES;
	} else
		return 0;

	frame = &ov511->frame[framenum];

	PDEBUG (4, "framenum = %d, width = %d, height = %d", framenum, 
		frame->width, frame->height);

	frame->grabstate = FRAME_GRABBING;
	frame->scanstate = STATE_SCANNING;
	frame->scanlength = 0;		/* accumulated in ov511_parse_data() */
	frame->snapshot = 0;

	ov511->curframe = framenum;

	/* Make sure it's not too big */
	if (frame->width > ov511->maxwidth)
		frame->width = ov511->maxwidth;

	frame->width &= ~7L;		/* Multiple of 8 */

	if (frame->height > ov511->maxheight)
		frame->height = ov511->maxheight;

	frame->height &= ~3L;		/* Multiple of 4 */

	return 0;
}

/****************************************************************************
 *
 * Buffer management
 *
 ***************************************************************************/
static int ov511_alloc(struct usb_ov511 *ov511)
{
	int i;

	PDEBUG(4, "entered");
	down(&ov511->buf_lock);

	if (ov511->buf_state == BUF_PEND_DEALLOC) {
		ov511->buf_state = BUF_ALLOCATED;
		del_timer(&ov511->buf_timer);
	}

	if (ov511->buf_state == BUF_ALLOCATED)
		goto out;

	ov511->fbuf = rvmalloc(OV511_NUMFRAMES * MAX_DATA_SIZE);
	if (!ov511->fbuf)
		goto error;

	ov511->rawfbuf = vmalloc(OV511_NUMFRAMES * MAX_RAW_DATA_SIZE);
	if (!ov511->rawfbuf) {
		rvfree(ov511->fbuf, OV511_NUMFRAMES * MAX_DATA_SIZE);
		ov511->fbuf = NULL;
		goto error;
	}

	for (i = 0; i < OV511_NUMFRAMES; i++) {
		ov511->frame[i].grabstate = FRAME_UNUSED;
		ov511->frame[i].data = ov511->fbuf + i * MAX_DATA_SIZE;
		ov511->frame[i].rawdata = ov511->rawfbuf + i * MAX_RAW_DATA_SIZE;
		PDEBUG(4, "frame[%d] @ %p", i, ov511->frame[i].data);

		ov511->sbuf[i].data = kmalloc(FRAMES_PER_DESC *
			MAX_FRAME_SIZE_PER_DESC, GFP_KERNEL);
		if (!ov511->sbuf[i].data) {
			while (--i) {
				kfree(ov511->sbuf[i].data);
				ov511->sbuf[i].data = NULL;
				ov511->frame[i].data = NULL;
				ov511->frame[i].rawdata = NULL;
			}
			vfree(ov511->rawfbuf);
			ov511->rawfbuf = NULL;
			rvfree(ov511->fbuf, OV511_NUMFRAMES * MAX_DATA_SIZE);
			ov511->fbuf = NULL;
			goto error;
		}
		PDEBUG(4, "sbuf[%d] @ %p", i, ov511->sbuf[i].data);
	}
	ov511->buf_state = BUF_ALLOCATED;
out:
	up(&ov511->buf_lock);
	PDEBUG(4, "leaving");
	return 0;
error:
	ov511->buf_state = BUF_NOT_ALLOCATED;
	up(&ov511->buf_lock);
	PDEBUG(4, "errored");
	return -ENOMEM;
}

/* 
 * - You must acquire buf_lock before entering this function.
 * - Because this code will free any non-null pointer, you must be sure to null
 *   them if you explicitly free them somewhere else!
 */
static void ov511_do_dealloc(struct usb_ov511 *ov511)
{
	int i;
	PDEBUG(4, "entered");

	if (ov511->fbuf) {
		rvfree(ov511->fbuf, OV511_NUMFRAMES * MAX_DATA_SIZE);
		ov511->fbuf = NULL;
	}

	if (ov511->rawfbuf) {
		vfree(ov511->rawfbuf);
		ov511->rawfbuf = NULL;
	}

	for (i = 0; i < OV511_NUMFRAMES; i++) {
		if (ov511->sbuf[i].data) {
			kfree(ov511->sbuf[i].data);
			ov511->sbuf[i].data = NULL;
		}
		ov511->frame[i].data = NULL;
		ov511->frame[i].rawdata = NULL;

		/* Allocated by decompressor */
		FREE(ov511->frame[i].m_pTempBuffer);
		ov511->frame[i].m_pTempBuffer = NULL;
	}

	PDEBUG(4, "buffer memory deallocated");
	ov511->buf_state = BUF_NOT_ALLOCATED;
	PDEBUG(4, "leaving");
}

static void ov511_buf_callback(unsigned long data)
{
	struct usb_ov511 *ov511 = (struct usb_ov511 *)data;
	PDEBUG(4, "entered");
	down(&ov511->buf_lock);

	if (ov511->buf_state == BUF_PEND_DEALLOC)
		ov511_do_dealloc(ov511);

	up(&ov511->buf_lock);
	PDEBUG(4, "leaving");
}

static void ov511_dealloc(struct usb_ov511 *ov511, int now)
{
	struct timer_list *bt = &(ov511->buf_timer);
	PDEBUG(4, "entered");
	down(&ov511->buf_lock);

	PDEBUG(4, "deallocating buffer memory %s", now ? "now" : "later");

	if (ov511->buf_state == BUF_PEND_DEALLOC) {
		ov511->buf_state = BUF_ALLOCATED;
		del_timer(bt);
	}

	if (now)
		ov511_do_dealloc(ov511);
	else {
		ov511->buf_state = BUF_PEND_DEALLOC;
		init_timer(bt);
		bt->function = ov511_buf_callback;
		bt->data = (unsigned long)ov511;
		bt->expires = jiffies + buf_timeout * HZ;
		add_timer(bt);
	}
	up(&ov511->buf_lock);
	PDEBUG(4, "leaving");
}

/****************************************************************************
 *
 * V4L API
 *
 ***************************************************************************/

static int ov511_open(struct video_device *dev, int flags)
{
	struct usb_ov511 *ov511 = (struct usb_ov511 *)dev;
	int err;

	MOD_INC_USE_COUNT;
	PDEBUG(4, "opening");
	down(&ov511->lock);

	err = -EBUSY;
	if (ov511->user) 
		goto out;

	err = -ENOMEM;
	if (ov511_alloc(ov511))
		goto out;

	ov511->sub_flag = 0;

	err = ov511_init_isoc(ov511);
	if (err) {
		ov511_dealloc(ov511, 0);
		goto out;
	}

	ov511->user++;
	
	if (ov511->led_policy == LED_AUTO)
		ov511_led_control(ov511, 1);

out:
	up(&ov511->lock);

	if (err)
		MOD_DEC_USE_COUNT;

	return err;
}

static void ov511_close(struct video_device *dev)
{
	struct usb_ov511 *ov511 = (struct usb_ov511 *)dev;

	PDEBUG(4, "ov511_close");
	
	down(&ov511->lock);

	ov511->user--;
	ov511_stop_isoc(ov511);

	if (ov511->led_policy == LED_AUTO)
		ov511_led_control(ov511, 0);

	if (ov511->dev)
		ov511_dealloc(ov511, 0);

	up(&ov511->lock);

	if (!ov511->dev) {
		ov511_dealloc(ov511, 1);
		video_unregister_device(&ov511->vdev);
		kfree(ov511);
		ov511 = NULL;
	}

	MOD_DEC_USE_COUNT;
}

static int ov511_init_done(struct video_device *dev)
{
#if defined(CONFIG_PROC_FS) && defined(CONFIG_VIDEO_PROC_FS)
	create_proc_ov511_cam((struct usb_ov511 *)dev);
#endif

	return 0;
}

static long ov511_write(struct video_device *dev, const char *buf, unsigned long count, int noblock)
{
	return -EINVAL;
}

static int ov511_ioctl(struct video_device *vdev, unsigned int cmd, void *arg)
{
	struct usb_ov511 *ov511 = (struct usb_ov511 *)vdev;

	PDEBUG(4, "IOCtl: 0x%X", cmd);

	if (!ov511->dev)
		return -EIO;	

	switch (cmd) {
	case VIDIOCGCAP:
	{
		struct video_capability b;

		PDEBUG (4, "VIDIOCGCAP");

		strcpy(b.name, "OV511 USB Camera");
		b.type = VID_TYPE_CAPTURE | VID_TYPE_SUBCAPTURE;
		b.channels = 1;
		b.audios = 0;
		b.maxwidth = ov511->maxwidth;
		b.maxheight = ov511->maxheight;
		b.minwidth = 160;
		b.minheight = 120;

		if (copy_to_user(arg, &b, sizeof(b)))
			return -EFAULT;
				
		return 0;
	}
	case VIDIOCGCHAN:
	{
		struct video_channel v;

		if (copy_from_user(&v, arg, sizeof(v)))
			return -EFAULT;
		if (v.channel != 0)
			return -EINVAL;

		v.flags = 0;
		v.tuners = 0;
		v.type = VIDEO_TYPE_CAMERA;
		v.norm = 0;
		strcpy(v.name, "Camera");

		if (copy_to_user(arg, &v, sizeof(v)))
			return -EFAULT;
				
		return 0;
	}
	case VIDIOCSCHAN:
	{
		int v;

		if (copy_from_user(&v, arg, sizeof(v)))
			return -EFAULT;

		if (v != 0)
			return -EINVAL;

		return 0;
	}
	case VIDIOCGPICT:
	{
		struct video_picture p;

		PDEBUG (4, "VIDIOCGPICT");

		if (ov7610_get_picture(ov511, &p))
			return -EIO;
							
		if (copy_to_user(arg, &p, sizeof(p)))
			return -EFAULT;

		return 0;
	}
	case VIDIOCSPICT:
	{
		struct video_picture p;
		int i;

		PDEBUG (4, "VIDIOCSPICT");

		if (copy_from_user(&p, arg, sizeof(p)))
			return -EFAULT;

		if (!ov511_get_depth(p.palette))
			return -EINVAL;
			
		if (ov7610_set_picture(ov511, &p))
			return -EIO;

		PDEBUG(4, "Setting depth=%d, palette=%d", p.depth, p.palette);
		for (i = 0; i < OV511_NUMFRAMES; i++) {
			ov511->frame[i].depth = p.depth;
			ov511->frame[i].format = p.palette;
			ov511->frame[i].segsize = GET_SEGSIZE(p.palette);
		}

		return 0;
	}
	case VIDIOCGCAPTURE:
	{
		int vf;

		PDEBUG (4, "VIDIOCGCAPTURE");

		if (copy_from_user(&vf, arg, sizeof(vf)))
			return -EFAULT;
		ov511->sub_flag = vf;
		return 0;
	}
	case VIDIOCSCAPTURE:
	{
		struct video_capture vc;

		if (copy_from_user(&vc, arg, sizeof(vc)))
			return -EFAULT;
		if (vc.flags)
			return -EINVAL;
		if (vc.decimation)
			return -EINVAL;

		vc.x &= ~3L;
		vc.y &= ~1L;
		vc.y &= ~31L;

		if (vc.width == 0)
			vc.width = 32;

		vc.height /= 16;
		vc.height *= 16;
		if (vc.height == 0)
			vc.height = 16;

		ov511->subx = vc.x;
		ov511->suby = vc.y;
		ov511->subw = vc.width;
		ov511->subh = vc.height;

		return 0;
	}
	case VIDIOCSWIN:
	{
		struct video_window vw;
		int i, result;

		if (copy_from_user(&vw, arg, sizeof(vw)))
			return -EFAULT;

		PDEBUG (4, "VIDIOCSWIN: width=%d, height=%d",
			vw.width, vw.height);

#if 0
		if (vw.flags)
			return -EINVAL;
		if (vw.clipcount)
			return -EINVAL;
		if (vw.height != ov511->maxheight)
			return -EINVAL;
		if (vw.width != ov511->maxwidth)
			return -EINVAL;
#endif

		/* If we're collecting previous frame wait
		   before changing modes */
		interruptible_sleep_on(&ov511->wq);
		if (signal_pending(current)) return -EINTR;

		result = mode_init_regs(ov511, vw.width, vw.height,
			ov511->frame[0].format, ov511->sub_flag);
		if (result < 0)
			return result;

		for (i = 0; i < OV511_NUMFRAMES; i++) {
			ov511->frame[i].width = vw.width;
			ov511->frame[i].height = vw.height;
		}

		return 0;
	}
	case VIDIOCGWIN:
	{
		struct video_window vw;

		vw.x = 0;		/* FIXME */
		vw.y = 0;
		vw.width = ov511->frame[0].width;
		vw.height = ov511->frame[0].height;
		vw.chromakey = 0;
		vw.flags = 30;

		PDEBUG (4, "VIDIOCGWIN: %dx%d", vw.width, vw.height);

		if (copy_to_user(arg, &vw, sizeof(vw)))
			return -EFAULT;

		return 0;
	}
	case VIDIOCGMBUF:
	{
		struct video_mbuf vm;

		memset(&vm, 0, sizeof(vm));
		vm.size = OV511_NUMFRAMES * MAX_DATA_SIZE;
		vm.frames = OV511_NUMFRAMES;
		vm.offsets[0] = 0;
		vm.offsets[1] = MAX_FRAME_SIZE + sizeof (struct timeval);

		if (copy_to_user((void *)arg, (void *)&vm, sizeof(vm)))
			return -EFAULT;

		return 0;
	}
	case VIDIOCMCAPTURE:
	{
		struct video_mmap vm;
		int ret, depth;

		if (copy_from_user((void *)&vm, (void *)arg, sizeof(vm)))
			return -EFAULT;

		PDEBUG(4, "CMCAPTURE");
		PDEBUG(4, "frame: %d, size: %dx%d, format: %d",
			vm.frame, vm.width, vm.height, vm.format);

		depth = ov511_get_depth(vm.format);
		if (!depth) {
			err("VIDIOCMCAPTURE: invalid format (%d)", vm.format);
			return -EINVAL;
		}

		if ((vm.frame != 0) && (vm.frame != 1)) {
			err("VIDIOCMCAPTURE: invalid frame (%d)", vm.frame);
			return -EINVAL;
		}

		if (vm.width > ov511->maxwidth || vm.height > ov511->maxheight) {
			err("VIDIOCMCAPTURE: requested dimensions too big");
			return -EINVAL;
		}

		if (ov511->frame[vm.frame].grabstate == FRAME_GRABBING)
			return -EBUSY;

		if ((ov511->frame[vm.frame].width != vm.width) ||
		    (ov511->frame[vm.frame].height != vm.height) ||
		    (ov511->frame[vm.frame].format != vm.format) ||
		    (ov511->frame[vm.frame].sub_flag !=
		     ov511->sub_flag)) {
			/* If we're collecting previous frame wait
			   before changing modes */
			interruptible_sleep_on(&ov511->wq);
			if (signal_pending(current)) return -EINTR;
			ret = mode_init_regs(ov511, vm.width, vm.height,
				vm.format, ov511->sub_flag);
#if 0
			if (ret < 0)
				return ret;
#endif
		}

		ov511->frame[vm.frame].width = vm.width;
		ov511->frame[vm.frame].height = vm.height;
		ov511->frame[vm.frame].format = vm.format;
		ov511->frame[vm.frame].sub_flag = ov511->sub_flag;
		ov511->frame[vm.frame].segsize = GET_SEGSIZE(vm.format);
		ov511->frame[vm.frame].depth = depth;

		/* Mark it as ready */
		ov511->frame[vm.frame].grabstate = FRAME_READY;

		return ov511_new_frame(ov511, vm.frame);
	}
	case VIDIOCSYNC:
	{
		int frame;

		if (copy_from_user((void *)&frame, arg, sizeof(int)))
			return -EFAULT;

		PDEBUG(4, "syncing to frame %d, grabstate = %d", frame,
		       ov511->frame[frame].grabstate);

		switch (ov511->frame[frame].grabstate) {
		case FRAME_UNUSED:
			return -EINVAL;
		case FRAME_READY:
		case FRAME_GRABBING:
		case FRAME_ERROR:
redo:
			if (!ov511->dev)
				return -EIO;

			do {
#if 0
				init_waitqueue_head(&ov511->frame[frame].wq);
#endif
				interruptible_sleep_on(&ov511->frame[frame].wq);
				if (signal_pending(current)) {
					if (retry_sync) {
						PDEBUG(3, "***retry sync***");

						/* Polling apps will destroy frames with that! */
						ov511_new_frame(ov511, frame);
						ov511->curframe = -1;

						/* This will request another frame. */
						if (waitqueue_active(&ov511->frame[frame].wq))
							wake_up_interruptible(&ov511->frame[frame].wq);

						return 0;
 					} else {
						return -EINTR;
					}
				}
			} while (ov511->frame[frame].grabstate == FRAME_GRABBING);

			if (ov511->frame[frame].grabstate == FRAME_ERROR) {
				int ret;

				if ((ret = ov511_new_frame(ov511, frame)) < 0)
					return ret;
				goto redo;
			}			
		case FRAME_DONE:
			if (ov511->snap_enabled && !ov511->frame[frame].snapshot) {
				int ret;
				if ((ret = ov511_new_frame(ov511, frame)) < 0)
					return ret;
				goto redo;
			}

			ov511->frame[frame].grabstate = FRAME_UNUSED;

			/* Reset the hardware snapshot button */
			/* FIXME - Is this the best place for this? */
			if ((ov511->snap_enabled) &&
			    (ov511->frame[frame].snapshot)) {
				ov511->frame[frame].snapshot = 0;
				ov51x_clear_snapshot(ov511);
			}

			if (ttpp)
				ov511_postprocess(&(ov511->frame[frame]));

			break;
		} /* end switch */

		return 0;
	}
	case VIDIOCGFBUF:
	{
		struct video_buffer vb;

		memset(&vb, 0, sizeof(vb));
		vb.base = NULL;	/* frame buffer not supported, not used */

		if (copy_to_user((void *)arg, (void *)&vb, sizeof(vb)))
			return -EFAULT;

		return 0;
	}
	case VIDIOCKEY:
		return 0;
	case VIDIOCCAPTURE:
		return -EINVAL;
	case VIDIOCSFBUF:
		return -EINVAL;
	case VIDIOCGTUNER:
	case VIDIOCSTUNER:
		return -EINVAL;
	case VIDIOCGFREQ:
	case VIDIOCSFREQ:
		return -EINVAL;
	case VIDIOCGAUDIO:
	case VIDIOCSAUDIO:
		return -EINVAL;
	default:
		return -ENOIOCTLCMD;
	} /* end switch */

	return 0;
}

static inline long ov511_read(struct video_device *dev, char *buf, unsigned long count, int noblock)
{
	struct usb_ov511 *ov511 = (struct usb_ov511 *)dev;
	int i;
	int frmx = -1;
	volatile struct ov511_frame *frame;

	PDEBUG(4, "%ld bytes, noblock=%d", count, noblock);

	if (!dev || !buf)
		return -EFAULT;

	if (!ov511->dev)
		return -EIO;

	/* See if a frame is completed, then use it. */
	if (ov511->frame[0].grabstate >= FRAME_DONE)	/* _DONE or _ERROR */
		frmx = 0;
	else if (ov511->frame[1].grabstate >= FRAME_DONE)/* _DONE or _ERROR */
		frmx = 1;

	/* If nonblocking we return immediately */
	if (noblock && (frmx == -1))
		return -EAGAIN;

	/* If no FRAME_DONE, look for a FRAME_GRABBING state. */
	/* See if a frame is in process (grabbing), then use it. */
	if (frmx == -1) {
		if (ov511->frame[0].grabstate == FRAME_GRABBING)
			frmx = 0;
		else if (ov511->frame[1].grabstate == FRAME_GRABBING)
			frmx = 1;
	}

	/* If no frame is active, start one. */
	if (frmx == -1)
		ov511_new_frame(ov511, frmx = 0);

	frame = &ov511->frame[frmx];

restart:
	if (!ov511->dev)
		return -EIO;

	/* Wait while we're grabbing the image */
	PDEBUG(4, "Waiting image grabbing");
	while (frame->grabstate == FRAME_GRABBING) {
		interruptible_sleep_on(&ov511->frame[frmx].wq);
		if (signal_pending(current))
			return -EINTR;
	}
	PDEBUG(4, "Got image, frame->grabstate = %d", frame->grabstate);

	if (frame->grabstate == FRAME_ERROR) {
		frame->bytes_read = 0;
		err("** ick! ** Errored frame %d", ov511->curframe);
		if (ov511_new_frame(ov511, frmx))
			err("read: ov511_new_frame error");
		goto restart;
	}


	/* Repeat until we get a snapshot frame */
	if (ov511->snap_enabled)
		PDEBUG (4, "Waiting snapshot frame");
	if (ov511->snap_enabled && !frame->snapshot) {
		frame->bytes_read = 0;
		if (ov511_new_frame(ov511, frmx))
			err("ov511_new_frame error");
		goto restart;
	}

	/* Clear the snapshot */
	if (ov511->snap_enabled && frame->snapshot) {
		frame->snapshot = 0;
		ov51x_clear_snapshot(ov511);
	}

	if (ttpp)
		ov511_postprocess((struct ov511_frame *) frame);

	PDEBUG(4, "frmx=%d, bytes_read=%ld, scanlength=%ld", frmx,
		frame->bytes_read, frame->scanlength);

	/* copy bytes to user space; we allow for partials reads */
//	if ((count + frame->bytes_read) > frame->scanlength)
//		count = frame->scanlength - frame->bytes_read;

	/* FIXME - count hardwired to be one frame... */
	count = frame->width * frame->height * (frame->depth >> 3);

	PDEBUG(4, "Copy to user space: %ld bytes", count);
	if ((i = copy_to_user(buf, frame->data + frame->bytes_read, count))) {
		PDEBUG(4, "Copy failed! %d bytes not copied", i);
		return -EFAULT;
	}

	frame->bytes_read += count;
	PDEBUG(4, "{copy} count used=%ld, new bytes_read=%ld",
		count, frame->bytes_read);

	if (frame->bytes_read >= frame->scanlength) { /* All data has been read */
		frame->bytes_read = 0;

		/* Mark it as available to be used again. */
		ov511->frame[frmx].grabstate = FRAME_UNUSED;
		if (ov511_new_frame(ov511, !frmx))
			err("ov511_new_frame returned error");
	}

	PDEBUG(4, "read finished, returning %ld (sweet)", count);

	return count;
}

static int ov511_mmap(struct video_device *dev, const char *adr,
	unsigned long size)
{
	struct usb_ov511 *ov511 = (struct usb_ov511 *)dev;
	unsigned long start = (unsigned long)adr;
	unsigned long page, pos;

	if (ov511->dev == NULL)
		return -EIO;

	PDEBUG(4, "mmap: %ld (%lX) bytes", size, size);

	if (size > (((OV511_NUMFRAMES * MAX_DATA_SIZE) + PAGE_SIZE - 1) & ~(PAGE_SIZE - 1)))
		return -EINVAL;

	pos = (unsigned long)ov511->fbuf;
	while (size > 0) {
		page = kvirt_to_pa(pos);
		if (remap_page_range(start, page, PAGE_SIZE, PAGE_SHARED))
			return -EAGAIN;
		start += PAGE_SIZE;
		pos += PAGE_SIZE;
		if (size > PAGE_SIZE)
			size -= PAGE_SIZE;
		else
			size = 0;
	}

	return 0;
}

static struct video_device ov511_template = {
	name:		"OV511 USB Camera",
	type:		VID_TYPE_CAPTURE,
	hardware:	VID_HARDWARE_OV511,
	open:		ov511_open,
	close:		ov511_close,
	read:		ov511_read,
	write:		ov511_write,
	ioctl:		ov511_ioctl,
	mmap:		ov511_mmap,
	initialize:	ov511_init_done,
};

/****************************************************************************
 *
 * OV511 and sensor configuration
 *
 ***************************************************************************/

/* This initializes the OV7610, OV7620, or OV7620AE sensor. The OV7620AE uses
 * the same register settings as the OV7610, since they are very similar.
 */
static int ov7xx0_configure(struct usb_ov511 *ov511)
{
	struct usb_device *dev = ov511->dev;
	int i, success;
	int rc;

	/* Lawrence Glaister <lg@jfm.bc.ca> reports:
	 *
	 * Register 0x0f in the 7610 has the following effects:
	 *
	 * 0x85 (AEC method 1): Best overall, good contrast range
	 * 0x45 (AEC method 2): Very overexposed
	 * 0xa5 (spec sheet default): Ok, but the black level is
	 *	shifted resulting in loss of contrast
	 * 0x05 (old driver setting): very overexposed, too much
	 *	contrast
	 */
	static struct ov511_regvals aRegvalsNorm7610[] = {
		{ OV511_I2C_BUS, 0x10, 0xff },
		{ OV511_I2C_BUS, 0x16, 0x06 },
		{ OV511_I2C_BUS, 0x28, 0x24 },
		{ OV511_I2C_BUS, 0x2b, 0xac },
		{ OV511_I2C_BUS, 0x12, 0x00 },
		{ OV511_I2C_BUS, 0x38, 0x81 },
		{ OV511_I2C_BUS, 0x28, 0x24 },	/* 0c */
		{ OV511_I2C_BUS, 0x0f, 0x85 },	/* lg's setting */
		{ OV511_I2C_BUS, 0x15, 0x01 },
		{ OV511_I2C_BUS, 0x20, 0x1c },
		{ OV511_I2C_BUS, 0x23, 0x2a },
		{ OV511_I2C_BUS, 0x24, 0x10 },
		{ OV511_I2C_BUS, 0x25, 0x8a },
		{ OV511_I2C_BUS, 0x27, 0xc2 },
		{ OV511_I2C_BUS, 0x2a, 0x04 },
		{ OV511_I2C_BUS, 0x2c, 0xfe },
		{ OV511_I2C_BUS, 0x30, 0x71 },
		{ OV511_I2C_BUS, 0x31, 0x60 },
		{ OV511_I2C_BUS, 0x32, 0x26 },
		{ OV511_I2C_BUS, 0x33, 0x20 },
		{ OV511_I2C_BUS, 0x34, 0x48 },
		{ OV511_I2C_BUS, 0x12, 0x24 },
		{ OV511_I2C_BUS, 0x11, 0x01 },
		{ OV511_I2C_BUS, 0x0c, 0x24 },
		{ OV511_I2C_BUS, 0x0d, 0x24 },
		{ OV511_DONE_BUS, 0x0, 0x00 },
	};

	static struct ov511_regvals aRegvalsNorm7620[] = {
		{ OV511_I2C_BUS, 0x10, 0xff },
		{ OV511_I2C_BUS, 0x16, 0x06 },
		{ OV511_I2C_BUS, 0x28, 0x24 },
		{ OV511_I2C_BUS, 0x2b, 0xac },
		{ OV511_I2C_BUS, 0x12, 0x00 },
		{ OV511_I2C_BUS, 0x28, 0x24 },
		{ OV511_I2C_BUS, 0x0f, 0x85 },	/* lg's setting */
		{ OV511_I2C_BUS, 0x15, 0x01 },
		{ OV511_I2C_BUS, 0x23, 0x00 },
		{ OV511_I2C_BUS, 0x24, 0x10 },
		{ OV511_I2C_BUS, 0x25, 0x8a },
		{ OV511_I2C_BUS, 0x27, 0xe2 },
		{ OV511_I2C_BUS, 0x2a, 0x00 },
		{ OV511_I2C_BUS, 0x2c, 0xfe },
		{ OV511_I2C_BUS, 0x30, 0x71 },
		{ OV511_I2C_BUS, 0x31, 0x60 },
		{ OV511_I2C_BUS, 0x32, 0x26 },
		{ OV511_I2C_BUS, 0x33, 0x20 },
		{ OV511_I2C_BUS, 0x34, 0x48 },
		{ OV511_I2C_BUS, 0x12, 0x24 },
		{ OV511_I2C_BUS, 0x11, 0x01 },
		{ OV511_I2C_BUS, 0x0c, 0x24 },
		{ OV511_I2C_BUS, 0x0d, 0x24 },
		{ OV511_DONE_BUS, 0x0, 0x00 },
	};

	PDEBUG (4, "starting configuration");

	/* This looks redundant, but is necessary for WebCam 3 */
	if (ov51x_set_slave_ids(ov511, OV7xx0_I2C_WRITE_ID,
				OV7xx0_I2C_READ_ID) < 0)
		return -1;

	if (ov51x_init_ov_sensor(ov511) >= 0) {
		PDEBUG(1, "OV7xx0 sensor initalized (method 1)");
	} else {
		/* Reset the 76xx */
		if (ov511_i2c_write(dev, 0x12, 0x80) < 0) return -1;

		/* Wait for it to initialize */
		schedule_timeout (1 + 150 * HZ / 1000);

		i = 0;
		success = 0;
		while (i <= i2c_detect_tries) {
			if ((ov511_i2c_read(dev, OV7610_REG_ID_HIGH) == 0x7F) &&
			    (ov511_i2c_read(dev, OV7610_REG_ID_LOW) == 0xA2)) {
				success = 1;
				break;
			} else {
				i++;
			}
		}

// Was (i == i2c_detect_tries) previously. This obviously used to always report
// success. Whether anyone actually depended on that bug is unknown
		if ((i >= i2c_detect_tries) && (success == 0)) {
			err("Failed to read sensor ID. You might not have an OV7610/20,");
			err("or it may be not responding. Report this to mmcclell@bigfoot.com");
			err("This is only a warning. You can attempt to use your camera anyway");
// Only issue a warning for now  
//			return -1;
		} else {
			PDEBUG(1, "OV7xx0 initialized in %d attempt(s) (method 2)", i+1);
		}
	}

	/* Detect sensor if user didn't use override param */
	if (sensor == 0) {
		rc = ov511_i2c_read(dev, OV7610_REG_COM_I);

		if (rc < 0) {
			err("Error detecting sensor type");
			return -1;
		} else if((rc & 3) == 3) {
			info("Sensor is an OV7610");
			ov511->sensor = SEN_OV7610;
		} else if((rc & 3) == 1) {
			
			/* I don't know what's different about the 76BE yet */
			if (ov511_i2c_read(dev, 0x15) & 1)
				info("Sensor is an OV7620AE");
			else
				info("Sensor is an OV76BE");

			/* OV511+ will return all zero isoc data unless we
			 * configure the sensor as a 7620. Someone needs to
			 * find the exact reg. setting that causes this. */
			if (ov511->bridge == BRG_OV511PLUS) {
				info("Enabling 511+/7620AE workaround");
				ov511->sensor = SEN_OV7620;
			} else {
				ov511->sensor = SEN_OV7620AE;
			}
		} else if((rc & 3) == 0) {
			info("Sensor is an OV7620");
			ov511->sensor = SEN_OV7620;
		} else {
			err("Unknown image sensor version: %d", rc & 3);
			return -1;
		}
	} else {	/* sensor != 0; user overrode detection */
		ov511->sensor = sensor;
		info("Sensor set to type %d", ov511->sensor);
	}

	if (ov511->sensor == SEN_OV7620) {
		PDEBUG(4, "Writing 7620 registers");
		if (ov511_write_regvals(dev, aRegvalsNorm7620))
			return -1;
	} else {
		PDEBUG(4, "Writing 7610 registers");
		if (ov511_write_regvals(dev, aRegvalsNorm7610))
			return -1;
	}

	/* Set sensor-specific vars */
	ov511->maxwidth = 640;
	ov511->maxheight = 480;

	if (aperture < 0) {          /* go with the default */
		if (ov511_i2c_write(dev, 0x26, 0xa2) < 0) return -1;
	} else if (aperture <= 0xf) {  /* user overrode default */
		if (ov511_i2c_write(dev, 0x26, (aperture << 4) + 2) < 0)
			return -1;
	} else {
		err("Invalid setting for aperture; legal value: 0 - 15");
		return -1;
	}

	if (autoadjust) {
		if (ov511_i2c_write(dev, 0x13, 0x01) < 0) return -1;
		if (ov511_i2c_write(dev, 0x2d, 
		     ov511->sensor==SEN_OV7620?0x91:0x93) < 0) return -1;
	} else {
		if (ov511_i2c_write(dev, 0x13, 0x00) < 0) return -1;
		if (ov511_i2c_write(dev, 0x2d, 
		     ov511->sensor==SEN_OV7620?0x81:0x83) < 0) return -1;
		ov511_i2c_write(dev, 0x28, ov511_i2c_read(dev, 0x28) | 8);
	}

	return 0;
}

/* This initializes the OV6620, OV6630, OV6630AE, or OV6630AF sensor. */
static int ov6xx0_configure(struct usb_ov511 *ov511)
{
	struct usb_device *dev = ov511->dev;
	int rc;

	static struct ov511_regvals aRegvalsNorm6x20[] = {
		{ OV511_I2C_BUS, 0x12, 0x80 }, /* reset */
		{ OV511_I2C_BUS, 0x11, 0x01 },
		{ OV511_I2C_BUS, 0x03, 0x60 },
		{ OV511_I2C_BUS, 0x05, 0x7f }, /* For when autoadjust is off */
		{ OV511_I2C_BUS, 0x07, 0xa8 },
		/* The ratio of 0x0c and 0x0d  controls the white point */
		{ OV511_I2C_BUS, 0x0c, 0x24 },
		{ OV511_I2C_BUS, 0x0d, 0x24 },
		{ OV511_I2C_BUS, 0x12, 0x28 }, /* Enable AGC */
		{ OV511_I2C_BUS, 0x14, 0x04 },
		/* 0x16: 0x06 helps frame stability with moving objects */
		{ OV511_I2C_BUS, 0x16, 0x06 },
//		{ OV511_I2C_BUS, 0x20, 0x30 }, /* Aperture correction enable */
		{ OV511_I2C_BUS, 0x26, 0xb2 }, /* BLC enable */
		{ OV511_I2C_BUS, 0x28, 0x05 }, /* Selects RGB format if RGB on */
		{ OV511_I2C_BUS, 0x2a, 0x04 }, /* Disable framerate adjust */
//		{ OV511_I2C_BUS, 0x2b, 0xac }, /* Framerate; Set 2a[7] first */
		{ OV511_I2C_BUS, 0x34, 0xd2 }, /* Max A/D range */
		{ OV511_I2C_BUS, 0x38, 0x8b },
		{ OV511_I2C_BUS, 0x39, 0x40 },
		
		{ OV511_I2C_BUS, 0x3c, 0x39 }, /* Enable AEC mode changing */
		{ OV511_I2C_BUS, 0x3c, 0x3c }, /* Change AEC mode */
		{ OV511_I2C_BUS, 0x3c, 0x24 }, /* Disable AEC mode changing */

		{ OV511_I2C_BUS, 0x3d, 0x80 },
		/* These next two registers (0x4a, 0x4b) are undocumented. They 
		 * control the color balance */
		{ OV511_I2C_BUS, 0x4a, 0x80 },
		{ OV511_I2C_BUS, 0x4b, 0x80 },
		{ OV511_I2C_BUS, 0x4d, 0xd2 }, /* This reduces noise a bit */
		{ OV511_I2C_BUS, 0x4e, 0xc1 },
		{ OV511_I2C_BUS, 0x4f, 0x04 },
// Do 50-53 have any effect?
// Toggle 0x12[2] off and on here?
		{ OV511_DONE_BUS, 0x0, 0x00 },
	};

	/* This chip is undocumented so many of these are guesses. OK=verified,
	 * A=Added since 6620, U=unknown function (not a 6620 reg) */
	static struct ov511_regvals aRegvalsNorm6x30[] = {
	/*OK*/	{ OV511_I2C_BUS, 0x12, 0x80 }, /* reset */
	/*00?*/	{ OV511_I2C_BUS, 0x11, 0x01 },
	/*OK*/	{ OV511_I2C_BUS, 0x03, 0x60 },
	/*0A?*/	{ OV511_I2C_BUS, 0x05, 0x7f }, /* For when autoadjust is off */
		{ OV511_I2C_BUS, 0x07, 0xa8 },
		/* The ratio of 0x0c and 0x0d  controls the white point */
	/*OK*/	{ OV511_I2C_BUS, 0x0c, 0x24 },
	/*OK*/	{ OV511_I2C_BUS, 0x0d, 0x24 },
	/*A*/	{ OV511_I2C_BUS, 0x0e, 0x20 },
//	/*24?*/	{ OV511_I2C_BUS, 0x12, 0x28 }, /* Enable AGC */
//		{ OV511_I2C_BUS, 0x12, 0x24 }, /* Enable AGC */
//	/*A*/	{ OV511_I2C_BUS, 0x13, 0x21 },
//	/*04?*/	{ OV511_I2C_BUS, 0x14, 0x80 },
		/* 0x16: 0x06 helps frame stability with moving objects */
	/*03?*/	{ OV511_I2C_BUS, 0x16, 0x06 },
//	/*OK*/	{ OV511_I2C_BUS, 0x20, 0x30 }, /* Aperture correction enable */
		// 21 & 22? The suggested values look wrong. Go with default
	/*A*/	{ OV511_I2C_BUS, 0x23, 0xc0 },
	/*A*/	{ OV511_I2C_BUS, 0x25, 0x9a }, // Check this against default
//	/*OK*/	{ OV511_I2C_BUS, 0x26, 0xb2 }, /* BLC enable */
//	/*04?*/	{ OV511_I2C_BUS, 0x28, 0x05 }, /* Selects RGB format if RGB on */
	/*OK*/	{ OV511_I2C_BUS, 0x2a, 0x04 }, /* Disable framerate adjust */
//	/*OK*/	{ OV511_I2C_BUS, 0x2b, 0xac }, /* Framerate; Set 2a[7] first */
//	/*U*/	{ OV511_I2C_BUS, 0x2c, 0xa0 },
//	/*A*/	{ OV511_I2C_BUS, 0x33, 0x26 }, // Reserved bits on 6620
		// Reg 0x34: More 6620 reserved junk
//	/*d2?*/	{ OV511_I2C_BUS, 0x34, 0x03 }, /* Max A/D range */
//	/*U*/	{ OV511_I2C_BUS, 0x36, 0x8f }, // May not be necessary
//	/*U*/	{ OV511_I2C_BUS, 0x37, 0x80 }, // May not be necessary
//	/*8b?*/	{ OV511_I2C_BUS, 0x38, 0x83 },
//	/*40?*/	{ OV511_I2C_BUS, 0x39, 0xc0 }, // 6630 adds bit 7
		//Check these again
//		{ OV511_I2C_BUS, 0x3c, 0x39 }, /* Enable AEC mode changing */
//		{ OV511_I2C_BUS, 0x3c, 0x3c }, /* Change AEC mode */
//		{ OV511_I2C_BUS, 0x3c, 0x24 }, /* Disable AEC mode changing */

	/*OK*/	{ OV511_I2C_BUS, 0x3d, 0x80 },
//	/*A*/	{ OV511_I2C_BUS, 0x3f, 0x0e }, // Reserved bits on 6620. Default may be safer.

//	/*U*/	{ OV511_I2C_BUS, 0x40, 0x00 },
//	/*U*/	{ OV511_I2C_BUS, 0x41, 0x00 },
//	/*U*/	{ OV511_I2C_BUS, 0x42, 0x80 },
//	/*U*/	{ OV511_I2C_BUS, 0x43, 0x3f },
//	/*U*/	{ OV511_I2C_BUS, 0x44, 0x80 },
//	/*U*/	{ OV511_I2C_BUS, 0x45, 0x20 },
//	/*U*/	{ OV511_I2C_BUS, 0x46, 0x20 },
//	/*U*/	{ OV511_I2C_BUS, 0x47, 0x80 },
//	/*U*/	{ OV511_I2C_BUS, 0x48, 0x7f },
//	/*U*/	{ OV511_I2C_BUS, 0x49, 0x00 },

		/* These next two registers (0x4a, 0x4b) are undocumented. They 
		 * control the color balance */
//	/*OK?*/	{ OV511_I2C_BUS, 0x4a, 0x80 }, // Check these
//	/*OK?*/	{ OV511_I2C_BUS, 0x4b, 0x80 },
//	/*U*/	{ OV511_I2C_BUS, 0x4c, 0xd0 }, // These high registers are always suspect
	/*d2?*/	{ OV511_I2C_BUS, 0x4d, 0x10 }, /* This reduces noise a bit */
	/*c1?*/	{ OV511_I2C_BUS, 0x4e, 0x40 },
	/*04?*/	{ OV511_I2C_BUS, 0x4f, 0x07 },

//	/*U*/	{ OV511_I2C_BUS, 0x50, 0xff }, // These are all undocumented (by 6620 docs)
	/*U*/	{ OV511_I2C_BUS, 0x54, 0x23 },
//	/*U*/	{ OV511_I2C_BUS, 0x55, 0xff },
//	/*U*/	{ OV511_I2C_BUS, 0x56, 0x12 },
	/*U*/	{ OV511_I2C_BUS, 0x57, 0x81 },
//	/*U*/	{ OV511_I2C_BUS, 0x58, 0x75 },
	/*U*/	{ OV511_I2C_BUS, 0x59, 0x01 },
	/*U*/	{ OV511_I2C_BUS, 0x5a, 0x2c },
	/*U*/	{ OV511_I2C_BUS, 0x5b, 0x0f },
//	/*U*/	{ OV511_I2C_BUS, 0x5c, 0x10 },

// Set 0x3d to 0x80 here?
// Set 0x27 to 0xa6 here?
// Toggle 0x12[2] off and on here?
		{ OV511_DONE_BUS, 0x0, 0x00 },
	};

	PDEBUG (4, "starting sensor configuration");
	
	if (ov51x_init_ov_sensor(ov511) < 0) {
		err("Failed to read sensor ID. You might not have an OV6xx0,");
		err("or it may be not responding. Report this to mmcclell@bigfoot.com");
		return -1;
	} else {
		PDEBUG(1, "OV6xx0 sensor detected");
	}

	/* Detect sensor if user didn't use override param */
	if (sensor == 0) {
		rc = ov511_i2c_read(dev, OV7610_REG_COM_I);

		if (rc < 0) {
			err("Error detecting sensor type");
			return -1;
		} else if((rc & 3) == 0) {
			info("Sensor is an OV6630");
			ov511->sensor = SEN_OV6630;
		} else if((rc & 3) == 1) {
			info("Sensor is an OV6620");
			ov511->sensor = SEN_OV6620;
		} else if((rc & 3) == 2) {
			info("Sensor is an OV6630AE");
			ov511->sensor = SEN_OV6630;
		} else if((rc & 3) == 3) {
			info("Sensor is an OV6630AF");
			ov511->sensor = SEN_OV6630;
		} 
	} else {	/* sensor != 0; user overrode detection */
		ov511->sensor = sensor;
		info("Sensor set to type %d", ov511->sensor);
	}

	/* Set sensor-specific vars */
	ov511->maxwidth = 352;
	ov511->maxheight = 288;

	if (ov511->sensor == SEN_OV6620) {
		PDEBUG(4, "Writing 6x20 registers");
		if (ov511_write_regvals(dev, aRegvalsNorm6x20))
			return -1;
	} else {
		PDEBUG(4, "Writing 6x30 registers");
		if (ov511_write_regvals(dev, aRegvalsNorm6x30))
			return -1;
	}
	
	if (aperture < 0) {          /* go with the default */
		if (ov511_i2c_write(dev, 0x26, 0xa2) < 0) return -1;
	} else if (aperture <= 0xf) {  /* user overrode default */
		if (ov511_i2c_write(dev, 0x26, (aperture << 4) + 2) < 0)
			return -1;
	} else {
		err("Invalid setting for aperture; legal value: 0 - 15");
		return -1;
	}

	if (autoadjust) {
		if (ov511_i2c_write_mask(dev, 0x13, 0x01, 0x01) < 0) return -1;
		if (ov511_i2c_write(dev, 0x2d, 0x99) < 0) return -1;
	} else {
		if (ov511_i2c_write_mask(dev, 0x13, 0x00, 0x01) < 0) return -1;
		if (ov511_i2c_write(dev, 0x2d, 0x89) < 0) return -1;
		if (ov511_i2c_write_mask(dev, 0x28, 0x08, 0x08) < 0) return -1;
	}

	return 0;
}

/* This initializes the OV511/OV511+ and the sensor */
static int ov511_configure(struct usb_ov511 *ov511)
{
	struct usb_device *dev = ov511->dev;
	int i;

	static struct ov511_regvals aRegvalsInit511[] = {
		{ OV511_REG_BUS, OV511_REG_SYSTEM_RESET, 0x7f },
	 	{ OV511_REG_BUS, OV511_REG_SYSTEM_INIT, 0x01 },
	 	{ OV511_REG_BUS, OV511_REG_SYSTEM_RESET, 0x7f },
		{ OV511_REG_BUS, OV511_REG_SYSTEM_INIT, 0x01 },
		{ OV511_REG_BUS, OV511_REG_SYSTEM_RESET, 0x3f },
		{ OV511_REG_BUS, OV511_REG_SYSTEM_INIT, 0x01 },
		{ OV511_REG_BUS, OV511_REG_SYSTEM_RESET, 0x3d },
		{ OV511_DONE_BUS, 0x0, 0x00},
	};

	static struct ov511_regvals aRegvalsNorm511[] = {
		{ OV511_REG_BUS, OV511_REG_DRAM_ENABLE_FLOW_CONTROL, 0x01 },
		{ OV511_REG_BUS, OV511_REG_SYSTEM_SNAPSHOT, 0x01 },
		{ OV511_REG_BUS, OV511_REG_SYSTEM_SNAPSHOT, 0x03 },
		{ OV511_REG_BUS, OV511_REG_SYSTEM_SNAPSHOT, 0x01 },
		{ OV511_REG_BUS, OV511_REG_FIFO_BITMASK, 0x1f },
		{ OV511_REG_BUS, OV511_OMNICE_ENABLE, 0x00 },
		{ OV511_REG_BUS, OV511_OMNICE_LUT_ENABLE, 0x03 },
		{ OV511_DONE_BUS, 0x0, 0x00 },
	};

	PDEBUG(4, "");

	ov511->customid = ov511_reg_read(dev, OV511_REG_SYSTEM_CUSTOM_ID);
	if (ov511->customid < 0) {
		err("Unable to read camera bridge registers");
		return -EIO;
	}

	ov511->desc = -1;
	PDEBUG (4, "CustomID = %d", ov511->customid);
	for (i = 0; clist[i].id >= 0; i++) {
		if (ov511->customid == clist[i].id) {
			info("camera: %s", clist[i].description);
			ov511->desc = i;
			break;
		}
	}

	/* Lifeview USB Life TV not supported */
	if (clist[i].id == 38) {
		err("This device is not supported yet.");
		return -EIO;
	}

	if (clist[i].id == -1) {
		err("Camera type (%d) not recognized", ov511->customid);
		err("Please notify mmcclell@bigfoot.com of the name, manufacturer, model,");
		err("and camera type number of your camera. Also include the");
		err("complete output of the detection process.");
	}

	memcpy(&ov511->vdev, &ov511_template, sizeof(ov511_template));

	for (i = 0; i < OV511_NUMFRAMES; i++)
		init_waitqueue_head(&ov511->frame[i].wq);

	init_waitqueue_head(&ov511->wq);

	if (video_register_device(&ov511->vdev, VFL_TYPE_GRABBER) < 0) {
		err("video_register_device failed");
		return -EBUSY;
	}

	if (ov511_write_regvals(dev, aRegvalsInit511)) goto error;

	if (ov511->led_policy == LED_OFF || ov511->led_policy == LED_AUTO)
		ov511_led_control(ov511, 0);

	if (ov511_write_regvals(dev, aRegvalsNorm511)) goto error;
	if (compress) {
		ov511_reg_write(dev, 0x70, phy);
		ov511_reg_write(dev, 0x71, phuv);
		ov511_reg_write(dev, 0x72, pvy);
		ov511_reg_write(dev, 0x73, pvuv);
		ov511_reg_write(dev, 0x74, qhy);
		ov511_reg_write(dev, 0x75, qhuv);
		ov511_reg_write(dev, 0x76, qvy);
		ov511_reg_write(dev, 0x77, qvuv);
		if (ov511_upload_quan_tables(dev) < 0)
			err("Error uploading quantization tables");
	}

	ov511_set_packet_size(ov511, 0);

	ov511->snap_enabled = snapshot;	

	/* Test for 7xx0 */
	if (ov51x_set_slave_ids(ov511, OV7xx0_I2C_WRITE_ID,
				OV7xx0_I2C_READ_ID) < 0)
		return -1;

	if (ov511_i2c_write(dev, 0x12, 0x80) < 0) {
		/* Test for 6xx0 */
		if (ov51x_set_slave_ids(ov511, OV6xx0_I2C_WRITE_ID,
					OV6xx0_I2C_READ_ID) < 0)
			return -1;

		if (ov511_i2c_write(dev, 0x12, 0x80) < 0) {
			/* Test for 8xx0 */
			if (ov51x_set_slave_ids(ov511, OV8xx0_I2C_WRITE_ID,
						OV8xx0_I2C_READ_ID))
				return -1;

			if (ov511_i2c_write(dev, 0x12, 0x80) < 0) {
				err("Can't determine sensor slave IDs");
 				goto error;
			} else {
				err("Detected unsupported OV8xx0 sensor");
				goto error;
			}
		} else {
			if(ov6xx0_configure(ov511) < 0) {
				err("failed to configure OV6xx0");
 				goto error;
			}
		}
	} else {
		if(ov7xx0_configure(ov511) < 0) {
			err("failed to configure OV7xx0");
	 		goto error;
		}
	}

	if (ov51x_set_default_params(ov511) < 0)
		goto error;

	return 0;

error:
	err("OV511 Config failed");

	#if defined(CONFIG_PROC_FS) && defined(CONFIG_VIDEO_PROC_FS)
	/* Safe to call even if entry doesn't exist */
	destroy_proc_ov511_cam((struct usb_ov511 *)dev);
	#endif

	video_unregister_device(&ov511->vdev);
	usb_driver_release_interface(&ov511_driver,
		&dev->actconfig->interface[ov511->iface]);

	return -EBUSY;
}

/* This initializes the OV518/OV518+ and the sensor */
static int ov518_configure(struct usb_ov511 *ov511)
{
	struct usb_device *dev = ov511->dev;
	int i;

	static struct ov511_regvals aRegvalsInit518[] = {
		{ OV511_REG_BUS, OV511_REG_SYSTEM_RESET, 0x40 },
	 	{ OV511_REG_BUS, OV511_REG_SYSTEM_INIT, 0xe1 },
	 	{ OV511_REG_BUS, OV511_REG_SYSTEM_RESET, 0x3e },
		{ OV511_REG_BUS, OV511_REG_SYSTEM_INIT, 0xe1 },
		{ OV511_REG_BUS, OV511_REG_SYSTEM_RESET, 0x00 },
		{ OV511_REG_BUS, OV511_REG_SYSTEM_INIT, 0xe1 },
		{ OV511_REG_BUS, 0x5d, 0x03 },
		{ OV511_DONE_BUS, 0x0, 0x00},
	};

#if 0
	/* Old sequence - Note: the mnemonics for these registers may not be
	 * right with the OV518, but their values are right. */
	static struct ov511_regvals aRegvalsNorm518[] = {
		{ OV511_REG_BUS, OV511_REG_DRAM_ENABLE_FLOW_CONTROL, 0x09 },
		{ OV511_REG_BUS, OV511_REG_SYSTEM_SNAPSHOT, 0x02 },
		{ OV511_REG_BUS, OV511_REG_SYSTEM_SNAPSHOT, 0x00 },
		{ OV511_REG_BUS, OV511_REG_FIFO_BITMASK, 0x0f },
		{ OV511_REG_BUS, 0x24, 0x9f },
		{ OV511_REG_BUS, 0x25, 0x90 },
		{ OV511_REG_BUS, OV511_REG_SYSTEM_CLOCK_DIVISOR, 0x04 }, /* ? */
		{ OV511_REG_BUS, OV511_OMNICE_ENABLE, 0x00 },
		{ OV511_REG_BUS, OV511_OMNICE_LUT_ENABLE, 0x00 },
		{ OV511_DONE_BUS, 0x0, 0x00 },
	};
#else
	/* New values, based on Windows driver. Since what they do is not
	 * known yet, this may be incorrect. */
	static struct ov511_regvals aRegvalsNorm518[] = {
		{ OV511_REG_BUS, 0x52, 0x02 },
		{ OV511_REG_BUS, 0x52, 0x01 },
		{ OV511_REG_BUS, 0x31, 0x0f },
		{ OV511_REG_BUS, 0x5d, 0x03 },
		{ OV511_REG_BUS, 0x24, 0x9f },
		{ OV511_REG_BUS, 0x25, 0x90 },
		{ OV511_REG_BUS, 0x20, 0x08 },
		{ OV511_REG_BUS, 0x51, 0x04 },
//		{ OV511_REG_BUS, 0x71, 0x19 },
		{ OV511_DONE_BUS, 0x0, 0x00 },
	};
#endif
	PDEBUG(4, "");

	/* First 5 bits of custom ID reg are a revision ID on OV518 */
	info("Camera revision %d", 0x1F & ov511_reg_read(dev, OV511_REG_SYSTEM_CUSTOM_ID));

	memcpy(&ov511->vdev, &ov511_template, sizeof(ov511_template));

	for (i = 0; i < OV511_NUMFRAMES; i++)
		init_waitqueue_head(&ov511->frame[i].wq);

	init_waitqueue_head(&ov511->wq);

	if (video_register_device(&ov511->vdev, VFL_TYPE_GRABBER) < 0) {
		err("video_register_device failed");
		return -EBUSY;
	}

	if (ov511_write_regvals(dev, aRegvalsInit518)) goto error;

	/* LED is off by default with OV518; have to explicitly turn it on */
	if (ov511->led_policy == LED_OFF || ov511->led_policy == LED_AUTO)
		ov511_led_control(ov511, 0);
	else
		ov511_led_control(ov511, 1);

	if (compress) {
		compress = 0;
		warn("Compression disabled on OV518+");
	}

	if (ov511_write_regvals(dev, aRegvalsNorm518)) goto error;
	/* The next setting may not be necessary */
	if (ov511_reg_write_mask(dev, 0x57,0x00, 0x02) < 0) goto error;

	ov511_set_packet_size(ov511, 0);

	ov511->snap_enabled = snapshot;

	/* Test for 76xx */
	if (ov51x_set_slave_ids(ov511, OV7xx0_I2C_WRITE_ID,
				OV7xx0_I2C_READ_ID) < 0)
		return -1;

	/* The OV518 must be more aggressive about sensor detection since
	 * I2C write will never fail if the sensor is not present. We have
	 * to try to initialize the sensor to detect its presence */

	if (ov51x_init_ov_sensor(ov511) < 0) {
		/* Test for 6xx0 */
		if (ov51x_set_slave_ids(ov511, OV6xx0_I2C_WRITE_ID,
					OV6xx0_I2C_READ_ID) < 0)
			return -1;

		if (ov51x_init_ov_sensor(ov511) < 0) {
			/* Test for 8xx0 */
			if (ov51x_set_slave_ids(ov511, OV6xx0_I2C_WRITE_ID,
						OV6xx0_I2C_READ_ID) < 0)
				return -1;

			if (ov51x_init_ov_sensor(ov511) < 0) {
				err("Can't determine sensor slave IDs");
 				goto error;
			} else {
				err("Detected unsupported OV8xx0 sensor");
				goto error;
			}
		} else {
			if(ov6xx0_configure(ov511) < 0) {
				err("failed to configure OV6xx0");
 				goto error;
			}
		}
	} else {
		if(ov7xx0_configure(ov511) < 0) {
			err("failed to configure OV7xx0");
	 		goto error;
		}
	}

	if (ov51x_set_default_params(ov511) < 0)
		goto error;

	return 0;

error:
	err("OV518 Config failed");

	#if defined(CONFIG_PROC_FS) && defined(CONFIG_VIDEO_PROC_FS)
	/* Safe to call even if entry doesn't exist */
	destroy_proc_ov511_cam((struct usb_ov511 *)dev);
	#endif

	video_unregister_device(&ov511->vdev);
	usb_driver_release_interface(&ov511_driver,
		&dev->actconfig->interface[ov511->iface]);

	return -EBUSY;
}


/****************************************************************************
 *
 *  USB routines
 *
 ***************************************************************************/

static void *
ov511_probe(struct usb_device *dev, unsigned int ifnum,
	const struct usb_device_id *id)
{
	struct usb_interface_descriptor *interface;
	struct usb_ov511 *ov511;

	PDEBUG(1, "probing for device...");

	/* We don't handle multi-config cameras */
	if (dev->descriptor.bNumConfigurations != 1)
		return NULL;

	interface = &dev->actconfig->interface[ifnum].altsetting[0];

	/* Checking vendor/product should be enough, but what the hell */
	if (interface->bInterfaceClass != 0xFF)
		return NULL;
	if (interface->bInterfaceSubClass != 0x00)
		return NULL;

	/* Since code below may sleep, we use this as a lock */
	MOD_INC_USE_COUNT;

	if ((ov511 = kmalloc(sizeof(*ov511), GFP_KERNEL)) == NULL) {
		err("couldn't kmalloc ov511 struct");
		goto error;
	}

	memset(ov511, 0, sizeof(*ov511));

	ov511->dev = dev;
	ov511->iface = interface->bInterfaceNumber;
	ov511->led_policy = led;

	switch (dev->descriptor.idProduct) {
	case 0x0511:
		info("USB OV511 camera found");
		ov511->bridge = BRG_OV511;
		break;
	case 0xA511:
		info("USB OV511+ camera found");
		ov511->bridge = BRG_OV511PLUS;
		break;
	case 0x0518:
		info("USB OV518+ camera found");
		ov511->bridge = BRG_OV518PLUS;
		break;
	case 0x0002:
		if (dev->descriptor.idVendor != 0x0813)
			goto error;
		info("Intel Play Me2Cam (OV511+) found");
		ov511->bridge = BRG_OV511PLUS;
		break;
	default:
		err("Unknown product ID 0x%x", dev->descriptor.idProduct);
		goto error;
	}

	/* Workaround for some applications that want data in RGB
	 * instead of BGR. */
	if (force_rgb)
		info("data format set to RGB");

	if (ov511->bridge == BRG_OV518PLUS) {
		if (ov518_configure(ov511) < 0)
			goto error;
	} else {
		if (ov511_configure(ov511) < 0)
			goto error;
	}

	init_MUTEX(&ov511->lock);	/* to 1 == available */
	init_MUTEX(&ov511->buf_lock);
	init_MUTEX(&ov511->param_lock);
	ov511->buf_state = BUF_NOT_ALLOCATED;

#ifdef OV511_DEBUG
	if (dump_bridge)
		ov511_dump_regs(dev);
#endif

	PDEBUG(1, "Camera initialization successful");

	MOD_DEC_USE_COUNT;
     	return ov511;

error:
	err("Camera initialization failed");
	if (ov511) {
		kfree(ov511);
		ov511 = NULL;
	}

	MOD_DEC_USE_COUNT;
	return NULL;
}


static void
ov511_disconnect(struct usb_device *dev, void *ptr)
{
	struct usb_ov511 *ov511 = (struct usb_ov511 *) ptr;
	int n;

	MOD_INC_USE_COUNT;

	PDEBUG(3, "");

	/* We don't want people trying to open up the device */
	if (!ov511->user)
		video_unregister_device(&ov511->vdev);
	else
		PDEBUG(3, "Device open...deferring video_unregister_device");

	for (n = 0; n < OV511_NUMFRAMES; n++)
		ov511->frame[n].grabstate = FRAME_ERROR;

	ov511->curframe = -1;

	/* This will cause the process to request another frame */
	for (n = 0; n < OV511_NUMFRAMES; n++)
		if (waitqueue_active(&ov511->frame[n].wq))
			wake_up_interruptible(&ov511->frame[n].wq);
	if (waitqueue_active(&ov511->wq))
		wake_up_interruptible(&ov511->wq);

	ov511->streaming = 0;

	/* Unschedule all of the iso td's */
	for (n = OV511_NUMSBUF - 1; n >= 0; n--) {
		if (ov511->sbuf[n].urb) {
			ov511->sbuf[n].urb->next = NULL;
			usb_unlink_urb(ov511->sbuf[n].urb);
			usb_free_urb(ov511->sbuf[n].urb);
			ov511->sbuf[n].urb = NULL;
		}
	}

	usb_driver_release_interface(&ov511_driver,
		&ov511->dev->actconfig->interface[ov511->iface]);
	ov511->dev = NULL;

#if defined(CONFIG_PROC_FS) && defined(CONFIG_VIDEO_PROC_FS)
        destroy_proc_ov511_cam(ov511);
#endif

	/* Free the memory */
	if (ov511 && !ov511->user) {
		ov511_dealloc(ov511, 1);
		kfree(ov511);
		ov511 = NULL;
	}

	MOD_DEC_USE_COUNT;
}

static struct usb_driver ov511_driver = {
	name:		"ov511",
	id_table:       device_table,
	probe:		ov511_probe,
	disconnect:	ov511_disconnect
};


/****************************************************************************
 *
 *  Module routines
 *
 ***************************************************************************/

static int __init usb_ov511_init(void)
{
#if defined(CONFIG_PROC_FS) && defined(CONFIG_VIDEO_PROC_FS)
        proc_ov511_create();
#endif

	if (usb_register(&ov511_driver) < 0)
		return -1;

	info("ov511 driver version %s registered", version);

	return 0;
}

static void __exit usb_ov511_exit(void)
{
	usb_deregister(&ov511_driver);
	info("driver deregistered");

#if defined(CONFIG_PROC_FS) && defined(CONFIG_VIDEO_PROC_FS)
        proc_ov511_destroy();
#endif
}

module_init(usb_ov511_init);
module_exit(usb_ov511_exit);
